/*
 * bbftpc/bbftp.c
 * Copyright (C) 1999, 2000, 2001, 2002 IN2P3, CNRS
 * bbftp@in2p3.fr
 * http://doc.in2p3.fr/bbftp
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */ 

/****************************************************************************


  
  
 bbftp.c  v 2.0.0  2001/03/01   - Complete rewriting for version 2.0.0
          v 2.0.1  2001/04/19   - Correct indentation 
                                - Verify if -s is present on shhremotecmd
                                  and add it if not (in order to fit v 1.9.4-tja1
                                  behaviour)
          v 2.1.0 2001/05/21    - Add debug
                                - Add -m option to have special output
                                - Set -v before all options
                                - Correct case where the last line in control
                                  file has no CR
          v 2.1.2 2001/11/19    - Fix COS 0 case
          v 2.2.0 2001/10/03    - Add certificate authentication mode

*****************************************************************************/
#include "bbftp.h"
#include <sys/types.h>

#include <errno.h> 
#include <fcntl.h>
#include <netdb.h>
#include <pwd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
//#include <sys/time.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "client.h"
#include "structures.h"
#include "client_proto.h"
#include "common.h"
#include "config.h"
#include "version.h"

extern int getoptt (int argc, char *const *argv, const char *shortopts);

#define SETTOZERO    0
#define SETTOONE     1

#define SSHREMOTECMD "bbftpd -s -e -l information"
#define SSHCMD "ssh -q"


#ifdef PRIVATE_AUTH
#define OPTIONS "qbcC:de:f:i:l:mM:o:O:p:P:r:R:tu:vVw:D::Whxj:n:"
#else
#define OPTIONS "qbcC:de:E:f:g::i:I:l:L:mM:o:O:p:r:R:s::S::tu:vVw:D::Whxj:n:"
#endif
/*
#endif
*/


/*
 * Range for the ephemeral ports for data connections
 */
int     pasvport_min = 0 ;
int     pasvport_max = 0 ;

/*
** timestamp:
**      Every message on standart error and standart output are
**      timestamped
*/
int     timestamp = SETTOZERO ;

/*
** debug:
**      Set to one to print more debugging information
*/
int     debug = SETTOZERO ;
/*
** verbose:
**      Set to one to print  information
*/
int     verbose = SETTOZERO ;
/*
** warning:
**      Set to one to print warning to stderr
*/
int     warning = SETTOZERO ;
/*
** statoutput:
**      Set to one for special output
*/
int     statoutput = SETTOZERO ;
/*
** globaltrymax:
**      Number of try in case or recoverable error
*/
int     globaltrymax = NBTRYMAX ;
/*
** newcontrolport:
**      Control port to be used 
*/
int     newcontrolport = CONTROLPORT ;
/*
** usessh:
**      Set to one when using ssh to start the remote daemon
*/
int     usessh = SETTOZERO ;
/*
** sshbatchmode:
**      This is set to non-zero if running in batch mode (that is, password
**      and passphrase queries are not allowed). 
*/
int     sshbatchmode  = SETTOZERO ;
/*
** For ssh/RSA
*/
char    *sshidentityfile = "~/.ssh/id_rsa";
char    *sshremotecmd = NULL ;
char    *sshcmd = NULL ;
/*
** usecert:
**		Set to one if using certificate authentifaction
*/
int		usecert = SETTOZERO ;
/*
** useprivate:
**      Set to one if using private authentication
*/
int     useprivate = SETTOZERO ;
/*
** privatestr:
**      Pointer to a private string used for private authentication
**
*/
char    *privatestr = NULL ;
/* 
** hisctladdr:
**      the remote address
*/
struct sockaddr_in hisctladdr;

char *localips = "";

int natip = 0;

/*
** bbftprc:
**      Where to store the bbftprc file
*/
char    *bbftprc = NULL ;
/*
** Variable defining the local options :
** 
** localumask:
**      Local umask taken by the umask command at start and
**      modified by the setlocalumask command
**
** localrfio:
**      set to one when using rfio for local files
**
*/
int     localumask ;
/*
** Variables defining both side options :
**
** usetmpfile:
**      Set to one when using tmpname for file creation
**
** usegzipcompress:
**      Set to one when using tmpname for file creation
**
** keepaccess:
**      Set to one when keeping access time and modification
**
** keepmode:
**      Set to one when keeping file mode
**
** creatdir:
**      Set to one when automatic directory creation is needed
*/
int     sendwinsize     = 256 ;
int     recvwinsize     = 256 ;

int     nbport = 2 * MAXSTR;

int		ackto			= ACKTO;
int		recvcontrolto	= CONTROLSOCKTO;
int		sendcontrolto	= SENDCONTROLTO;
int		datato			= DATASOCKTO;

int crtfathers = 0;	/* Current number of simultaneous father processes */
int maxfathers = 1;	/* Maximum number of fathers, -1 means no limit */

/*
** Variables remote side options :
**
** remoterfio:
**      Set to one when rfio for remote file
**
** remoteumask:
**      if set to zero do not set remote umask
**
** remotecos:
**      if not set to zero do set remote cos
**
** remotedir :
**      if not set to NULL change dir after connection
**
*/
int     remoteumask = SETTOZERO ;
int     remotecos   = SETTOZERO ;
char    *remotedir = NULL ;
/*
** incontrolsock :
**      Define the control socket for reading
** outcontrolsock :
**      Define the control socket for writing
**
**      For normal use : incontrolsock = outcontrolsock
*/
int     incontrolsock ;
int     outcontrolsock ;

char    *hostname   = NULL ;
char    *username   = NULL ;
char    *password   = NULL ;

#ifdef CERTIFICATE_AUTH
char    *service    = NULL ;
char	*fullhostname = NULL;
#endif

/*
** castfd:
**      CASTOR file descriptor
*/
int     castfd = -1 ;
/*
** Parameters describing the transfer ********************
*/
int     transferoption = TROPT_TMP | TROPT_DIR | TROPT_MODE | TROPT_ACC | TROPT_UNIX /*| TROPT_MPATH*/;
int     filemode ;
char    lastaccess[9];
char    lastmodif[9];
int     buffersizeperstream = 256 ;
int     requestedstreamnumber ;
my64_t  filesize ;

/*
** curfilename :
**      Define the pointer to the current file
*/
char    *curfilename = NULL ;
/*
** realfilename :
**      Define the pointer to the real file (= curfilename if TROPT_TMP not
**      set)
*/
char    *realfilename   = NULL ;
unsigned short int     *myports        = NULL ;
int     *mysockets      = NULL ;


xbuffer	logmessage = { NULL, 0, 0 };	/* universal usage */
int	errcode = 0;			/* program exit code */

char *incoming = NULL;
char *outgoing = NULL;

int     background  = SETTOZERO ;

int main(argc, argv, envp)
    int argc;
    char **argv;
    char **envp;
{
    
    extern char *optarg;
    extern int optind, opterr, optopt;
/*
** Variable set by options
*/
    char    *inputfile  = NULL ;
    char    *resultfile = NULL ;
    char    *outputfile = NULL ;
    char    *errorfile  = NULL ;
    char    *bbftpcmd   = NULL ;

/*
** For local user
*/
    struct  passwd  *mypasswd = NULL;
    char    *bbftprcfile = NULL ;
    
    sSTAT    statbuf ;
    
    int     retcode ;

    int        i, k ;

    int     j ;
    int     infd = -1;
    int     resfd = -1;
    int     alluse ;

/*
** Get local umask 
*/
    localumask = umask(0) ;
/*
** and reset it to the correct value
*/
    umask(localumask) ;


/*
** First check for timestamp in order to have a common output
*/
    opterr = 0 ;
    while ((j = getoptt(argc, argv, OPTIONS)) != -1) {
        switch (j) {
            case 't' :{
                timestamp = SETTOONE ;
                break ;
            }
        }
    }

#ifdef PRIVATE_AUTH
    useprivate = SETTOONE ;
    usessh = SETTOZERO ;
#endif
    
/*
** Check for -v option
*/
    opterr = 0 ;
    optind = 1 ;
    while ((j = getoptt(argc, argv, OPTIONS)) != -1) {
        switch (j) {
            case 'v' :{
                printmessage(stderr,CASE_NORMAL,0,timestamp,"bbftpPRO version %s\n",VERSION) ;
                printmessage(stderr,CASE_NORMAL,0,timestamp,"Compiled with  :   default port %d\n",CONTROLPORT) ;


#ifdef PORT_RANGE
                printmessage(stderr,CASE_NORMAL,0,timestamp,"                   data ports range = %s \n", PORT_RANGE) ;
#endif


#ifdef WITH_GZIP
                printmessage(stderr,CASE_NORMAL,0,timestamp,"                   compression \n") ;
#endif
#ifdef WITH_SSL
                printmessage(stderr,CASE_NORMAL,0,timestamp,"                   encryption \n") ;
#endif
#ifdef WITH_RFIO
                printmessage(stderr,CASE_NORMAL,0,timestamp,"                   RFIO interface \n") ;
#endif
#ifdef WITH_RFIO64
                printmessage(stderr,CASE_NORMAL,0,timestamp,"                   RFIO64 interface \n") ;
#endif
#ifdef AFS
                printmessage(stderr,CASE_NORMAL,0,timestamp,"                   AFS authentication \n") ;
#endif
#ifdef PRIVATE_AUTH
                printmessage(stderr,CASE_NORMAL,0,timestamp,"                   private authentication \n") ;
#else
                printmessage(stderr,CASE_NORMAL,0,timestamp,"                   default ssh command = %s \n",SSHCMD) ;
                printmessage(stderr,CASE_NORMAL,0,timestamp,"                   default ssh remote command = %s \n",SSHREMOTECMD) ;
# ifdef CERTIFICATE_AUTH
				printmessage(stderr,CASE_NORMAL,0,timestamp,"                   certificate authentication enabled\n") ;
# endif
#endif
                printmessage(stderr,CASE_NORMAL,0,timestamp,"                   default number of tries = %d  \n",NBTRYMAX) ;
                printmessage(stderr,CASE_NORMAL,0,timestamp,"                   default sendwinsize = %d Kbytes\n",sendwinsize) ;
                printmessage(stderr,CASE_NORMAL,0,timestamp,"                   default recvwinsize = %d Kbytes\n",recvwinsize) ;
                printmessage(stderr,CASE_NORMAL,0,timestamp,"                   default number of streams = %d \n",nbport) ;
                
		exit( errno = 0 );
            }
        }
    }

/*
** Check for STDERR replacement
*/
    opterr = 0 ;
    optind = 1 ;
    while ((j = getoptt(argc, argv, OPTIONS)) != -1) {
        switch (j) {
            case 'f' :{
                errorfile = optarg ;
                break ;
            }
        }
    }
    if ( errorfile != NULL ) {
	int stdfd;
	
        if ( (stdfd = OPEN(errorfile,O_CREAT|O_WRONLY|O_SYNC|O_TRUNC,0777)) < 0 ) {
            printmessage(stderr,CASE_FATAL_ERROR,10,timestamp,"Error openning error file (%s) : %s\n",errorfile,strerror(errno)) ;
        }
        CLOSE(STDERR_FILENO);
        if ( fcntl(stdfd,F_DUPFD,STDERR_FILENO) != STDERR_FILENO ) {
            printmessage(stderr,CASE_FATAL_ERROR,11,timestamp,"Error dup on error file (%s) : %s\n",errorfile,strerror(errno)) ;
        }
    }
/*
** Check for STDOUT replacement
*/
    opterr = 0 ;
    optind = 1 ;
    while ((j = getoptt(argc, argv, OPTIONS)) != -1) {
        switch (j) {
            case 'o' :{
                outputfile = optarg ;
                break ;
            }
        }
    }
    if ( outputfile != NULL ) {
	int stdfd;
	
        if ( (stdfd = OPEN(outputfile,O_CREAT|O_WRONLY|O_SYNC|O_TRUNC,0777)) < 0 ) {
            printmessage(stderr,CASE_FATAL_ERROR,12,timestamp,"Error openning output file (%s) : %s\n",outputfile,strerror(errno)) ;
        }
        CLOSE(STDOUT_FILENO);
        if ( fcntl(stdfd,F_DUPFD,STDOUT_FILENO) != STDOUT_FILENO ) {
            printmessage(stderr,CASE_FATAL_ERROR,13,timestamp,"Error dup on output file (%s) : %s\n",outputfile,strerror(errno)) ;
        }
    }

/*
** Check for warning
*/
    opterr = 0 ;
    optind = 1 ;
    while ((j = getoptt(argc, argv, OPTIONS)) != -1) {
        switch (j) {
            case 'W' :{
                warning = SETTOONE ;
                break ;
	    }
        }
    }

/*
** Check for data listener
*/
    opterr = 0 ;
    optind = 1 ;
    while ((j = getoptt(argc, argv, OPTIONS)) != -1) {
        switch (j) {
            case 'D' : {  
	      if( optarg ) {
                if( 2 == sscanf( optarg, "%d:%d", &i, &k ) && i <= k ) {
                  pasvport_min = i; 
		  pasvport_max = k;
                } 
		else {
                  printmessage( stderr, CASE_FATAL_ERROR, 4, timestamp, "Invalid port range : %s\n", optarg );
                }
              } 
	      else {
#ifdef PORT_RANGE
                sscanf(PORT_RANGE,"%d:%d",&pasvport_min, &pasvport_max) ;
#endif
                if( 0 == pasvport_max ) {
                  pasvport_min = 0;
                  pasvport_max = 1;
                }
              }
              break;
	    }
        }
    }


/*
** Now all the others
*/
    opterr = 0 ;
    optind = 1 ;
    while ((j = getoptt(argc, argv, OPTIONS)) != -1) {
        switch (j) {

	    /* skip options already accounted for */
	    case 't':
	    case 'f':
	    case 'o':
	    case 'W':
	    case 'D':
	      break;

	    case 'h': 
#ifdef	WITH_SSL
	      transferoption |= TROPT_HIDDEN;
#else
	      printmessage( stderr, CASE_ERROR, 7, timestamp, "switch -h is unavailable: bbftpPRO was built without encryption\n" );
#endif	/*WITH_SSL*/
	      break;

	    case 'x':
	      transferoption |= TROPT_UNIX;
	      break;

	    case 'j':
	      /*transferoption |= TROPT_MPATH;*/
	      localips = optarg;
	      break;

	    case 'n':
	      if( pasvport_max ) 
		      natip = get_ip( optarg, NULL );
	      else if( warning ) 
		      printmessage( stderr, CASE_WARNING, 66, timestamp, "Switch [-n %s] safely ignored\n", optarg );
	      break;

            case 'b' :{
                background = SETTOONE ;
                break ;
            }
            case 'c' :{
                transferoption |= TROPT_GZIP;
                break ;
            }

	    case 'C': { /*incoming piped command*/
		if( optarg ) incoming = optarg;
		break;
	    }
	    case 'O': { /*outgoing piped command*/
		if( optarg ) outgoing = optarg;
		break;
	    }

            case 'd' :{
                debug = 1 ;
                break ;
            }
            case 'e' : {
                bbftpcmd = optarg ;
                break ;
            }
            case 'E' : {
                sshremotecmd = optarg ;
                usessh = 1 ;
                break ;
            }
		      
#ifdef CERTIFICATE_AUTH
            case 'g' :{
		usecert = SETTOONE;
		if( optarg )
                  service = optarg;
                break ;
            }
#endif         

            case 'i' :{
                inputfile = optarg ;
                if ( STAT(inputfile,&statbuf) < 0 ) {
                    printmessage(stderr,CASE_FATAL_ERROR,7,timestamp,"Input file (%s) cannot be stated\n",inputfile) ;
                }
                if ( (resultfile = (char *) malloc(strlen(inputfile) + 5 )) == NULL ) {
                    printmessage(stderr,CASE_FATAL_ERROR,1,timestamp,"Cannot malloc space for result file name\n") ;
                }
		sprintf( resultfile, "%s.res", inputfile );
                break ;
            }
            case 'I' :{
                sshidentityfile = optarg ;
                break ;
            }
            case 'L' : {
                sshcmd = optarg ;
                usessh = 1 ;
                break ;
            }
            case 'm' :{
                statoutput = SETTOONE ;
                break ;
            }
            case 'M' :{
                retcode = sscanf(optarg,"%d",&alluse) ;
                if( retcode != 1 || alluse < -1 || 0 == alluse ) {
                    printmessage(stderr,CASE_FATAL_ERROR,3,timestamp,"Maximum number of simultaneous transfers must be positive or -1\n");
                }
		maxfathers = alluse;
                break ;
            }
            case 'P' :{
                privatestr = optarg ;
                break ;
            }
            case 'q' :{
                transferoption |= TROPT_QBSS;
		break ;
	    }
            case 'p' :{
                retcode = sscanf(optarg,"%d",&alluse) ;
                if ( retcode != 1 || alluse <= 0) {
                    printmessage(stderr,CASE_FATAL_ERROR,3,timestamp,"Number of streams must be numeric and > 0\n") ;
                }
                nbport = alluse;
                break ;
            }
            case 'r' :{
                retcode = sscanf(optarg,"%d",&alluse) ;
                if ( retcode != 1 || alluse <= 0) {
                    printmessage(stderr,CASE_FATAL_ERROR,4,timestamp,"Number of tries must be numeric > 0\n") ;
                }
                globaltrymax = alluse ;
                break ;
            }
            case 'R' :{
                bbftprcfile = optarg ;
                break ;
            }
	
	    case 'S':
		sshbatchmode = SETTOONE;
		/* fall through */
            case 's':
                usessh = SETTOONE;
		if( optarg )
		  sshremotecmd = optarg;
                break ;

            case 'u' :{
                username = optarg ;
                break ;
            }
            case 'V':{
                verbose = SETTOONE ;
                break ;
            }
            case 'w' :{
                retcode = sscanf(optarg,"%d",&alluse) ;
                if ( retcode != 1 || alluse <= 0) {
                    printmessage(stderr,CASE_FATAL_ERROR,4,timestamp,"Control port must be numeric\n") ;
                }
                newcontrolport = alluse ;
                break ;
            }
            default : {
                Usage() ;
                printmessage(stderr,CASE_FATAL_ERROR,6,timestamp,"Error on command line (unsupported option -%c)\n",optopt) ;
            }
        }
    }
    
    /*
    ** Reset all outputs variable if statoutput is set to one
    */
    if ( statoutput ) {
        debug     = SETTOZERO ;
        verbose   = SETTOZERO ;
        warning   = SETTOZERO ;
        timestamp = SETTOZERO ;
    }
 
/*
** Check hostname
*/          
    if( optind != argc-1 ) {
      if( warning ) 
	printmessage( stderr, CASE_WARNING, 7, timestamp, "No hostname, localhost assumed\n" );
      hostname = "localhost";
    } 
    else
      hostname = argv[optind];

    hisctladdr.sin_addr.s_addr = get_ip( hostname, 

#ifdef	CERTIFICATE_AUTH
					 &fullhostname
#else
					 NULL
#endif	/*CERTIFICATE_AUTH*/
				       );

/*
 * Compute default client IP, override with switch -n, 
 * if listener for data connections and no switch -j
 *
*/
    if( 0 == natip && pasvport_max/*listener*/ && 0 == strlen( localips )/*no choice*/ ) {
	    /* best guess */
	    if( htonl( INADDR_LOOPBACK ) == hisctladdr.sin_addr.s_addr )
                        natip = htonl( INADDR_LOOPBACK );
	    else {
		    static char lhost[256];
		    
		    gethostname( lhost, sizeof( lhost ) );
		    natip = get_ip( lhost, NULL );
	    }
    }
    
/*
** Check all ssh stuff
*/
    if ( usessh) {
            char *tmpsshremotecmd = NULL;

        if ( sshremotecmd == NULL ) {
            /*
            ** Malloc space and set the default
            */
            
	    if( (sshremotecmd = (char *)malloc(strlen(SSHREMOTECMD) + 1)) == NULL )
		    goto cms001;

            strcpy(sshremotecmd,SSHREMOTECMD) ;
        } else {

	    /*
            ** Verify if -s is present if not add it (in order to fit v 1.9.4-tja1
            ** behaviour)
            */
            if ( strstr(sshremotecmd," -s") == NULL ) {
                if( (tmpsshremotecmd = (char *)malloc(strlen(sshremotecmd) + 4)) == NULL ) {
		    cms001:
                    printmessage(stderr,CASE_FATAL_ERROR,8,timestamp,"Cannot malloc space for ssh remote cmd\n") ;
                }
                sprintf(tmpsshremotecmd,"%s -s",sshremotecmd) ;
                sshremotecmd = tmpsshremotecmd ;
            }
        }

        if ( sshcmd == NULL ) {
            if ( (sshcmd = (char *) malloc(strlen(SSHCMD)+1) ) == NULL ) {
                  printmessage(stderr,CASE_FATAL_ERROR,9,timestamp,"Cannot malloc space for ssh cmd\n") ;
            }
            strcpy(sshcmd,SSHCMD) ;
        }
    }

/*
** Check for input file or command line
*/
    if ( inputfile == NULL && bbftpcmd == NULL ) {
        printmessage(stderr,CASE_FATAL_ERROR,20,timestamp,"Error on command line (-i or -e option are mandatory)\n") ;
    }
    
/*
** Check username if not in certificate authentication mode
*/          

    mypasswd = getpwuid( getuid() );
    if( NULL == mypasswd )
      printmessage(stderr,CASE_FATAL_ERROR,18,timestamp,"You don't exist, go away !\n");
    
    if( NULL == username )
      username = mypasswd->pw_name;
    
/*
** Check for the local user in order to find the .bbftprc file
*/
    if( NULL == bbftprcfile ) {
        /*
        ** look for the local user in order to find the .bbftprc file
        */
        if( NULL == mypasswd->pw_dir ) {
            if( warning ) 
	      printmessage( stderr, CASE_WARNING, 7, timestamp, "No home directory, .bbftprc will not be used\n" );
        } 
	else
          bbftprcfile = printxb( &logmessage, "%s/.bbftprc", mypasswd->pw_dir );
    }

    if( bbftprcfile ) {
      if( strncmp( bbftprcfile, "none", 4 ) != 0 ) {
	int rcfd;
	  
        rcfd  = OPEN( bbftprcfile, O_RDONLY, 0666 );
	
	if( rcfd < 0 ) {
          if( warning ) 
	    printmessage( stderr, CASE_WARNING, 11, timestamp, "Open(%s): %s\n", bbftprcfile, strerror( errno ) );
        }
	else {
	  interpreter( rcfd, -1, "" );

	  CLOSE( rcfd );
	      
	}
      }
    }


/*XXX*/
#if 0
    if ( debug ) {
        printmessage(stderr,CASE_NORMAL,0,timestamp,"Starting parameters -----------------\n") ;
        printmessage(stderr,CASE_NORMAL,0,timestamp,"number of tries   = %d\n",globaltrymax) ;
        printmessage(stderr,CASE_NORMAL,0,timestamp,"number of streams = %d\n",nbport) ;
        printmessage(stderr,CASE_NORMAL,0,timestamp,"localumask        = %03o\n",localumask) ;
        printmessage(stderr,CASE_NORMAL,0,timestamp,"remoteumask       = %03o\n",remoteumask) ;
        printmessage(stderr,CASE_NORMAL,0,timestamp,"remotecos         = %d\n",remotecos) ;
        printmessage(stderr,CASE_NORMAL,0,timestamp,"buffersize        = %d KB\n",buffersizeperstream) ;
        printmessage(stderr,CASE_NORMAL,0,timestamp,"sendwinsize       = %d KB\n",sendwinsize) ;
        printmessage(stderr,CASE_NORMAL,0,timestamp,"recvwinsize       = %d KB\n",recvwinsize) ;
        printmessage(stderr,CASE_NORMAL,0,timestamp,"ackto             = %d s\n",ackto) ;
        printmessage(stderr,CASE_NORMAL,0,timestamp,"recvcontrolto     = %d s\n",recvcontrolto) ;
        printmessage(stderr,CASE_NORMAL,0,timestamp,"sendcontrolto     = %d s\n",sendcontrolto) ;
        printmessage(stderr,CASE_NORMAL,0,timestamp,"datato     = %d s\n",datato) ;
		printmessage(stderr,CASE_NORMAL,0,timestamp,"option %s\n", (transferoption & TROPT_DIR) ? "createdir" : "nocreatedir") ;
		printmessage(stderr,CASE_NORMAL,0,timestamp,"option %s\n", (transferoption & TROPT_TMP) ? "tmpfile" : "notmpfile") ;
		printmessage(stderr,CASE_NORMAL,0,timestamp,"option %s\n", (transferoption & TROPT_RFIO) ? "remoterfio" : "noremoterfio") ;
		printmessage(stderr,CASE_NORMAL,0,timestamp,"option %s\n", (transferoption & TROPT_MODE) ? "keepmode" : "nokeepmode") ;
		printmessage(stderr,CASE_NORMAL,0,timestamp,"option %s\n", (transferoption & TROPT_ACC) ? "keepaccess" : "nokeepaccess") ;
#ifdef WITH_GZIP
		printmessage(stderr,CASE_NORMAL,0,timestamp,"option %s\n", (transferoption & TROPT_GZIP) ? "gzip" : "nogzip") ;
#endif        
		printmessage(stderr,CASE_NORMAL,0,timestamp,"option %s\n", (transferoption & TROPT_QBSS) ? "qbss" : "noqbss") ;
#ifdef	WITH_SSL
		printmessage(stderr,CASE_NORMAL,0,timestamp,"option %s\n", (transferoption & TROPT_HIDDEN) ? "encrypt" : "noencrypt");
#endif	/*WITH_SSL*/
		printmessage(stderr,CASE_NORMAL,0,timestamp,"option %s\n", (transferoption & TROPT_RECS) ? "recursive" : "norecursive");
		printmessage(stderr,CASE_NORMAL,0,timestamp,"option %s\n", (transferoption & TROPT_EXCL) ? "exclude" : "noexclude");
		printmessage(stderr,CASE_NORMAL,0,timestamp,"option %s\n", (transferoption & TROPT_UNIX) ? "unix" : "nounix");
		printmessage(stderr,CASE_NORMAL,0,timestamp,"option %s\n", (transferoption & TROPT_MPATH) ? "mpath" : "nompath");
        printmessage(stderr,CASE_NORMAL,0,timestamp,"-------------------------------------\n") ;
        printmessage(stderr,CASE_NORMAL,0,timestamp,"Connection mode ---------------------\n") ;
        if ( usecert ) {
/*			if ( username == NULL ) {*/
				printmessage(stderr,CASE_NORMAL,0,timestamp,"Using certificate authentication\n") ;
/*			} else {
				printmessage(stderr,CASE_NORMAL,0,timestamp,"Using standard bbftpPRO mode\n") ;
			}*/
        } else if ( useprivate ) {
            printmessage(stderr,CASE_NORMAL,0,timestamp,"Using private authentication\n") ;
        } else if ( usessh) {
            printmessage(stderr,CASE_NORMAL,0,timestamp,"Using ssh mode\n") ;
            printmessage(stderr,CASE_NORMAL,0,timestamp,"     ssh command        = %s \n",sshcmd) ;
            printmessage(stderr,CASE_NORMAL,0,timestamp,"     ssh remote command = %s \n",sshremotecmd) ;
        } else {
            printmessage(stderr,CASE_NORMAL,0,timestamp,"Using standard bbftpPRO mode\n",localumask) ;
        }
        printmessage(stderr,CASE_NORMAL,0,timestamp,"-------------------------------------\n") ;
   }
#endif


/*
** Open the input file if needed
*/
    if ( inputfile != NULL ) {
	infd = OPEN(inputfile,O_RDONLY,0666);
        if ( infd < 0 ) {
            printmessage(stderr,CASE_FATAL_ERROR,21,timestamp,"Error opening input file (%s) : %s\n",inputfile,strerror(errno)) ;
        }
	resfd = OPEN(resultfile,O_CREAT|O_WRONLY|O_SYNC|O_TRUNC,0777);
        if ( resfd < 0 ) {
            printmessage(stderr,CASE_FATAL_ERROR,22,timestamp,"Error opening result file (%s) : %s\n",resultfile,strerror(errno)) ;
        }
    }
     

/*
** Set the signals
*/
    if( bbftp_setfathersignals() < 0 )
      exit( errno );
/*
** Now we've got all informations to make the connection
*/

#ifdef  WITH_SSL
    stir();
#endif  /* WITH_SSL */
    
    if ( debug )
        printmessage(stderr,CASE_NORMAL,0,timestamp,"Connecting to server ----------------\n",localumask) ;
    reconnecttoserver() ;
    if ( debug )
         printmessage(stderr,CASE_NORMAL,0,timestamp,"Connecting end ---------------------\n",localumask) ;
   
    interpreter( infd, resfd, bbftpcmd );

    if( 0 == background ) { 
extern int waitforchildprocess( int *counter );
	    /* wait for all fathers to finish their transfers */
	    while( crtfathers > 0 ) errcode |= waitforchildprocess( &crtfathers );
    }

    writepackage( MSG_CLOSE_CONN, 0 );

    if ( inputfile != NULL ) {
        CLOSE( infd );
        CLOSE( resfd );
    }

    if( errcode || debug ) printmessage( stderr, errcode ? CASE_ERROR : CASE_NORMAL, errcode, timestamp, "bbftpPRO ends with status: %s\n", strerror( errcode ) );

    return( errno = errcode ); /* Tricky trick */
}
