/*
 * bbftpc/bbftp_cert.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 

#include "bbftp.h"
#include <sys/types.h>

#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <sys/time.h>

#include "client.h"
#include "structures.h"
#include "client_proto.h"
#include "common.h"

#include <gssapi.h>
#include <gfw.h>

extern  int     timestamp ;
extern  int     incontrolsock ;
extern  int     outcontrolsock ;
extern  int     newcontrolport ;
extern  int     verbose ;
extern  int     warning ;
extern  int     debug ;

extern	char	*service ;
extern	char	*fullhostname;

#ifdef	WITH_SSL
extern	int	crypton;
#endif	/* WITH_SSL */


int f_handler_cert( char **buffer, int *len, int *code ) {
  OM_uint32 maj_stat, min_stat;
  int ret;

start:
  if( *code != MSG_CRYPT )
    return( BB_RET_NONE );
 
  if( getcrypt( (struct mess_sec *)(*buffer) ) != BB_RET_OK )
    return( BB_RET_ERROR );

#ifdef	WITH_SSL
  if( crypton != 0 ) {
    if( genmyprsa() != BB_RET_OK )
      return( BB_RET_ERROR );

    if( sendrsa( MSG_NKEY ) != BB_RET_OK )
      return( BB_RET_ERROR );
  }
#endif	/* WITH_SSL */
  
  ret = writepackage( MSG_CERT_LOG, 0 );
  if( ret < 0 ) {

#ifdef	WITH_SSL
    if( crypton != 0 ) {
      crypton = 0;
      bbftp_close_control();
      
      incontrolsock = bbftp_connect( -1, htons( newcontrolport ), 0 );
      if( incontrolsock < 0 ) {
        printmessage(stderr,CASE_ERROR,52,timestamp,"bbftp_connect: %s\n", strerror( -incontrolsock ) );
        return( BB_RET_ERROR );
      }
      outcontrolsock = incontrolsock;

      if (debug) printmessage(stderr,CASE_NORMAL,0,timestamp,"Connection established\n") ;
      
      *buffer = readpackage( len, code, NULL );
      goto start;
    }
#endif	/* WITH_SSL */
    
    printmessage(stderr,CASE_ERROR,58,timestamp,"Error sending %s message : %s\n",MESSAGE_NAMES[MSG_CERT_LOG],strerror(errno)) ;
    return( BB_RET_ERROR );
  }
	  
  /* 
  ** The authentication process is run
  */
  maj_stat = gfw_init_sec_context(&min_stat, incontrolsock, outcontrolsock, (service ? service : fullhostname));
  if (maj_stat != GSS_S_COMPLETE && maj_stat != -1) {
      gfw_msgs_list *messages = NULL;
      gfw_status_to_strings(maj_stat, min_stat, &messages) ;
      if (verbose) {
          while (messages != NULL) {
              printmessage(stderr,CASE_ERROR,27,timestamp,"%s\n", messages->msg);
              messages = messages->next;
          }
      } else {
          printmessage(stderr,CASE_ERROR,27,timestamp,"%s\n", messages->msg);
      }
      return( BB_RET_ERROR );
  }
    
  /*
  ** Certificate authentication seems OK on client side so wait for the MSG_OK
  ** message
  */
  if (debug) printmessage(stderr,CASE_NORMAL,0,timestamp,"Client certificate authentication OK\n") ;
  if (debug) printmessage(stderr,CASE_NORMAL,0,timestamp,"Waiting for server answer\n") ;

  *buffer = readpackage( len, code, NULL );
  if( MSG_OK == *code || MSG_WARN == *code ) {
    if( verbose ) printmessage( stderr,
		                (MSG_OK == *code ? CASE_NORMAL : CASE_WARNING),
				(MSG_OK == *code ? 0 : 100),
				timestamp,
				"<< %s\n",
				*buffer
			      );
    return( BB_RET_OK );
  }
  return( BB_RET_NONE );
}


/*******************************************************************************
** bbftp_cert_connect :                                                     *
**                                                                             *
**      This routine opens a connection with the server and simply calls       *
**      the GSS Framework (GFW) library to validate user's certificates        *
**      routine returns 0 to set up correctly the global parameters            *
**                                                                             *
**      OUPUT variable :                                                       *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          incontrolsock       MAY BE MODIFIED                                *
**          outcontrolsock      MAY BE MODIFIED                                *
**          tmpctrlsock        MODIFIED                                       *
**                                                                             *
**                                                                             *
**      RETURN:                                                                *
**          -1  Unrecoverable error                                            *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/

int bbftp_cert_connect() 
{
    int     ret;
    int     errcode;

    incontrolsock = bbftp_connect( -1, htons( newcontrolport ), 0 );
    if( incontrolsock < 0 ) {
      printmessage(stderr,CASE_ERROR,52,timestamp, "bbftp_connect: %s\n", strerror( -incontrolsock ) );
      return( BB_RET_ERROR );
    }
   
    outcontrolsock = incontrolsock;
    
    if (debug) printmessage(stderr,CASE_NORMAL,0,timestamp,"Connection established\n") ;
   
    ret = handlepackage( handler_cert, &errcode );
    if( BB_RET_ERROR == ret || BB_RET_FT_NR == ret )
      bbftp_close_control();
    return( ret );
}
