/*
 * bbftpc/bbftp_get.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 

#include "bbftp.h"
#include <sys/types.h>

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <sys/wait.h>
#include <unistd.h>

#include <string.h>
#include <stdlib.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/stat.h>
#include <fcntl.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#include <math.h>

#include "client.h"
#include "structures.h"
#include "client_proto.h"
#include "common.h"


extern  int     debug ;
extern  int     warning ;
extern  int     verbose ;
extern  int     timestamp ;
extern  int     statoutput ;
extern  int     *mysockets;
extern  unsigned short int     *myports;

extern  int     transferoption  ; 
extern  int     filemode ;
extern  char    lastaccess[9];
extern  char    lastmodif[9];
extern  int     sendwinsize ;
extern  int     recvwinsize ;
extern  int     buffersizeperstream ;
extern  int     requestedstreamnumber ;
extern  my64_t  filesize ;
extern  char    *curfilename ;
extern  char    *realfilename;
extern	int	sendcontrolto ;
extern  int     nbport ;

extern  int     protocol ;

extern	pid_t	sshchildpid;


extern	xbuffer logmessage;
extern	int errcode;

extern	xbuffer	thelcd;

extern int crtfathers;
extern int maxfathers;


#ifndef HAVE_NTOHLL
my64_t ntohll(my64_t v) ;
#define	htonll	ntohll
#endif

int gtpt; /* GET or PUT */

static int end_file() {
  int retcode;

  do {
    retcode = bbftp_storeclosecastfile( realfilename );
    if( retcode < 0 ) {
      errcode = 80;
      break;
    }
    if( transferoption & TROPT_ACC ) {
      struct  utimbuf ftime;

      unsigned int am;

      sscanf( lastaccess, "%08x", &am );
      ftime.actime = am;
      sscanf( lastmodif, "%08x", &am );
      ftime.modtime = am;
      /*
      sscanf( lastaccess, "%08x", (unsigned int *)(&ftime.actime) );
      sscanf( lastmodif, "%08x", (unsigned int *)(&ftime.modtime) );
      */

      retcode = bbftp_storeutime( realfilename, &ftime );
      if( retcode < 0 )
	break;
    }
    if( transferoption & TROPT_MODE ) {
      retcode = bbftp_storechmod( realfilename, filemode );
      if( retcode < 0 )
        break;
    }
    if( transferoption & TROPT_TMP ) {
      retcode = bbftp_storerename( realfilename, curfilename );
    }
  }while( 0 );

  return( retcode );
}

static int brokenreport( int code, char *message, int retcode ) {
  errcode = code;
  printmessage( stderr, CASE_ERROR, code, timestamp, message );
  bbftp_free_f_var();
  bbftp_close_control();
  return( retcode );
}

static int sendamsg( int code, int brkcode, int retcode ) {
  if( writepackage( code, 0 ) < 0 ) {
    printxb( &logmessage, "Error sending %s message\n", MESSAGE_NAMES[code] );
    return( brokenreport( 64, ptrxb( &logmessage ), brkcode ) );
  }
  bbftp_free_f_var();
  return( retcode );
}


int dochildren() {
  int code;
  int *buffer = NULL;

  /*
  ** Now we are going to start the children
  */
  code = start_pumps();
  if( code < 0 ) {
    printmessage(stderr,CASE_ERROR,errcode,timestamp,"%s\n", ptrxb( &logmessage ) );
    return( BB_RET_ERROR );
  }

  ++crtfathers;
  
  return( BB_RET_OK );
}


static int getstatus( pid_t pid ) {
  int status;
  int ret;

  waitpid( pid, &status, 0 );

  if( WIFEXITED( status ) )
    ret = WEXITSTATUS( status );
  else
    ret = EINTR; /* killed by a signal? */

  return( ret );
}


/* Stats PUT file bytes_transfered seconds kbytes/s Mbits/s */
static struct timespec tstart;
void printstats( my64_t size ) {
  struct timespec tend;
  struct timeval tdiff;
  float s, bs;

  clock_gettime( CLOCK_MONOTONIC, &tend );
  tdiff.tv_sec = tend.tv_sec - tstart.tv_sec ;
  tdiff.tv_usec = (tend.tv_nsec - tstart.tv_nsec) / 1000;
  if( tdiff.tv_usec < 0 ) tdiff.tv_sec--, tdiff.tv_usec += 1000000;
  s = tdiff.tv_sec + (tdiff.tv_usec / 1000000.);
  #define nz(x)   ((x) == 0 ? 1 : (x))
  bs = size / nz(s);

  if( verbose ) {
	  printmessage( stderr, CASE_NORMAL, 0, timestamp, "\r%s %s%s%" LONG_LONG_FORMAT" bytes in %.3g secs (%.3g Kbytes/sec or %.3g Mbits/s)\n", gtpt ? "Got" : "Sent", transferoption & TROPT_UNIX ? "" : curfilename, transferoption & TROPT_UNIX ? "" : " ", size, s, bs / 1024.0,(8.0*bs) / (1024.0 * 1024.0) );
  }
  /*else */if( statoutput ) {
	  printmessage( stderr, CASE_NORMAL, 0, 0, "\r%s %s%s%" LONG_LONG_FORMAT" %.3f %d %d %d %d %s %s\n", gtpt ? "Got" : "Sent", transferoption & TROPT_UNIX ? "" : curfilename, transferoption & TROPT_UNIX ? "" : " ", size, s, buffersizeperstream, sendwinsize, recvwinsize, requestedstreamnumber, (transferoption & TROPT_GZIP) ? "compressed" : "uncompressed", (transferoption & TROPT_HIDDEN) ? "encrypted" : "unencrypted" );
  }
}


int waitforchildprocess( int *counter ) {
	int status;
	int ecode;
	pid_t pid;
    
	for(;;) {
	       	pid = wait( &status );	/* wait for any child process */
		if( -1 == (int)(pid) ) {
			int err;
			
			err = errno;
			if( EINTR == err ) continue;
			if( debug ) printmessage( stderr, CASE_ERROR, 66, timestamp,"Wait: %s\n", strerror( err ) );
			if( ECHILD == err ) return( 0 ); /* no child */
			continue;
		}
		if( pid == sshchildpid ) continue; /* thankfully, sshchildpid is NOT a child of the ringleader! */
		break;
	}

	--(*counter);

	if( WIFEXITED( status ) ) ecode = WEXITSTATUS( status ); /* normal termination */
	else ecode = EINTR; /* killed by a signal? */

	/*XXX*/ /*if( ecode && debug ) printmessage( stderr, CASE_NORMAL, 0, timestamp, "Process %d ends with error: %s\n", pid, strerror( ecode ) );*/

	return( ecode );
}


extern  char *buckets;
extern  spadetype *spades;
extern	volatile int *die;
extern	my64_t *speeds;
extern  int shmid;
extern	int shaping;
    

static int waitforallchildren( int cnt ) {
	int childerr = 0;
	
      	while( cnt > 0 ) {
		int tc;
		
	      	tc = waitforchildprocess( &cnt );
      		if( tc && 0 == childerr && (((transferoption & TROPT_UNIX) && 0 == (*die & 1)) || 0 == (transferoption & TROPT_UNIX)) ) {
			childerr = tc;
			if( cnt ) kill( 0, SIGTERM ); /* swiftly */
		}

		/*if( debug ) printmessage( stderr, CASE_NORMAL, 0, timestamp, "Children left: %d\n", cnt );*/
	}

      	return( childerr );
}


void endgame() {

	clock_gettime( CLOCK_MONOTONIC, &tstart );

	if( transferoption & TROPT_UNIX ) {
		int ptar;   /* pipe from/to tar */
		spadetype * const endspades = spades + requestedstreamnumber;
		int osend = 0;

		pid_t tarpid;

		int lmd;
		int mlmd;

		my64_t tsize = 0;	/* transferred size in Unix mode */

	    	if( 0 == (transferoption & TROPT_PIPE) ) {
			int p[2]; 
    
			if( -1 == pipe( p ) ) {
				printmessage( stderr, CASE_ERROR, 41, timestamp, "Pipe: %s\n", strerror( errno ) );
        
				mal:
				*die |= 32; /* pipe error */
				goto end2;
			}

			tarpid = fork();
			switch( (int)(tarpid) ) {
				char *argv[4];
				char *dir;
				char *file;
	       
				case 0:
					argv[0] = "sh";
					argv[1] = "-c";
					argv[3] = NULL;
					if( 0 == gtpt ) {
						close( p[0] );
						if( 1 != p[1] ) {
							dup2( p[1], 1 );
							close( p[1] );
						}
	 
						dir = strrchr( curfilename, '/' );
						if( dir ) {
							file = dir + 1;
							if( dir == curfilename )
								dir = "/";
							else {
								*dir = 0;
								dir = curfilename;
							}
						} else {
							dir = ( lenxb( &thelcd ) ? ptrxb( &thelcd ) : "." );
							file = curfilename;
						}
	    
#if defined( WITH_RFIO ) || defined( WITH_RFIO64 )
						{
							char *w = strpbrk( file, "[]{}*?+^$," );
#endif
	      
							printxb( &logmessage,

#if defined( WITH_RFIO ) || defined( WITH_RFIO64 )
									( NULL == w ? "rftar %s %s -c -f - --totals -C %s %s%s" : 
									  "rftar %s %s -c -f - --totals -C %s `rfdir %s | awk '{print $9}' | grep -E %s`" 
									  ), 
#else
									"tar %s %s -c -f - --totals -C %s `cd %s && ls -d %s`",
#endif

									( verbose ? "-v" : "" ), ( transferoption & TROPT_GZIP ? "-z" : "" ), dir,
									(
			
#if defined( WITH_RFIO ) || defined( WITH_RFIO64 )
									 NULL == w ? "" :
#endif
			 
									 dir 
									 ), 
									file
										);

#if defined( WITH_RFIO ) || defined( WITH_RFIO64 )
						}
#endif
	    
					} else {
						close( p[1] );
						if( 0 != p[0] ) {
							dup2( p[0], 0 );
							close( p[0] );
						}
	
						if( '/' == *curfilename )
							dir = curfilename;
						else {
							static xbuffer xdir = { NULL, 0, 0 };

							dir = printxb( &xdir, "%s/%s", (lenxb( &thelcd ) ? ptrxb( &thelcd ) : "."), curfilename );
						}
	  
						printxb( &logmessage, "%smkdir -p %s %s %star %s %s %s %s -x -f - --totals -C %s", 

#if defined( WITH_RFIO ) || defined( WITH_RFIO64 )
								"rf",
#else
								"",
#endif
                                                                                                                            
								dir,
                                                                                                                            
#if defined( WITH_RFIO ) || defined( WITH_RFIO64 )
								";", "rf",
#else
								"&&", "",
#endif
                                                                                                                            
								( transferoption & TROPT_ACC ? "--atime-preserve" : "" ),
								( transferoption & TROPT_MODE ? "-p" : "" ),
								( transferoption & TROPT_GZIP ? "-z" : "" ),
								( verbose ? "-v" : "" ),
								dir
									);
					}
					argv[2] = ptrxb( &logmessage );
	
					if( debug )
						printmessage( stderr/*XXX*/, CASE_NORMAL, 0, timestamp, "sh -c '%s'\n", ptrxb( &logmessage ) );  

					if( 0 == verbose && 0 == statoutput ) {
						int fd;

						fd = OPEN( "/dev/null", O_WRONLY );
						if( -1 == fd )
							printmessage( stderr, CASE_FATAL_ERROR, 46, timestamp, "Open: %s\n", strerror( errno ) );

						if( 2 != fd ) {
							dup2( fd, 2 );
							close( fd );
						}
					}

					execvp( argv[0], argv );
					printmessage( stderr, CASE_FATAL_ERROR, 46, timestamp, "Execvp: %s\n", strerror( errno ) );
					/* game over */

				case -1:
					printmessage( stderr, CASE_ERROR, 45, timestamp, "Fork: %s\n", strerror( errno ) );
					close( p[0] );
					close( p[1] );
					goto mal;

				default:
					if( 0 == gtpt ) {
						close( p[1] );
						ptar = p[0];
					}	   
					else {
						close( p[0] );
						ptar = p[1];
					}
					break;
			}
		}
		else
			ptar = gtpt;

		if( shaping ) {
			lmd = 6 * 4 / requestedstreamnumber;
			if( lmd < 2 ) 
				lmd = mlmd = 2;
			else 
				mlmd = lmd;
		}

    
		for(;;) {
			register char *lbucket;
			register spadetype *lspade;
			my64_t mint;
			my64_t st;
			my64_t stsq;

			for( lspade = spades, lbucket = buckets + 2 * sizeof( int ); lspade < endspades; ++lspade, lbucket += (buffersizeperstream * 1024 + 3 * sizeof( int )) ) {

				for( ; 0 == *die && FILLED == lspade->state; ) {	
				  	struct timeval timer;

					timer.tv_sec  = 0;
					timer.tv_usec = 1111;
					select( 0, 0, 0, 0, &timer );
				}

				if( FILLED == lspade->state ) {
					for( ; (*die & 1) && FILLED == lspade->state; ) {	
						struct timeval timer;

						timer.tv_sec  = 0;
						timer.tv_usec = 1111;
						select( 0, 0, 0, 0, &timer );
					}
					if( FILLED == lspade->state ) goto inside; /* Bail out now */
				}

				if( 0 == gtpt ) {
					if( shaping ) {
						int k = lspade - spades;
						register my64_t *lspeed = speeds + k;

						if( -1 == lmd ) {
							if( -1 == mint )
								lspade->length = buffersizeperstream * 1024;
							else {
								lspade->length = (int)(buffersizeperstream * 1024.0 * mint / (float)(*lspeed) * 1.3);
								if( lspade->length < 1024 ) lspade->length = 1024;
								else if( lspade->length > buffersizeperstream * 1024 )
									lspade->length = buffersizeperstream * 1024;
							}
						}
						else if( lmd >= 0 )
							lspade->length = buffersizeperstream * 1024;

						if( 0 == lmd )
							if( 0 == k )
								mint = *lspeed;
							else if( *lspeed < mint )
								mint = *lspeed;

						if( 0 == lmd || -12 == lmd )
							if( 0 == k ) {
								st = *lspeed;
								stsq = st * st;
							}
							else {
								st += (*lspeed);
								stsq += (*lspeed) * (*lspeed);
							}

						if( lmd > 0 || lmd <= -9 ) *lspeed = -mlmd;
						else if( -1 == lmd ) *lspeed = -2;
					}
					lspade->length = readpipe( ptar, lbucket, shaping ? lspade->length : buffersizeperstream * 1024 );
					if( lspade->length <= 0 ) {

						inside:
						lspade->length = 0;
						lspade->state = FILLED;
						goto end;
					}
					tsize += lspade->length;
				}
				else {
					int len = 0;
	  
					if( lspade->length > 0 )
						len = writepipe( ptar, lbucket, lspade->length );

					if( lspade->length <= 0 || len != lspade->length )
						goto end;
					
					tsize += len;
				}
				lspade->state = FILLED;
			}

			if( shaping ) {
				static my64_t psig6 = -1;
				static my64_t sig6;
				static my64_t pop;
				static my64_t msig6;
				static my64_t msd;

				if( 0 == lmd ) msig6 = st / requestedstreamnumber, msd = sqrt( (stsq - (st / requestedstreamnumber * st)) / (const int)(requestedstreamnumber - 1) );

				if( -8 == lmd ) mlmd = 3;

				if( -12 == lmd ) {
					static int clk = 0;

					sig6 = st / requestedstreamnumber;
					/*if( debug ) printmessage( stderr, CASE_NORMAL, 0, timestamp, "AVRG=%" LONG_LONG_FORMAT"\n", sig6 / 1000 );*/

					if( psig6 < 0 ) {
						if( -1 == psig6 ) {
							my64_t sd;

							sd = sqrt( (stsq - (st / requestedstreamnumber * st)) / (const int)(requestedstreamnumber - 1) );
							if( sig6 > msig6 || sd > msd ) {
								if( sig6 > msig6 && sd > msd ) {
									/*if( debug ) printmessage( stderr, CASE_NORMAL, 0, timestamp, "Reshaping to measured\n" );*/
									psig6 = -2;
									lmd = -1;
									mint = -1;
									continue;
								}
								/*if( debug ) printmessage( stderr, CASE_NORMAL, 0, timestamp, "Reshaping\n" );*/
								lmd = -13;
								mlmd = 1;
								continue;
							}
						}
						psig6 = sig6;
						lmd = -10;
						pop = 1;
						continue;
					}

					clk = (clk + 1) % (psig6 > 20000 ? 32 : 64);
					if( 3 * sig6 > 4 * psig6 )
						/*if( debug ) printmessage( stderr, CASE_NORMAL, 0, timestamp, "Reshaping down\n" )*/;
					else if( 0 == clk )
						/*if( debug ) printmessage( stderr, CASE_NORMAL, 0, timestamp, "Reshaping clocked\n" )*/;
					else {
						psig6 = (pop * psig6 + sig6) / (pop + 1);
						pop = (pop + 1) % 3; if( 0 == pop ) ++pop;

						lmd = -10;
						continue;
					}

					clk = 0;
					mlmd = 1;
				}

				if( -13 == lmd ) {
					lmd = 6 * 4 / requestedstreamnumber;
					if( lmd < 2 ) 
						lmd = mlmd = 2;
					else 
						mlmd = lmd;	
					psig6 = -1;
					continue;
				}

				--lmd; continue;
			}

		}
      
    
		end:
		close( ptar );

		if( 0 == (transferoption & TROPT_PIPE) ) {
			int tarend = getstatus( tarpid );

			osend = tarend;

			if( debug ) printmessage( stderr, tarend ? CASE_NORMAL : CASE_ERROR, tarend, timestamp, "Piped process ends, status: %s\n", strerror( tarend ) );
		}

    
		end2:
		errcode = waitforallchildren( requestedstreamnumber );
		errcode |= osend;

		printstats( tsize );

		if( errcode || debug ) printmessage( stderr, errcode ? CASE_ERROR : CASE_NORMAL, errcode, timestamp, errcode ? "Piped transfer lost\n" : "Piped transfer successful\n" );

		/*shmdt( buckets );*/
		shmctl( shmid, IPC_RMID, NULL );

		return;	
	}

	errcode = waitforallchildren( requestedstreamnumber );
	if( errcode ) bbftp_clean_child();
	
	if( gtpt && 0 == errcode ) {
		int retcode;

		retcode = end_file();
		if( retcode < 0 ) {
			printmessage( stderr, CASE_ERROR, errcode, timestamp, "%s\n", ptrxb( &logmessage ) );
			bbftp_storeunlink( realfilename );
			return;
		}
	}

	if( 0 == errcode ) printstats( filesize );

	/*XXX*/ /*if( gtpt ) sync();*/
}


int *myips = NULL;

int startgame() {
    int    code;
    char   *buffer;
    struct mess_store_v2 *messy_ptr;
    int    transferholder;
    int    retcode;

    /*
    ** At this stage we are waiting for the file length and
    ** characteristics
    */
  
    messy_ptr = (struct mess_store_v2 *)(readpackage( &retcode, &code, NULL ));
    switch( code ) {

      case MSG_NONE:
	return( brokenreport( 61, "Error waiting MSG_OK (on MSG_RETR/STORE_V2) message\n", BB_RET_CONN_BROKEN ) );

      case MSG_BAD:
      case MSG_BAD_NO_RETRY:
        /*
        ** The transfer has failed so ...
        */
        printmessage(stderr,CASE_ERROR,100,timestamp,"%s\n",(char *)(messy_ptr)) ;
        if (verbose) printmessage(stderr,CASE_NORMAL,0,timestamp,"<< %s\n",(char *)(messy_ptr)) ;
        errcode = 100;
	bbftp_free_f_var();
	return( MSG_BAD == code ? BB_RET_ERROR : BB_RET_FT_NR );

      case MSG_TRANS_OK_V2:
      case MSG_TRANS_OK_V3:
#ifndef WORDS_BIGENDIAN
	transferholder          = ntohl( messy_ptr->transferoption );
        filemode                = ntohl(messy_ptr->filemode) ;
        requestedstreamnumber   = ntohl(messy_ptr->nbstream) ;
        filesize                = ntohll(messy_ptr->filesize) ;
#else
	transferholder          = messy_ptr->transferoption;
        filemode                = messy_ptr->filemode ;
        requestedstreamnumber   = messy_ptr->nbstream ;
        filesize                = messy_ptr->filesize ;
#endif
	strncpy( lastaccess, messy_ptr->lastaccess, 8 );
	lastaccess[8] = '\0';
	strncpy( lastmodif, messy_ptr->lastmodif, 8 );
	lastmodif[8] = '\0';

        if( (transferoption & TROPT_GZIP) && 0 == (transferholder & TROPT_GZIP) ) {
          if( warning ) printmessage(stderr,CASE_WARNING,27,timestamp,"The server cannot handle compression: option ignored\n");
          transferoption &= ~TROPT_GZIP ;
        }
	if( (transferoption & TROPT_HIDDEN) && 0 == (transferholder & TROPT_HIDDEN) ) {
	  if( warning ) printmessage(stderr,CASE_WARNING,27,timestamp,"The server cannot handle encryption: option ignored\n");
	  transferoption &= ~TROPT_HIDDEN;
	}
	if( (transferoption & TROPT_EXCL) && 0 == (transferholder & TROPT_EXCL) ) {
	  if( warning ) printmessage(stderr,CASE_WARNING,27,timestamp,"The server cannot handle exclusion: option ignored\n");
	  transferoption &= ~TROPT_EXCL;
	}
	if( (transferoption & TROPT_UNIX) && 0 == (transferholder & TROPT_UNIX) ) {
	  if( warning ) printmessage(stderr,CASE_WARNING,27,timestamp,"The server cannot handle unix style: option ignored\n");
	  transferoption &= ~TROPT_UNIX;
	}
	if( (transferoption & TROPT_MPATH) && 0 == (transferholder & TROPT_MPATH) ) {
	  if( warning ) printmessage(stderr,CASE_WARNING,27,timestamp,"The server cannot handle multi-path transfers: option ignored\n");
	  transferoption &= ~TROPT_MPATH;
	}

#ifndef	WITH_GZIP
	if( 0 == (transferoption & TROPT_UNIX) )
	  transferoption &= ~TROPT_GZIP;
#endif	/* WITH_GZIP */
	
	if( 0 == (transferoption & TROPT_UNIX) ) {
	  if( 0 == filesize )
	    break;
	
	  if( gtpt && (transferoption & TROPT_EXCL) ) {
            if( filelength( curfilename ) == filesize ) {
              retcode = sendamsg( MSG_CREATE_ZERO, BB_RET_CONN_BROKEN, BB_RET_OK );
              return( retcode );
            }
			            
	    if( writepackage( MSG_GOAHEAD, 0 ) < 0 )
              return( brokenreport( 64, "Error sending MSG_GOAHEAD message\n", BB_RET_CONN_BROKEN ) );
          }
	}

        mysockets = (int *)malloc( requestedstreamnumber*sizeof(int) );
        myports = (unsigned short int *)malloc( requestedstreamnumber*sizeof(unsigned short int) );
	myips = (int *)(malloc( requestedstreamnumber*sizeof(int) ));
        if( NULL == mysockets || NULL == myports || NULL == myips ) {
          printmessage( stderr, CASE_ERROR, 35, timestamp, "Error allocating memory for socks&ports: %s\n", strerror( errno ) );
          return( sendamsg( MSG_ABORT, BB_RET_CONN_BROKEN, BB_RET_ERROR ) );
        }

        if( MSG_TRANS_OK_V2 == code ) {
		
	  if( getdatasock( requestedstreamnumber ) < 0 )
	    return( sendamsg( MSG_ABORT, BB_RET_CONN_BROKEN, BB_RET_ERROR ) );
	  
	  /*XXX*/
	  retcode = writepackage( MSG_PORTS, requestedstreamnumber * (sizeof( unsigned short int ) + sizeof( int )), myports, requestedstreamnumber * sizeof( unsigned short int ), myips, requestedstreamnumber * sizeof( int ) );
	  if( retcode < 0 ) {
	    printxb( &logmessage, "Error sending %s message(%s)\n", MESSAGE_NAMES[MSG_PORTS], strerror( -retcode ) );
	    return( brokenreport( 64, ptrxb( &logmessage ), BB_RET_CONN_BROKEN ) );
	  }
	
	  break;
        }

	/* MSG_TRANS_OK_V3 */
	/*XXX*/
	buffer = readpackage( &retcode, &code, NULL );
        if( MSG_PORTS != code ) {
	  printxb( &logmessage, "Error, NOT getting %s message\n", MESSAGE_NAMES[MSG_PORTS] );
	  return( brokenreport( 64, ptrxb( &logmessage ), BB_RET_CONN_BROKEN ) );
	}
        memcpy( myports, buffer, requestedstreamnumber * sizeof( unsigned short int ) );
	memcpy( myips, (unsigned short int *)(buffer) + requestedstreamnumber, requestedstreamnumber * sizeof( int ) );

	if( makedatasock( requestedstreamnumber/*,(unsigned short int *)(buffer)*/ ) < 0 ) {
          return( sendamsg( MSG_ABORT, BB_RET_CONN_BROKEN, BB_RET_ERROR ) );
        }

	break;

      case MSG_OK:
        /*
        ** This case can only happen if filesize = 0 or TROPT_EXCL during a put
        */
	if (verbose) printmessage( stderr, CASE_NORMAL, 0, timestamp, "<< %s%sOK\n",(retcode ? (char *)(messy_ptr) : ""), (retcode ? " " : "") );
        bbftp_free_f_var();
        return BB_RET_OK;

      default:
	/*
        ** Receive unkwown message so something is
        ** going wrong. close the control socket
        ** and restart
        */
	if( VALID_MESSAGE(code) )
	  printxb( &logmessage, "unexpected message %s\n", MESSAGE_NAMES[code] );
	else
	  printxb( &logmessage, "unlisted message %d\n", code );
	return( brokenreport( 62, ptrxb( &logmessage ), BB_RET_CONN_BROKEN ) );
    }
									    
  return( OK777 );
}

int sendattributes( char *remotefilename ) {
  static struct mess_store_v2 messy;
  int	len;

  /* Don't do more than maxfathers in the same time! */
  if( -1 != maxfathers ) while( crtfathers >= maxfathers ) waitforchildprocess( &crtfathers );

  if ( verbose) printmessage( stderr, CASE_NORMAL, 0, timestamp, ">> COMMAND : %s %s %s\n", (transferoption & TROPT_PIPE) ? "pipe" : (gtpt ? "get" : "put"), (transferoption & TROPT_PIPE) ? remotefilename : (gtpt ? remotefilename : curfilename), (transferoption & TROPT_PIPE) ? "" : (gtpt ? curfilename : remotefilename) );
  /*
  ** Check if file is correct
  */
  
  requestedstreamnumber = nbport;
  
  len = (gtpt ? bbftp_storecheckfile( realfilename ) : bbftp_retrcheckfile( curfilename ));
  if( len != 0 ) {
    printmessage(stderr,CASE_ERROR,errcode,timestamp,"%s\n", ptrxb( &logmessage ) );
    bbftp_free_f_var() ;
    return( len > 0 ? BB_RET_ERROR : BB_RET_FT_NR );
  }

#ifndef WITH_GZIP
  if( 0 == (transferoption & TROPT_UNIX) )
    if( transferoption & TROPT_GZIP ) {
      transferoption &= ~TROPT_GZIP;
      if( warning )
	printmessage( stderr, CASE_WARNING, 27, timestamp, "gzip option disabled\n" );
    }
#endif /* WITH_GZIP */
  
  /*
  ** Set up the parameters
  */
  strncpy( messy.lastaccess, gtpt ? "" : lastaccess, 8 );
  strncpy( messy.lastmodif, gtpt ? "" : lastmodif, 8 );
#ifndef WORDS_BIGENDIAN
  messy.transferoption       = htonl( transferoption );
  messy.filemode             = gtpt ? 0 : htonl( filemode );
  messy.sendwinsize          = htonl( sendwinsize );
  messy.recvwinsize          = htonl( recvwinsize );
  messy.buffersizeperstream  = htonl( buffersizeperstream );
  messy.nbstream             = htonl( requestedstreamnumber );
  messy.filesize             = gtpt ? 0 : htonll( filesize );
#else
  messy.transferoption       = transferoption;
  messy.filemode             = gtpt ? 0 : filemode;
  messy.sendwinsize          = sendwinsize;
  messy.recvwinsize          = recvwinsize;
  messy.buffersizeperstream  = buffersizeperstream ;
  messy.nbstream             = requestedstreamnumber;
  messy.filesize             = gtpt ? 0 : filesize;
#endif

  if( writepackage( (gtpt ? MSG_RETR_V2 : MSG_STORE_V2), sizeof( messy ), &messy, sizeof( messy ) ) < 0 )
    return( brokenreport( 64, "Error sending MSG_RETR/STORE_V2 message\n", BB_RET_CONN_BROKEN ) );

  len = strlen( remotefilename );
  if( writepackage( MSG_FILENAME, len, remotefilename, len ) < 0 )
    return( brokenreport( 64, "Error sending MSG_FILENAME message\n", BB_RET_CONN_BROKEN ) );
  
  return( OK777 );
}


static int do_get( char *remotefilename ) {
    int retcode;

    /* Sending file attributes */
    retcode = sendattributes( remotefilename );
    if( retcode != OK777 )
      return( retcode );
    
    /* Getting file features */ 
    retcode = startgame();
    if( retcode != OK777 )
      return( retcode );
   
    if( 0 == (transferoption & TROPT_UNIX) ) {  
      if( 0 == filesize ) {
        retcode = bbftp_storecreatefile( realfilename );
        if( retcode != 0 ) {
          printmessage(stderr,CASE_ERROR,errcode,timestamp,"%s\n", ptrxb( &logmessage ) );
          return( sendamsg( MSG_ABORT,
                            (retcode < 0 ? BB_RET_FT_NR_CONN_BROKEN : BB_RET_CONN_BROKEN),
                            (retcode < 0 ? BB_RET_FT_NR : BB_RET_ERROR)
                          ));
        }
	    
        retcode = end_file();
        if( retcode < 0 ) {
          printmessage( stderr, CASE_ERROR, errcode, timestamp, "%s\n", ptrxb( &logmessage ) );
	  bbftp_storeunlink(realfilename) ;
	  bbftp_free_f_var() ;
	  return( sendamsg( MSG_ABORT, BB_RET_CONN_BROKEN, BB_RET_ERROR ) );
        }

        retcode = sendamsg( MSG_CREATE_ZERO, BB_RET_CONN_BROKEN, BB_RET_OK );
        return( retcode );
      }
    }

    if ( debug ) printmessage(stderr,CASE_NORMAL,0,timestamp,"Accepted stream number: %d\n",requestedstreamnumber);

    /* Starting children */
    return( dochildren() );
}


int bbftp_get( char *remotefilename ) {
  gtpt = 1;
  return( do_get( remotefilename ) );
}
