/*
 * bbftpc/bbftp_lcd.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 

#include "bbftp.h"
#include <sys/types.h>

#include <errno.h>
#include <stdio.h>
#include <netinet/in.h>

#include <unistd.h>
#include <string.h>

#include "client.h"
#include "structures.h"
#include "client_proto.h"
#include "common.h"


extern  int        timestamp;
extern  int        verbose;

extern int errcode;

xbuffer thelcd = { NULL, 0, 0 };


int bbftp_lcd( char *dirpath ) {
    int err;
	
    if ( verbose) printmessage(stderr,CASE_NORMAL,0,timestamp,">> COMMAND : lcd %s\n",dirpath) ;

    /*
    ** We change the directory
    */
    if( CHDIR( dirpath ) < 0 ) {
	
        err = ERRNO;
	printmessage( stderr, CASE_ERROR, 71, timestamp, "Error changing directory(%s): %s\n", dirpath, strerror( err ) );
	errcode = 71;
	return( ENOMEM == err || EBUSY == err ? BB_RET_ERROR :  BB_RET_FT_NR );
    } 
  
    emptyxb( &thelcd ); 
    for(;;) {
      if( NULL == appendxb( &thelcd, NULL, 128 ) ) {
	errcode = 1;
        printmessage( stderr, CASE_ERROR, 1, timestamp, "Realloc: %s\n", strerror( errno ) );
        return( BB_RET_ERROR );
      }
			          
      if( NULL == GETCWD( ptrxb( &thelcd ), lenxb( &thelcd ) ) ) {
          
        err = ERRNO;
        if( ERANGE == errno )
          continue;
        printmessage( stderr, CASE_ERROR, 71, timestamp, "Getcwd : %s\n", strerror( err ) );
	errcode = 71;
	
	emptyxb( &thelcd );
        return( BB_RET_FT_NR );
      }
      break;
    }
    
    if( verbose ) printmessage( stderr, CASE_NORMAL, 0, timestamp, "<< OK : Current local directory is %s\n", ptrxb( &thelcd ) );
    
    return( BB_RET_OK );
}
