/*
 * bbftpc/bbftp_list.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 

#include "bbftp.h"
#include <sys/types.h>

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#include "client.h"
#include "structures.h"
#include "client_proto.h"
#include "common.h"

extern  int     verbose;
extern  int     timestamp;
extern  int     transferoption;

extern int errcode;


static	char	*localfilelist;
static	int	localfilelistlen;
int f_handler_mls( char **buffer, int *len, int *code ) {
  if( *code != MSG_LIST_REPL_V2 )
    return( BB_RET_NONE );

  localfilelistlen = *len;
  if( *len > 0 ) {
    localfilelist = (char *)(malloc( *len ));
    if( NULL == localfilelist ) {
      printmessage(stderr,CASE_ERROR,64,timestamp,"Malloc: %s\n", strerror( errno ) );
      return( BB_RET_ERROR );
    }
    memcpy( localfilelist, *buffer, *len );
  }
  else
    localfilelist = NULL;
  return( BB_RET_OK );
}

int bbftp_list( char *prefixdir, char *pattern, char **filelist, int *filelistlen ) {
    int lpre;
    int lpat;
    int holder;

    lpre = strlen( prefixdir );
    lpat = strlen( pattern );

#ifndef WORDS_BIGENDIAN
    holder = htonl( transferoption );
#else
    holder = transferoption;
#endif
    if( writepackage( MSG_LIST_V2, sizeof( holder ) + lpre + lpat, &holder, sizeof( holder ), prefixdir, lpre, pattern, lpat ) < 0 ) {
      printmessage(stderr,CASE_ERROR,64,timestamp,"Error sending %s message\n", MESSAGE_NAMES[MSG_LIST_V2]);
      errcode = 64;
      bbftp_close_control() ;
      return BB_RET_CONN_BROKEN ; /* restart connection */
    }


    holder = handlepackage( handler_mls );

    if( holder != BB_RET_OK )
      return( holder );

    *filelist = localfilelist;
    *filelistlen = localfilelistlen;

    return( BB_RET_OK );
}
