/*
 * bbftpc/bbftp_mklnk.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */

#include "bbftp.h"
#include <sys/types.h>

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "client.h"
#include "structures.h"
#include "client_proto.h"
#include "common.h"


extern  int     timestamp;
extern  int     verbose;
extern	int	transferoption;

extern xbuffer logmessage;
extern int errcode;

int f_handler_mklnk( char **buffer, int *len, int *code ) {
  F_HANDLER_DEFAULT
}

int bbftp_mklnk( char *oldpath, char *newpath ) {
    int	olen;
    int nlen;
    int holder;

    if ( verbose) printmessage(stderr,CASE_NORMAL,0,timestamp,">> COMMAND : mklnk %s %s\n", oldpath, newpath );

    olen = strlen( oldpath );
    nlen = strlen( newpath );

#ifndef	WORDS_BIGENDIAN
    holder = htonl( transferoption );
#else
    holder = transferoption;
#endif
    
    if( writepackage( MSG_MKLNK, sizeof( holder ) + olen + 1 + nlen, &holder, sizeof( holder ), oldpath, olen + 1, newpath, nlen ) < 0 ) {
      printmessage( stderr, CASE_ERROR, 64, timestamp, "Error sending %s message\n", MESSAGE_NAMES[MSG_MKLNK] );
      errcode = 64;
      bbftp_close_control() ;
      return BB_RET_CONN_BROKEN ; /* restart connection */
    }

    return( handlepackage( handler_mklnk ) ); 
}

int bbftp_storemklnk( char *oldpath, char *newpath ) {
  for(;;) {
    int ret;
    int err;    
  
    if( 0 == SYMLINK( oldpath, newpath ) ) 
	    
      return( 0 );

    err = ERRNO;
    switch( err ) {
      char *end;

      case ENOMEM:
      case EBUSY:
        ret = 1;
        break;
      case EEXIST:
	ret = 0;
	
	if( -1 == UNLINK( newpath ) ) {
		
	  err = ERRNO;
	  ret = (ENOMEM == err || EBUSY == err ? 1 : -1);
	}
	break;
      case ENOENT:
	if( transferoption & TROPT_DIR ) {
          end = strrchr( newpath, '/' );
          *end = 0;
          ret = bbftp_storemkdir( newpath );
          *end = '/';
          if( ret != 0 )
            return( ret );
          break;
	}
	/* fall through */							
      default:
        ret = -1;
        break;
    }
    
    if( ret != 0 ) {
      errcode = 77;
      printxb( &logmessage, "Symlink(%s): %s", newpath, strerror( err ) );
      return( ret );
    }
  }
}
