/*
 * bbftpc/bbftp_retr.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 

#include "bbftp.h"
#include <sys/types.h>

#if HAVE_DIRENT_H
# include <dirent.h>
#else
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# elif HAVE_SYS_DIR_H
#  include <sys/dir.h>
# elif HAVE_NDIR_H
#  include <ndir.h>
# endif
# define dirent direct
# define dirent64 direct64
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <string.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#include <errno.h>
#include <fcntl.h>
#include <fnmatch.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>

#include "common.h"
#include "client.h"
#include "structures.h"
#include "client_proto.h"

#ifdef WITH_GZIP
#include <zlib.h>
#endif


extern  int     debug ;
extern  int     warning ;
extern  int     verbose ;
extern  int     transferoption ; 
extern  int     filemode ;
extern  char    lastaccess[9];
extern  char    lastmodif[9];

extern	int     sendwinsize;
extern	int	recvwinsize;

extern	int     buffersizeperstream ;
extern	int     requestedstreamnumber ;
extern  my64_t  filesize ;
extern  char    *curfilename;
extern  char    *realfilename;
extern  unsigned short int     *myports ;
extern  int     *mysockets ;

extern	int	*myips;

extern  int     incontrolsock;
extern  int     outcontrolsock;

extern	int	ackto ;
extern  int     nbport ;
extern  int     protocol ;

extern int    timestamp;

extern int gtpt;

#ifdef  WITH_SSL
extern	int	crypton;
#endif  /* WITH_SSL */

extern xbuffer logmessage;
extern int errcode;

extern int maxfathers;


/*******************************************************************************
** bbftp_retrlisdir :                                                         *
**                                                                             *
**      Routine to list a directory                                            *
**                                                                             *
**      OUPUT variable :                                                       *
**          filelist   :  Where to store the file list (to be freed by caller  *
**          filelistlen:  Length of filelist buffer                            *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**                                                                             *
**      RETURN:                                                                *
**          -1  Unrecoverable error                                            *
**           0  OK                                                             *
**           1  recoverable error                                              *
**                                                                             *
*******************************************************************************/
int bbftp_retrlistdir( char *prefixdir, char *pattern, char **filelist, int *filelistlen ) {
    DIR     *curdir;
   
    sSTAT  statbuf;

    sDIRENT *dp;
    
    int err;
    int prefixlen;

    xbuffer crtfile = { NULL, 0, 0 };    
    xbuffer list = { NULL, 0, 0 }; 
    
    
    *filelist = NULL;   /* play safe for subsequent FREE */
    *filelistlen = 0;
    
    curdir = OPENDIR( prefixdir );
    
    if( NULL == curdir ) {
      err = ERRNO ;
      printxb( &logmessage, "Opendir(%s): %s ", prefixdir, strerror( err ) );
      errcode = 86;
      return( EACCES == err ? -1 : 1 );
    }
   
    prefixlen = strlen( prefixdir );
    if( NULL == appendxb( &crtfile, prefixdir, prefixlen ) ) {
      mal:
      printxb( &logmessage, "Malloc: %s", strerror( ENOMEM ) );
      errcode = 35;
					              
      CLOSEDIR( curdir );
						                    
      return( 1 );
    }
    
    do {
        sERRNO = 0;

	dp = READDIR( curdir );

	if( NULL == dp )
	  break;

        if( 0 == fnmatch( pattern, dp->d_name, 0 ) ) {
	    static char mode[3] = { '\0', '\0', '\0' };
	    int namelen;

	    namelen =  strlen( dp->d_name ) + 1;
	    if( NULL == appendxb( &list, dp->d_name, namelen ) ) {
	      freexb( &crtfile );
	      goto mal;
	    }

	    shrinkxb( &crtfile,  prefixlen );
	    if( NULL == appendxb( &crtfile, dp->d_name, namelen ) ) {
	      freexb( &list );
	      goto mal;
	    }
	    
	    if( LSTAT( ptrxb( &crtfile ), &statbuf ) < 0 ) {
		    
                printxb( &logmessage, "Lstat(%s): %s ", ptrxb( &crtfile ), strerror( ERRNO ) );
		errcode = 89;

		cmal:
		
		CLOSEDIR( curdir );

		freexb( &crtfile );
		freexb( &list );
                return( 1 );
            }
	    
            if( S_IFLNK == (statbuf.st_mode & S_IFLNK) ) {
	        mode[0] = 'l';

		if( STAT( ptrxb( &crtfile ), &statbuf ) < 0 ) {

                    /*
                    ** That means that the link refer to an unexisting file
                    */
                    mode[1] = 'u';
                } else {
                    if( S_IFDIR == (statbuf.st_mode & S_IFDIR) )
                      mode[1] = 'd' ;
                    else
                      mode[1] = 'f' ;
                }
            } else {
                mode[0] = ' ';
                if( S_IFDIR == (statbuf.st_mode & S_IFDIR) ) {
                    mode[1] = 'd';
                } else {
                    mode[1] = 'f';
                }
            }
	    
	    if( NULL == appendxb( &list, mode, 3 ) ) {
	      fmal:
	      freexb( &crtfile );
	      goto mal;
	    }

	    if( 'l' == mode[0] ) {
	      int llen;
	      static int buflen;
              static char *buf = NULL;	/* It's a leak of memory ! */

	      if( buf )
		goto inside;	      
	      
	      for( buflen = 128; ; buflen += 128 ) {
		     
		buf = (char *)(realloc( buf, buflen ));
                if( NULL == buf ) {
                  freexb( &crtfile );
                  freexb( &list );
                  goto mal;
                }
	
		inside:	
		
		llen = READLINK( ptrxb( &crtfile ), buf, buflen );

		if( -1 == llen ) {
		  printxb( &logmessage, "Readlink(%s): %s ", ptrxb( &crtfile ), strerror( ERRNO ) );
		  errcode = 88;
		  goto cmal;
		}
		
		if( llen < buflen ) {
		  *(buf + llen) = 0;
		  break;
		}
	      }

	      if( NULL == chainxb( &list, buf, llen + 1 ) )
		goto fmal;
	    }
	    
        }
    }while( 1 );
		
    err = ERRNO;

    CLOSEDIR( curdir );

    freexb( &crtfile );
    
    if( err != 0 ) {
        printxb( &logmessage, "Readdir(%s): %s ", prefixdir, strerror( err ) );
	errcode = 88 ;

	freexb( &list );
        return( 1 );
    }

    *filelistlen = lenxb( &list );
    *filelist = ptrxb( &list );
   
    return( 0 );   
}

my64_t filelength( char *filename ) {
	  
  sSTAT  statbuf;
	    
  if( STAT( filename, &statbuf ) < 0 )
    return( -1 );
	      
  if( S_IFDIR == (statbuf.st_mode & S_IFDIR) )
    return( -1 );
	        
  return( statbuf.st_size );
}


/*******************************************************************************
** bbftp_retrcheckfile :                                                       *
**                                                                             *
**      Routine to check a file and set the global parameters                  *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                      NOT MODIFIED                   * 
**          filemode                            MODIFIED                       *
**          lastaccess                          MODIFIED                       *
**          lastmodif                           MODIFIED                       *
**          filesize                            MODIFIED                       *
**          requestedstreamnumber               MODIFIED                       *
**                                                                             *
**      RETURN:                                                                *
**          -1  Unrecoverable error                                            *
**           0  OK                                                             *
**           1  recoverable error                                              *
**                                                                             *
*******************************************************************************/

int bbftp_retrcheckfile( char *filename ) {
	
    sSTAT  statbuf;	

    int     tmpnbport;
    int     err;
   
    if( transferoption & TROPT_UNIX )
      return( 0 );
   
    if( STAT( filename, &statbuf ) < 0 ) {

        /*
        ** It may be normal to get an error if the file
        ** does not exist but some error code must lead
        ** to the interruption of the transfer:
        **        EACCES        : Search permission denied
        **        ELOOP         : To many symbolic links on path
        **        ENAMETOOLONG  : Path argument too long
        **        ENOENT        : The file does not exists
        **        ENOTDIR       : A component in path is not a directory
        */
        err = ERRNO;
        printxb( &logmessage, "Stat %s : %s ", filename, strerror( err ) );
        errcode = 72;
	
	return( ENOMEM == err || EBUSY == err ? 1 : -1 );
    }
    
    /*
    ** The file exists so check if it is a directory
    */
    if( S_IFDIR == (statbuf.st_mode & S_IFDIR) ) {
        printxb( &logmessage, "File %s is a directory", filename );
        errcode = 73;
        return( -1 );
    }
    
    if (S_ISREG(statbuf.st_mode)) {
        filemode = statbuf.st_mode & ~S_IFREG;
    } else {
        filemode = statbuf.st_mode;
    }
    snprintf( lastaccess, 9, "%08x", (unsigned int)(statbuf.st_atime) );
    snprintf( lastmodif, 9, "%08x", (unsigned int)(statbuf.st_mtime) );
    filesize = statbuf.st_size ;
    
    tmpnbport = 1 + filesize / (4*1024*1024);	/* 4MB/stream */
    if( tmpnbport < nbport ) {
      requestedstreamnumber = tmpnbport;
      if( debug ) printmessage(stderr,CASE_NORMAL,0,timestamp,"Adjusted stream number: %d\n", requestedstreamnumber );
    }
    
    return( 0 );   
}


static void slam( int as ) {
  char *sm[] = { "I'm spacing out, try typing louder, I might snap out of it.\r\n",
                 "~*~ I was on ~*~ but now I'm gone I ~*~ left my name ~*~ To turn you on!!! ~*~\r\n",
                 "404 error: Cannot locate (your SN), please try again in 5 minutes, or 10 minutes, or 30 sorry for whatever inconvienence this may have caused.\r\n",
                 "Did you know that when someone isn't talking for a while and then their SN goes grey that means they arn't around the computer? Seems like you just found that out!! (=\r\n",
                 "Oops...It seems as though I've wandered away from my computer...either that or I really don't want to talk to you.\r\n",
                 "Am I just here ignoring you, or am I really idle? While you are figuring that out you may as well leave a message!\r\n",
                 "Right now I'm sitting here looking at the screen trying to see things from your point of view but I can't get my head that far up my ass.\r\n",
                 "Everyone is entitled to be stupid, but %n abuses the privilege.\r\n",
                 "A pretty girl can kiss a guy* a bird can kiss a butterfly* the rising sun can kiss the grass* but you my friend!! yes you!! YOU CAN KISS MY ASS*******\r\n",
                 "Don't piss me off today, I'm running out of places to hide to bodies\r\n",
                 "Roses are red violets are blue, God made me pretty, what the hell happen to you?\r\n",
                 "I have always woundered why people bang their heads against brick walls..... then I met you. Don't bother leaving a message.\r\n",
                 "If you didn't have feet you wouldn't wear shoes.....then why do you wear a bra\?\?!\r\n",
                 "Mirrors don't talk but lucky for you %n they don't laugh\r\n",
                 "I'm out like you in a beauty contest.\r\n",
               };
  static int idx = 0;

  write( as, sm[idx], strlen( sm[idx] ) );
  close( as );
  idx = (idx + 1) % (sizeof( sm ) / sizeof( char * ));
}

void childcloseall( int i ) {
  int j;

  for( j = i + 1; j < requestedstreamnumber; ++j )
    close( mysockets[j] );
}

static int myconnect( int i  ) {
  int    retcode;
	  
  retcode = bbftp_connect( mysockets[i], myports[i], myips[i] );
  if( retcode < 0 ) {
    printxb( &logmessage, "Couldn't connect to %s:%d : %s\n", inet_ntoa( ((struct sockaddr_in *)(myips + i))->sin_addr ), ntohs( myports[i] ), strerror( -retcode ) );
    return( -96 );
  }

  return( mysockets[i] );
}


static int expect( unsigned short int *sp_ptr, unsigned short int hisport, int current ) {
  int i;

  for( i = current; i < requestedstreamnumber; ++i )
    if( hisport == *(sp_ptr + i) )
      return( i );
  return( -1 );
}


int myaccept() {
  struct sockaddr_in peer_addr;
#if defined(SUNOS) || defined(_HPUX_SOURCE) || defined(IRIX) || defined(OSF1)
  int    addrlen;
#else
  size_t addrlen;
#endif
  int    ns;
  int i;

  unsigned short int *sports;

  /* Get the source ports */
  {
          int len, code;

          sports = (unsigned short int *)(readpackage( &len, &code, NULL));
          if( MSG_POSITION != code ) {
                  printxb( &logmessage, "Expecting %s, got %s! ", MESSAGE_NAMES[MSG_POSITION], MESSAGE_NAMES[code] );
                  goto maerr;
          }
	  if( debug ) printmessage(stderr,CASE_NORMAL,91,timestamp,"Got %s\n", MESSAGE_NAMES[MSG_POSITION]);
  }
 
  for( i = 0; i < requestedstreamnumber; ++i ) {
    int retcode;
    fd_set mask;
    struct timeval timer;
    int as;

    as = mysockets[i];

    if( as < 0 ) continue; /* skip out of order sockets */

    FD_ZERO( &mask );
    FD_SET( as, &mask );
    timer.tv_sec = 30;        /* THIRTY seconds */
    timer.tv_usec = 0;
    retcode = select( as + 1, &mask, 0, 0, &timer );
    if( retcode <= 0 ) {
      if( 0 == retcode )
	errno = ETIMEDOUT;
      else if( EINTR == errno ) {
	--i;
	continue;
      }
      printxb( &logmessage, "Select on data port(%d) : %s\n", i, strerror( errno ) );
    maerr:
      for( i = 0; i < requestedstreamnumber; ++i ) {
	  if( mysockets[i] < 0 ) mysockets[i] = -mysockets[i] - 1; /* restore out of order sockets, then close them all */
          close( mysockets[i] );
      }
      return( -96 );
    }
    ns = accept( as, 0, 0 );
    if( ns < 0 ) {
      printxb( &logmessage, "Accept on data port(%d) : %s\n", i, strerror( errno ) );
      goto maerr;
    }
    addrlen = sizeof( peer_addr );
    if( getpeername( ns, (struct sockaddr *)&peer_addr, (socklen_t *)&addrlen ) < 0 ) {
      close( ns );
      printxb( &logmessage, "Getpeername on data port(%d) : %s\n", i, strerror(errno) );
      goto maerr;
    }

    /*if( peer_addr.sin_addr.s_addr == hisctladdr.sin_addr.s_addr )*/ {
      int place;

      place = expect( sports, peer_addr.sin_port, i );
      if( place != -1 ) {

	      /*XXX*/
	      if( debug ) {
		      printmessage( stderr, CASE_NORMAL, 91, timestamp, "Connection(%s:%d --> ", inet_ntoa( peer_addr.sin_addr ), ntohs( peer_addr.sin_port ) );
		      addrlen = sizeof( peer_addr ); /* peer_addr is reused */
		      if( getsockname( ns, (struct sockaddr *)&peer_addr, (socklen_t *)&addrlen ) < 0 ) {
			      close( ns );
			      printxb( &logmessage, "Getsockname on accepted socket(%d) : %s\n", i, strerror(errno));
			      goto maerr;
		      }
		      printmessage( stderr, CASE_NORMAL, 91, timestamp, "%s:%d)%s\n", inet_ntoa( peer_addr.sin_addr ), ntohs( peer_addr.sin_port ), (place == i ? "" : " -- out of order, restored!") );
	      }

	if( place == i ) { /* close listening socket if no longer needed */
		int j;

		for( j = i + 1; j < requestedstreamnumber; ++j )
			if( as == mysockets[j] ) break;
		if( j == requestedstreamnumber ) {
			close( as );

			/*XXX*/if( debug ) printmessage( stderr, CASE_NORMAL, 91, timestamp, "Socket(%s, %d) closed\n", inet_ntoa( peer_addr.sin_addr ), ntohs( peer_addr.sin_port ) ); 
		}
	}

        mysockets[place] = ( i == place ? ns : (--i, -(ns + 1)) ); /* make it negative if out of order and stay on same i */

        /* make eventually reading sockets nonblock !! */
        if( -1 == fcntl( ns, F_SETFL, O_NONBLOCK ) ) {
	  printxb( &logmessage, "fcntl: %s\n", strerror(errno) );
	  goto maerr;
	}

        continue;
      }
    }

    if( debug ) printmessage(stderr,CASE_NORMAL,91,timestamp,"Door slammed to %s:%d\n",inet_ntoa(peer_addr.sin_addr),ntohs(peer_addr.sin_port));
    slam( ns );     /* does close too */
    --i;
    continue;
  }

  for( i = 0; i < requestedstreamnumber; ++i )
	  if( mysockets[i] < 0 ) mysockets[i] = -mysockets[i] - 1; /* restore out of order sockets */

  return( 0 );
}


void bbftp_clean_child() {
	bbftp_storeclosecastfile( NULL );
	if( gtpt && 0 == (transferoption & TROPT_UNIX) ) {
		if( debug  && realfilename ) printmessage(stderr,CASE_NORMAL,0,timestamp, "\rFile %s is deleted!\n", realfilename );
		bbftp_storeunlink( realfilename );
	}
}


char *buckets;
spadetype *spades;
volatile int *die;
my64_t *speeds;
int shmid;
int shaping;

/*******************************************************************************
** bbftp_retrtransferfile :                                                    *
**                                                                             *
**      Routine to transfer a file                                             *
**                                                                             *
**      INPUT variable :                                                       *
**          filename    :  file to send    NOT MODIFIED                        *
**                                                                             *
**      OUTPUT variable :                                                      *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**                                                                             *
**      RETURN:                                                                *
**          -1  transfer failed unrecoverable error                            *
**           0  Keep the connection open (does not mean that the file has been *
**              successfully transfered)                                       *
**          >0  Recoverable error but calling has the cleaning to do           *
**                                                                             *
*******************************************************************************/

static int crt;
static char *phpath;

static int bbftp_transferfile() {
    
    OFF_T	nbperchild;
    OFF_T	nbtosend;
    OFF_T	startpoint;
    OFF_T	nbread;
    OFF_T	numberread;
    
    int	retcode;
    int	fd;
    int	ns;

    char *bucket;
    spadetype *spade;
    my64_t *speed;

    int en;
      
      if( transferoption & TROPT_UNIX ) {
        bucket = buckets + crt * (buffersizeperstream * 1024 + 3 * sizeof( int )) + 2 * sizeof( int );        
	spade = spades + crt;
	if( shaping )
	  speed = speeds + crt;

	nbtosend = buffersizeperstream * 1024;
      }
      else {

	nbperchild = filesize / requestedstreamnumber;

        startpoint = crt * nbperchild;
        if( crt == requestedstreamnumber - 1 )
          nbtosend = filesize - startpoint;
        else
          nbtosend = nbperchild;
      }

	if( bbftp_setchildsignals() < 0 ) {
	  if( transferoption & TROPT_UNIX ) {
	    *die |= 2; /* signal handler error */
	    spade->state = EMPTY;
	  }
	  
	  exit( errno );
	}

	/*XXX*//*if( debug ) printmessage( stderr, CASE_NORMAL, 91, timestamp, "Child %d starting\n", getpid() );*/
	
	childcloseall( crt );
	ns = mysockets[crt];

	retcode = 1024 * (gtpt ? recvwinsize : sendwinsize);
        setsockopt( ns, SOL_SOCKET, (gtpt ? SO_RCVBUF : SO_SNDBUF), (char *)(&retcode), sizeof( retcode ) );
	
        close( STDIN_FILENO );

	bbftp_close_control();
	
	incontrolsock = outcontrolsock = ns;	/* for read&writemessage&package */

#ifdef	WITH_SSL
	if( 0 == (transferoption & TROPT_HIDDEN) ) crypton = 0;
#endif	/* WITH_SSL */

	if( 0 == (transferoption & TROPT_UNIX) ) {
	  
	  if( gtpt ) {
	    sSTAT statbuf;

	    if( STAT( phpath, &statbuf ) < 0 ) {
                                                                                
	      en = ERRNO;
              close( ns );
              exit( errno = en );
            }
	  }

	  fd = OPEN( phpath, (gtpt ? O_WRONLY : O_RDONLY), 0666 );

	  if( fd < 0 ) {
            en = ERRNO;

	    if( gtpt )

	      UNLINK( phpath );

	    close( ns );

            exit( errno = en );
          }

	  /* Is this is the last writing process? */
	  if( gtpt && (crt == requestedstreamnumber - 1) ) {
		  /* Seek to the end of file, in advance! */
		  if( LSEEK( fd, filesize, SEEK_SET ) < 0 )
			  goto seek001;
		  /* Allocate the entire file now!? */
		  if( WRITE( fd, "", 0 ) < 0 )
			  goto seek001;
	  }

	  if( LSEEK( fd, startpoint, SEEK_SET ) < 0 ) {
		  
	    seek001:  
            en = ERRNO;
		
	    printmessage( stderr, CASE_ERROR, 0, timestamp, "LSEEK/WRITE(%s): %s\n", phpath, strerror( en ) );

            CLOSE( fd );

	    if( gtpt )

	      UNLINK( phpath );
		
	    close( ns );
            exit( errno = en );
          }

	}

        nbread = 0;
        while ( nbread < nbtosend ) {
	  static char *readbuffer = NULL;

          if( transferoption & TROPT_UNIX ) {
            for( ; 0 == *die && EMPTY == spade->state; ) {
              struct timeval timer;
	      
	      timer.tv_sec  = 0;
              timer.tv_usec = 1111;
              select( 0, 0, 0, 0, &timer );
            }
            if( EMPTY == spade->state) {
              close( ns );
              exit( errno = 0 );
            }
	    if( 0 == spade->length )
	      break;
          }

	  if( 0 == gtpt ) {
	    char *rb;
	    
	    retcode = 1;
	    if( transferoption & TROPT_UNIX ) {	/* read from pipe */
	      numberread = spade->length;
	      rb = bucket;
	    }
	    else {	/* read from file */
              if( NULL == readbuffer ) {
                readbuffer = (char *)(malloc( buffersizeperstream * 1024 ));
                if( NULL == readbuffer ) {
                  en = errno;
                  close( ns );
                                                                                
                  CLOSE( fd );
                                                                                
                  exit( errno = en );
                }
              }
              numberread = READ( fd, readbuffer, buffersizeperstream*1024 <= nbtosend - nbread ? buffersizeperstream*1024 : nbtosend-nbread );
                                                                                
              if( numberread <= 0 ) {
                en = ERRNO;
                close( ns );
                                                                                
                CLOSE( fd );
                                                                                
                exit( errno = en );
              }
                                                                                
              nbread += numberread;
	      rb = readbuffer;
	    }

#ifdef	WITH_GZIP
	    if( (transferoption & TROPT_GZIP) && ( (0 == (transferoption & TROPT_UNIX)) || (transferoption & TROPT_PIPE) ) ) {
	      uLong bufcomplen;
	      static Bytef *compbuffer = NULL;

	      if( NULL == compbuffer ) {
		compbuffer = (Bytef *)(malloc( buffersizeperstream * 1024 ));

		if( NULL == compbuffer ) {
		  en = errno;
		  close( ns );
                                                                                
		  CLOSE( fd );
                                                                                
		  exit( errno = en );
		}
	      }
                                                                                
	      bufcomplen = buffersizeperstream * 1024;

	      retcode = compress( compbuffer, &bufcomplen, (Bytef *)(rb), (uLong)(numberread) );
	      if( 0 == retcode ) {
		numberread = (OFF_T)(bufcomplen);
		rb = (char *)(compbuffer);
		goto wtn001;
	      }
	      else
		retcode = 1;
	    }
#endif	/* WITH_GZIP */
	    
	    /* write to network */
	    wtn001:
	    if( shaping && 0 > (*speed) ) stopwatch( START );
	    retcode = writepackage( (0 == retcode ? DATA_COMPRESS : DATA_NOCOMPRESS), (int)(numberread), rb, (int)(numberread) );
	    if( shaping && 0 > (*speed) ) {
	      static int avrg = 0;
              static my64_t as = 0;
              my64_t s;
              int mod;

              s = stopwatch( STOP );
	      mod = -(int)(*speed);
              as = (avrg * as + s) / (avrg + 1);
              if( avrg == mod - 1 ) {
                *speed = as;
                if( 0 == as ) ++(*speed);
              }
              avrg = (avrg + 1) % mod;
	    }
            if( retcode < 0 ) {
              if( transferoption & TROPT_UNIX ) {
                *die |= 4; /* write to network error */
		spade->state = EMPTY;
	      }
              else
                                                                                
                CLOSE( fd );
                                                                                
              close( ns );
              exit( errno = -retcode );
            }

	    if( transferoption & TROPT_UNIX ) {
              spade->state = EMPTY;
              continue;
            }
	  }
	  else {	/*XXX*/
	    char *buffer;
	    int lentowrite;
	    int tmplen;

	    /* read from network */

#ifdef	WITH_GZIP
	    if( transferoption & TROPT_GZIP ) {
	      if( (0 == (transferoption & TROPT_UNIX)) || (transferoption & TROPT_PIPE) )
		goto rfn001;
	    }
#endif	/* WITH_GZIP */

	    if( transferoption & TROPT_UNIX )
	      buffer = readpackage( &lentowrite, &retcode, bucket - 2 * sizeof( int ) );
	    else {
	      rfn001:
	      buffer = readpackage( &lentowrite, &retcode, NULL );
	    }
            if( MSG_NONE == retcode )
              retcode = -( buffer ? ECONNRESET : EMSGSIZE );

            if( retcode < 0 ) {
              if( transferoption & TROPT_UNIX ) {
		spade->state = EMPTY;
		spade->length = lentowrite;
		*die |= 8; /* read from network error */
	      }
              else {
                                                                                
                CLOSE( fd );
                                                                                
                UNLINK( phpath );
                                                                                
              }
              close( ns );
              exit( errno = -retcode );
            }

#ifdef	WITH_GZIP
	    if( (transferoption & TROPT_GZIP) && ((0 == (transferoption & TROPT_UNIX)) || (transferoption & TROPT_PIPE)) )
	      if( DATA_COMPRESS == retcode ) {
	        int rc;
	        uLong bufcomplen;
                static Bytef *compbuffer = NULL;
	     
	        if( transferoption & TROPT_PIPE )
		  compbuffer = bucket;
	        else if( NULL == compbuffer ) {
		  compbuffer = (Bytef *)(malloc( buffersizeperstream * 1024 ));
                  if( NULL == compbuffer ) {
                    en = errno;
                                                                                
                    CLOSE( fd );
                                                                                
                    UNLINK( phpath );
                                                                                
                    close( ns );
                    exit( errno = en );
		  }
                }

                bufcomplen = buffersizeperstream * 1024;
                                                                                
                rc = uncompress( compbuffer, &bufcomplen, (Bytef *)(buffer), (uLong)(lentowrite) );
                if( rc != 0 ) {
                                                                                
                  CLOSE( fd );
                                                                                
                  UNLINK( phpath );
                                                                                
                  close( ns );
                  exit( errno = EILSEQ );
                }
                                                                                
                lentowrite = bufcomplen;
                buffer = (char *)(compbuffer);
	      }
	      else if( transferoption & TROPT_PIPE )
	        memcpy( bucket, buffer, lentowrite );
#endif	/* WITH_GZIP */

	    /* write to pipe */
	    if( transferoption & TROPT_UNIX ) {
              spade->length = ( MSG_ACK == retcode ? 0 : lentowrite );
              spade->state = EMPTY;
	      if( MSG_ACK == retcode ) {
		      /*XXX*/if( debug ) printmessage( stderr, CASE_NORMAL, 0, timestamp, "Transfer is over, got MSG_ACK!\n" );
		      break; 
	      }
              continue;
            }

	    /* write to file */
            tmplen = 0;
	    while ( tmplen < lentowrite ) {
                                                                                
              retcode = WRITE( fd, buffer + tmplen, lentowrite - tmplen );
                                                                                
              if( retcode < 0 ) {
                en = ERRNO;
                                                                                
		printmessage( stderr, CASE_ERROR, 0, timestamp, "WRITE(%s): %s\n", phpath, strerror( en ) );

                CLOSE( fd );
                                                                                
                UNLINK( phpath );
                                                                                
                close( ns );

                exit( errno = en );
              }
                                                                                
              tmplen += retcode;
            }
                                                                                
            nbread += lentowrite;
	  }
	}

	/* close the file */
	if( 0 == (transferoption & TROPT_UNIX ) ) {

	  retcode = CLOSE( fd );
	  
	  if( retcode < 0 && gtpt ) {
	    en = ERRNO;
	    
	    UNLINK( phpath );

	    close( ns );

	    exit( errno = en );
	  }
	}

	/*XXX*/
	if( gtpt || (transferoption & TROPT_UNIX) ) {

		if( (transferoption & TROPT_UNIX) && gtpt ) *die |= 1;	/* early! */
	  
		retcode = writepackage( MSG_ACK, 0 );
      		if( retcode < 0 ) {
      			/*XXX*/if( debug ) printmessage( stderr, CASE_NORMAL, 0, timestamp, "Failed to send MSG_ACK!\n" );
      			if( transferoption & TROPT_UNIX ) {
      				*die |= 4;	/* write to network error */
      				if( gtpt ) *die &= ~1;	/* transfer may be over though */

			    bad001:
      				spade->state = EMPTY;
      			} else if( gtpt ) UNLINK( phpath );

		    bad002:
      			close( ns );
      			exit( errno = -retcode );
      		}
	      	/*XXX*/if( debug ) printmessage( stderr, CASE_NORMAL, 0, timestamp, "Sent MSG_ACK!\n" );
	}

	if( 0 == gtpt ) {
	    int len, code;

	    readpackage( &len, &code, NULL );
	    if( debug ) printmessage( stderr, MSG_ACK == code ? CASE_NORMAL : CASE_ERROR, 0, timestamp, 
			    MSG_ACK == code ? "Got %s!!\n" : "Did not get %s!!\n", MESSAGE_NAMES[MSG_ACK] );
	    if( MSG_ACK != code ) {
		    retcode = -ENETRESET;
		    if( transferoption & TROPT_UNIX ) {
			    *die |= 8; /* read from network error */
			    goto bad001;
		    }
		    goto bad002;
	    }
	}

	if( transferoption & TROPT_UNIX ) {

		if( 0 == gtpt ) *die |= 1;	/* late! */
		
		spade->state = EMPTY;
	}
	else {
	  /*XXX*/ /*if( debug ) printmessage( stderr, CASE_NORMAL, 0, timestamp, "Child has %s %" LONG_LONG_FORMAT " bytes, ends correctly\n", (gtpt ? "received" : "sent"), (my64_t)(nbread) );*/
	}

	/* close network */
	close( ns );

        exit( errno = 0 );
}


int start_pumps() {
  int ns;
  int i;
  int retcode;

    switch( protocol ) {
      case 12:
        /* Active mode */
        ns = myaccept();
        break;
      case 13:
        /* Passive mode */
        for( i = 0;  i < requestedstreamnumber; ++i ) {
          ns = myconnect( i );
          if( ns < 0 ) {
            childcloseall( -1 );
            break;
          }
        }
        break;
      default:
         break; /* Do nothing */
    }

    if( ns < 0 ) {
      errcode = -ns;
      bbftp_free_f_var();
      return( -1 );
    }
                                                                                
    retcode = fork();
    switch( retcode ) {
      case 0:
        break;
      case -1:
        printxb( &logmessage, "Fork: %s", strerror( errno ) );
        errcode = 36;
        /* fall through */
      default:
        childcloseall( -1 );
        bbftp_free_f_var();
        return( -1 == retcode ? -1 : 0 );
    }
                                                                                
    if( bbftp_setsonsignals() < 0 )
      exit( errno = EINVAL );

    if( transferoption & TROPT_UNIX ) {

      shaping = (0 == gtpt) && (transferoption & TROPT_MPATH) && (requestedstreamnumber > 1);

      shmid = shmget( IPC_PRIVATE, requestedstreamnumber * ( buffersizeperstream * 1024 + 3 * sizeof( int ) + sizeof( spadetype ) + (shaping ? sizeof( my64_t ) : 0) ) + sizeof( int ), IPC_CREAT | IPC_EXCL | 0770 );
      if( -1 == shmid ) {
        printmessage( stderr, CASE_ERROR, 35, timestamp, "Shmget: %s\n", strerror( errno ) );
        exit( errno );
      }
      buckets = shmat( shmid, NULL, 0 );
      if( (char *)(-1) == buckets ) {
	i = errno;
        printmessage( stderr, CASE_ERROR, 35, timestamp, "Shmat: %s\n", strerror( errno ) );
        shmctl( shmid, IPC_RMID, NULL );
        exit( errno = i );
      }
      spades = (spadetype *)(buckets + requestedstreamnumber * (buffersizeperstream * 1024 + 3 * sizeof( int )));
      memset( spades, gtpt ? FILLED : EMPTY, requestedstreamnumber * sizeof( spadetype ) );
      die = (int *)(spades + requestedstreamnumber);
      *die = 0;
      if( shaping )
	speeds = (my64_t *)(die + 1);
    }
    else {
      if( gtpt ) {
	retcode = bbftp_storecreatefile( realfilename );
	if( retcode != 0 ) {
	  printmessage( stderr, CASE_ERROR, errcode, timestamp, "%s\n", ptrxb( &logmessage ) );
	  exit( errno = EPERM );
 	}
                                                                                
#if defined( WITH_RFIO ) || defined( WITH_RFIO64 )
	if( 1 == is_rfio( realfilename ) ) {
	  if( 0 == rfio_HsmIf_FindPhysicalPath( realfilename, &phpath ) )
	    exit( errno = EPERM );
	}
	else
#endif
          phpath = realfilename;
      }
      else
	phpath = curfilename; /*XXX*/
    }

    if( 1 != maxfathers && 0 == (transferoption & TROPT_PIPE) ) setsid(); /* get a new process group */
    
    /*
    ** Now start all our children
    */
    for( i = 0; i < requestedstreamnumber; ++i ) {
      retcode = fork();
      switch( retcode ) {
	case 0:
	  crt = i;	
	  bbftp_transferfile();	/* new process, no return */
	  break;
	case -1:
	  retcode = errno;
	  if( transferoption & TROPT_UNIX ) {
	    *die |= 16; /* fork error */
	    shmctl( shmid, IPC_RMID, NULL );
	  }
	  if( i ) {
		  kill( 0, SIGTERM );
extern int waitforchildprocess( int *counter );
		  while( i > 0 ) waitforchildprocess( &i );
	  }
	  bbftp_clean_child();
	  exit( errno = retcode );
	default:
	  close( mysockets[i] );
	  continue;
      }
    }

    bbftp_free_n_var();
    endgame();

    if( debug ) printmessage( stderr, errcode ? CASE_ERROR : CASE_NORMAL, errcode, timestamp, "Leader ends with status: %s\n", strerror( errcode ) );
    exit( errno = errcode );
}
