/*
 * bbftpc/bbftp_utils.c
 * Copyright (C) 1999, 2000, 2001, 2002 IN2P3, CNRS
 * bbftp@in2p3.fr
 * http://doc.in2p3.fr/bbftp
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */ 

/****************************************************************************

  
  
 bbftp_utils.c v 2.0.0 2001/03/01   - Creation of the routines
               v 2.0.1 2001/04/19   - Correct indentation
               v 2.1.0 2001/05/29   - Add private authentication
                                    - Add -m option
			   v 2.2.0  2001/10/03  - Add certificate authentication

*****************************************************************************/
#include "bbftp.h"

#include <string.h>

/* To be sure of optional arguments; should go in the Makefile */
#include "getopt.c"

#ifndef OSF1
#include <unistd.h>
#endif /* OSF1 */

#include <stdio.h>
#include <time.h>
#include <signal.h>
#include <stdarg.h>
#include <netinet/in.h>

#include "common.h"
#include "client.h"

#ifdef HAVE_BYTESWAP_H
#include <byteswap.h>
#endif

extern  int     incontrolsock ;
extern  int     outcontrolsock ;
extern  char    *curfilename ;
extern  char    *realfilename ;
extern  unsigned short int     *myports ;
extern  int     *mysockets ;

extern	int	*myips;

extern  int     debug ;
extern  int     requestedstreamnumber ;

extern	int	*nips;
extern	int	sip;


my64_t convertlong(my64_t v) {
    struct bb {
        int    fb ;
        int sb ;
    } ;
    struct bb *bbpt ;
    int     tmp ;
    my64_t    tmp64 ;
    
    tmp64 = v ;
    bbpt = (struct bb *) &tmp64 ;
    tmp = bbpt->fb ;
    bbpt->fb = ntohl(bbpt->sb) ;
    bbpt->sb = ntohl(tmp) ;    
    return tmp64 ;
}

#ifndef HAVE_NTOHLL
my64_t ntohll(my64_t v) {
#ifdef HAVE_BYTESWAP_H
    return bswap_64(v);
#else
    long lo = v & 0xffffffff;
    long hi = v >> 32U;
    lo = ntohl(lo);
    hi = ntohl(hi);
    return ((my64_t) lo) << 32U | hi;
#endif
}
#define htonll ntohll
#endif

void printmessage(FILE *strm , int flag, int errcode, int tok, char *fmt, ...) 
{
    va_list ap;
    time_t  logtime ;
    char    clogtime[50] ;
    
    va_start(ap,fmt);
    
    /*
    ** If timestamp start to print the time
    */
    if (tok) {
        /*
        ** Get time
        */
        if ( time(&logtime) == -1 ) {
            strcpy(clogtime,"") ;
        } else {
            if ( strftime(clogtime,sizeof(clogtime),"%a %b %d %H:%M:%S (%Z)",localtime(&logtime)) == 0 )    {
                strcpy(clogtime,"") ;
            }
        }
        /*
        ** And print it
        */
        if ( strlen(clogtime) > 0 ) {
            (void) fprintf(strm,"%s : ",clogtime) ;
        }
    }
    /*
    ** Check if it is an error
    */
    if ( flag == CASE_ERROR || flag == CASE_FATAL_ERROR ) {
        /*
        ** It is an error
        */
        (void) fprintf(strm,"BBFTP-ERROR-%05d : ",errcode) ;
    } else if ( flag == CASE_WARNING ) {
        /*
        ** It is a warning
        */
        (void) fprintf(strm,"BBFTP-WARNING-%05d : ",errcode) ;
    }
    /*
    ** And print the requested string 
    */
    (void) vfprintf(strm, fmt, ap);
    fflush(strm) ;
    if ( flag == CASE_FATAL_ERROR ) {
        /*
        ** exit in case of error
        */
        exit( errcode );
    }
    va_end(ap);
}

void Usage() 
{
    printf("Usage bbftp -i inputfile | -e \"command line\" \n") ;
#ifdef CERTIFICATE_AUTH
    printf("           [-g service name]\n") ;
#endif
    printf("           [-u username]\n") ;
    printf("           [-b (background)]\n") ;
    printf("           [-c (gzip compress)]\n") ;
    printf("           [-f errorfile]\n") ;
#ifndef PRIVATE_AUTH
    printf("           [-E server command for ssh]\n") ;
    printf("           [-I ssh identity file]\n") ;
    printf("           [-L ssh command]\n") ;
    printf("           [-s (use ssh)]\n") ;
    printf("           [-S (use ssh in batch mode)]\n") ;
#else
    printf("           [-P Private authentication string]\n") ;
#endif
    printf("           [-m (special output for statistics)]\n") ;
    printf("           [-o outputfile]\n") ;
    printf("           [-p number of // streams]\n") ;
    printf("           [-q (for QBSS on)]\n") ;
    printf("           [-r number of tries]\n") ;
    printf("           [-R .bbftprc filename]\n") ;
    printf("           [-t (timestamp)]\n") ;
    printf("           [-V (verbose)]\n") ;
    printf("           [-w controlport]\n") ;
    printf("           [-W (print warning to stderr)]\n") ;
    printf("           [-C Incoming command]\n" );
    printf("           [-O Outgoing command]\n" );

    printf("           [-D[min:max] (Domain of Ephemeral Ports)]\n" );
#ifdef	WITH_SSL
    printf("           [-h (hidden, data(!) encrypted)]\n" );
#endif	/*WITH_SSL*/
    printf("           [-x (unix style)]\n" );
    printf("           [-M maximum number of simultaneous file transfers]\n" );
    printf("           [-n NAT-name-or-IP (server alike)]\n" );
    printf("           [-j local-IP-list (CSV, server alike, multi-path)]\n" );
    
    printf("           host\n") ;
    printf("      or \n") ;
    printf("      bbftp -v\n") ;
}

void bbftp_close_control() 
{
    close( incontrolsock );
    if( incontrolsock != outcontrolsock) close(outcontrolsock);
    /* Don't kill the ssh client, it dies when the sshd counterpart dies */
}

void bbftp_free_n_var() {
  if( 0 == sip ) FREE( nips )
  FREE( myips )
  FREE( myports )
  FREE( mysockets )
}

void bbftp_free_f_var() {
  bbftp_free_n_var();
  FREE( realfilename )
  FREE( curfilename )
}

char *appendxb( xbuffer *xb, char *str, int len ) {
  if( len > xb->lsize ) {
    char *ptr;
    int ext;
    int rem;

    ext = len - xb->lsize;
    rem = ext % GROWTH;
    if( rem )
      ext += GROWTH - rem;

    ptr = (char *)(malloc( xb->tsize + ext ));
    if( NULL == ptr ) {
      FREE( xb->line )
      xb->tsize = 0;
      xb->lsize = 0;
      return( NULL );
    }

    if( xb->tsize - xb->lsize )
      memcpy( ptr, xb->line, xb->tsize - xb->lsize );
    xb->tsize += ext;
    xb->lsize += ext;
    FREE( xb->line )
    xb->line = ptr;
  }
  if( len > 0 ) {
    if( str )
      memcpy( xb->line + xb->tsize - xb->lsize, str, len );
    xb->lsize -= len;
  }
  
  return( xb->line );
}

void emptyxb( xbuffer *xb ) {
  xb->lsize = xb->tsize;
}

void shrinkxb( xbuffer *xb, int kept ) {
  if( kept <= xb->tsize )
    xb->lsize = xb->tsize - kept;
}

char *chainxb( xbuffer *xb, char *str, int len ) {
  if( xb->lsize < xb->tsize )
    xb->lsize++;
  return( appendxb( xb, str, len ) );
}

char *catxb( xbuffer *xb, char *strz ) {
  if( xb->lsize < xb->tsize )
    xb->lsize++;
  return( appendxb( xb, strz, strlen( strz ) + 1 ) );
}

void freexb( xbuffer *xb ) {
  FREE( xb->line )
  xb->tsize = 0;
  xb->lsize = 0;
}

int lenxb( xbuffer *xb ) {
  return( xb->tsize - xb->lsize );
}

char *ptrxb( xbuffer *xb ) {
  return( xb->line );
}

char *printxb( xbuffer *xb, char *format, ... ) {
  va_list args;
  char *proc;
  char *s;
  int d;
  static char dtoa[128];
 
  emptyxb( xb );
 
  va_start( args, format );
  do {
    int len;
    
    proc = strchr( format, '%' );
    len = (proc ? proc - format : strlen( format ));
    if( len && NULL == appendxb( xb, format, len ) )
      break;
    if( NULL == proc ) {
      appendxb( xb, "", 1 );
      break;
    }
    format = proc + 2;
    switch( *(proc + 1) ) {
      case 's':
	s = va_arg( args, char * );
	len = strlen( s );
	if( len && NULL == appendxb( xb, s, len ) )
	  break;
	continue;
      case 'd':
	d = va_arg( args, int );
	sprintf( dtoa, "%d", d );
	len = strlen( dtoa );
	if( len && NULL == appendxb( xb, dtoa, len ) )
          break;
        continue;
      default:
	break;
    }
    break;
  }while( 1 );
  va_end( args );

  s = ptrxb( xb );
  return( s ? s : "" );
}

my64_t stopwatch( const int stsp ) {
  static struct timespec tst;

  if( 0 == stsp ) {
    struct timespec tnd;
    register struct timeval tdf;

    (void) clock_gettime( CLOCK_MONOTONIC, &tnd );
    tdf.tv_sec = tnd.tv_sec - tst.tv_sec;
    tdf.tv_usec = (tnd.tv_nsec - tst.tv_nsec) / 1000;
    if( tdf.tv_usec < 0 ) tdf.tv_sec--, tdf.tv_usec += 1000000;
    return( tdf.tv_sec * 1000000 + tdf.tv_usec );
  }

  clock_gettime( CLOCK_MONOTONIC, &tst );
  return( 0 );
}
