/*
 * bbftpc/connecttoserver.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */

#include "bbftp.h"
#include <sys/types.h>

#include <errno.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <time.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>

#include "client.h"
#include "structures.h"
#include "client_proto.h"
#include "common.h"
#include "config.h"

#ifdef WITH_SSL
RSA	*hisrsa = NULL;
RSA	*myprsa = NULL;
int	crypton = 1;
unsigned char ciphkey[16];
unsigned char ciphiv[8];
#endif

#define SETTOZERO    0
#define SETTOONE     1

#ifndef INADDR_LOOPBACK
#define INADDR_LOOPBACK        ((in_addr_t) 0x7f000001) /* Inet 127.0.0.1. */
#endif


extern  int     timestamp ;
extern  int     usessh ;
extern	int	usecert ;
extern  int     sshbatchmode ;
extern  char    *sshidentityfile ;
extern  char    *sshremotecmd  ;
extern  char    *sshcmd ;
extern  char    *hostname ;
extern  char    *username ;
extern  char    *password ;
extern  int     incontrolsock ;
extern  int     outcontrolsock ;
extern  int     globaltrymax ;
extern  int     warning ;
extern  int     debug ;
extern  int     verbose ;
extern  int     newcontrolport ;
extern  int     transferoption ;
extern  int     remoteumask ;
extern  int     remotecos ;
extern  char    *remotedir ;

extern  struct  sockaddr_in hisctladdr ;

extern xbuffer logmessage;
extern int errcode;

extern int pasvport_max;

pid_t	sshchildpid = (pid_t)(0);


static char *askpassword() {
	
#ifdef PRIVATE_AUTH
    if ( bbftp_private_getargs() < 0 ) {
      printmessage(stderr,CASE_FATAL_ERROR,19,timestamp,"Error while private authentication : %s\n", ptrxb( &logmessage ) );
    }
    return( password );
#endif	/* PRIVATE_AUTH */
    
    if ((!usessh && !usecert)) {
	static char passwordBuffer[512];
	static char passMsg[512];

	sprintf( passMsg, "%s@%s's Password: ", username, hostname );

#ifdef AFS
        char *reasonString;
        long code ;
	
        code = ka_UserReadPassword( passMsg, passwordBuffer, sizeof(passwordBuffer), &reasonString);
	if( code )
            printmessage(stderr,CASE_FATAL_ERROR,19,timestamp,"ka_UserReadPassword: %s\n", reasonString) ;
#else	/* AFS */
	readpassphrase( passMsg, passwordBuffer, sizeof( passwordBuffer ), 0 );
#endif	/* AFS */
	return( passwordBuffer );
    }
}


int get_ip( char *host, char **fhost ) {
  struct sockaddr_in sn;
  char *alist[1] = { NULL };
  struct in_addr defaddr;
  struct hostent *hp, def;

  hp = gethostbyname( host );
  if( NULL == hp ) {
    defaddr.s_addr = inet_addr( host );
    if( -1 == defaddr.s_addr )
      return( INADDR_ANY );	/* zero */
    
    def.h_name = host;
    def.h_addr_list = alist;
    def.h_addr = (char *)&defaddr;
    def.h_length = sizeof( struct in_addr );
    def.h_addrtype = AF_INET;
    def.h_aliases = 0;
    hp = &def;
  }
  if( hp->h_length > (int)sizeof( sn.sin_addr ) )
    hp->h_length = sizeof( sn.sin_addr );
  memcpy( &sn.sin_addr, hp->h_addr, hp->h_length );

  if( fhost )
    *fhost = strdup( hp->h_name );

  return( sn.sin_addr.s_addr );
}


int bbftp_connect( int sock, int port, int ip ) {
  struct sockaddr_in data_source;
#if defined(SUNOS) || defined(_HPUX_SOURCE) || defined(IRIX) || defined(OSF1)
  int    addrlen;
  int    erlen;
#else
  size_t addrlen;
  size_t erlen;
#endif
  int    retcode;

  if( -1 == sock ) {	  
    sock = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );
    if( sock < 0 )
      return( -errno );

    data_source.sin_family = AF_INET;
    data_source.sin_addr.s_addr = INADDR_ANY;
    data_source.sin_port = 0;
    if( bind( sock, (struct sockaddr *)(&data_source), sizeof( data_source ) ) < 0) {
      int er;

      er = errno;
      close( sock );
      return( -er );
    }
  }
  
  data_source.sin_family = AF_INET;
  data_source.sin_addr.s_addr = INADDR_ANY == (in_addr_t)(ip) ? hisctladdr.sin_addr.s_addr : (in_addr_t)(ip);
  data_source.sin_port = port;

  addrlen = sizeof(data_source) ;

  retcode = connect( sock, (struct sockaddr *)(&data_source), addrlen );
  if( retcode < 0 ) {
    if( EINPROGRESS == errno ) {
      fd_set mask;
      struct timeval timer;
      int er;
     
      for(;;) {
        FD_ZERO( &mask );
        FD_SET( sock, &mask );
        timer.tv_sec = 30;        /* thirty seconds */
        timer.tv_usec = 0;
        retcode = select( sock + 1, 0, &mask, 0, &timer );
        if( retcode <= 0 ) {
	  if( 0 == retcode )
	    errno = ETIMEDOUT;
	  else if( EINTR == errno )
	    continue;
          return( -errno );
        }
	break;
      }
      erlen = sizeof( er );
      if( getsockopt( sock, SOL_SOCKET, SO_ERROR, (char *)&er, (socklen_t *)&erlen ) < 0 )
	return( -errno );
      if( 0 == er )
	goto bc001;
      errno = er;
    }
    return( -errno );
  }
 
bc001:
  /*XXX*/
  if( debug ) {
	  printmessage(stderr,CASE_NORMAL,91,timestamp,"Connection(%s:%d <-- ",inet_ntoa(data_source.sin_addr),ntohs( data_source.sin_port ));
	  addrlen = sizeof( data_source ); /* reused */
	  if( getsockname( sock, (struct sockaddr *)&data_source, (socklen_t *)&addrlen ) < 0 ) {
		  int er;

		  er = errno;
		  close( sock );
		  return( -er );
	  }
	  printmessage(stderr,CASE_NORMAL,91,timestamp,"%s:%d)\n",inet_ntoa(data_source.sin_addr),ntohs( data_source.sin_port ));
  }

  return( sock );
}


int getcrypt( struct mess_sec *msg_sec ) {
  int    crtype;
  int    lenkey ;
  int    lenexpo ;
  unsigned char *pubkey ;
  unsigned char *pubexponent ;

#ifndef WORDS_BIGENDIAN
  crtype = ntohl( msg_sec->crtype );
  lenkey = ntohl( msg_sec->pubkeylen );
  lenexpo = ntohl( msg_sec->expolen );
#else
  crtype = msg_sec->crtype;
  lenkey = msg_sec->pubkeylen ;
  lenexpo = msg_sec->expolen ;
#endif

  if( (crtype & CRYPT_RSA_PKCS1_OAEP_PADDING ) != CRYPT_RSA_PKCS1_OAEP_PADDING ) {
      printmessage(stderr,CASE_ERROR,57,timestamp,"Unkown encryption method \n") ;
      return( BB_RET_ERROR );
  }

#ifdef WITH_SSL
  if( hisrsa )
    RSA_free( hisrsa );
  
  hisrsa = key_new();
  if( NULL == hisrsa ) {
    printmessage(stderr,CASE_ERROR,56,timestamp,"RSA_new: %s\n",(char *) ERR_error_string(ERR_get_error(),NULL)) ;
    return( BB_RET_ERROR );
  }

  /*
  ** Copy the key and exponent received
  */
  pubkey  = (unsigned char *)(msg_sec + 1);
  pubexponent = pubkey + lenkey ;

  {
	  BIGNUM *n, *e;
	  
	  RSA_get0_key( hisrsa,(const BIGNUM **)(&n), (const BIGNUM **)(&e), NULL );

	  if( NULL == BN_mpi2bn( pubkey, lenkey, n ) || NULL == BN_mpi2bn( pubexponent, lenexpo, e ) ) {
		  printmessage(stderr,CASE_ERROR,56,timestamp,"BN_mpi2bn: %s\n",(char *) ERR_error_string(ERR_get_error(),NULL)) ;
		  return( BB_RET_ERROR );
	  }
  }

  /*XXX*//*if( debug ) RSA_print_fp( stderr, hisrsa, 8 );*/
#endif	/* WITH_SSL */

  return( BB_RET_OK );
}


int saversa( char *buffer ) {
#ifdef	WITH_SSL
  int bnlen;

  BIGNUM *n, *e;

  hisrsa = key_new();

  if( NULL == hisrsa ) {
    printmessage( stderr, CASE_ERROR, 56, timestamp, "RSA_new: %s\n", (char *)(ERR_error_string( ERR_get_error(), NULL )) );
    return( BB_RET_ERROR );
  }
	
  memcpy( (char *)(&bnlen), buffer, sizeof( int ) );  
#ifndef WORDS_BIGENDIAN
  bnlen = ntohl( bnlen );
#endif
  
  n = BN_mpi2bn( buffer + sizeof( int ), bnlen, NULL );
  if( NULL == n ) {
    printmessage( stderr, CASE_ERROR, 56, timestamp, "BN_mpi2bn: %s\n", (char *)(ERR_error_string( ERR_get_error(), NULL )) );
    return( BB_RET_ERROR );
  }

  buffer += sizeof( int ) + bnlen;
     
  memcpy( (char *)(&bnlen), buffer, sizeof( int ) ); 
#ifndef WORDS_BIGENDIAN
  bnlen = ntohl( bnlen );
#endif

  e = BN_mpi2bn( buffer + sizeof( int ), bnlen, NULL );
  if( NULL == e ) {
    printmessage( stderr, CASE_ERROR, 56, timestamp, "BN_mpi2bn: %s\n", (char *)(ERR_error_string( ERR_get_error(), NULL )) );
    return( BB_RET_ERROR );
  }

  RSA_set0_key( hisrsa, n, e, NULL );

  /*XXX*//*if( debug ) RSA_print_fp( stderr, hisrsa, 8 );*/

  return( BB_RET_OK );
#endif	/* WITH_SSL */

  return( BB_RET_ERROR );
}

#ifdef  WITH_SSL
static BIGNUM *saveBN( char *buffer ) {
  BIGNUM *bn = NULL;
  int bnlen;
  
  bn = BN_new();
  if( NULL == bn ) {
    printmessage( stderr, CASE_ERROR, 56, timestamp, "BN_new: %s\n", (char *)(ERR_error_string( ERR_get_error(), NULL )) );
    return( NULL );
  }
  
  memcpy( (char *)(&bnlen), buffer, sizeof( int ) );
#ifndef WORDS_BIGENDIAN
  bnlen = ntohl( bnlen );
#endif

  bn = BN_mpi2bn( buffer + sizeof( int ), bnlen, bn );
  if( NULL == bn ) {
    printmessage( stderr, CASE_ERROR, 56, timestamp, "BN_mpi2bn: %s\n", (char *)(ERR_error_string( ERR_get_error(), NULL )) );
    return( NULL );
  }

  return( bn );

  return( NULL );
}
#endif	/* WITH_SSL */


static void gencipher() {
#ifdef	WITH_SSL
  unsigned int seed ;
  struct timespec tp ;
  int    i;

  clock_gettime( CLOCK_MONOTONIC, &tp );
  seed = tp.tv_nsec;
  srandom(seed);

  for( i = 0; i < sizeof( ciphkey ); ++i )
    ciphkey[i] = random();

  for( i = 0; i < sizeof( ciphiv ); ++i )
    ciphiv[i] = random();
#endif	/* WITH_SSL */
}


int sendrsa( int code ) {
#ifdef	WITH_SSL
  int lene;
  int lenn;
  unsigned char *to;
  int tolen;
  int retcode;

  const BIGNUM *n, *e;

  RSA_get0_key( myprsa, &n, &e, NULL );

  lenn = BN_bn2mpi( n, NULL );
  lene = BN_bn2mpi( e, NULL );

  tolen = sizeof( int ) + lenn + sizeof( int ) + lene;
  to = (unsigned char *)(malloc( tolen ) );
  if( NULL == to ) {
    printmessage(stderr,CASE_ERROR,52,timestamp,"Malloc: %s\n", strerror( errno ) );
    return( BB_RET_ERROR );
  }
  BN_bn2mpi( n, to + sizeof( int ) );
  BN_bn2mpi( e, to + sizeof( int ) + lenn + sizeof( int ) );

#ifndef WORDS_BIGENDIAN
  retcode = htonl( lenn );
  memcpy( to, (char *)(&retcode), sizeof( int ) );
  retcode = htonl( lene );
  memcpy( to + sizeof( int ) + lenn, (char *)(&retcode), sizeof( int ) );
#else
  memcpy( to, (char *)(&lenn), sizeof( int ) );
  memcpy( to + sizeof( int ) + lenn, (char *)(&lene), sizeof( int ) );
#endif

  /* cipher too */
  gencipher();

  retcode = writepackage( code, tolen + 16 + 8, to, tolen, ciphkey, 16, ciphiv, 8 );	/* Either MSG_NKEY or MSG_KEY */
  FREE( to )
  if( retcode < 0 ) {
    printmessage(stderr,CASE_ERROR,58,timestamp,"Error sending %s message\n", MESSAGE_NAMES[MSG_KEY] );
    return( BB_RET_ERROR );
  }

  return( BB_RET_OK );
#endif	/* WITH_SSL */

  return( BB_RET_ERROR );
}


int genmyprsa() {
#ifdef	WITH_SSL
  ////myprsa = RSA_generate_key(NBITSINKEY,35,NULL,NULL);

	BIGNUM *e;
	int ret;

	myprsa = RSA_new();
	if( NULL == myprsa ) {
		gp1:
		printmessage( stderr, CASE_ERROR, 27, timestamp, "genmyprsa: %s\n", ERR_error_string( ERR_get_error(), NULL) );
		return( BB_RET_ERROR );
	}

	e = BN_new();
	if( NULL == e ) goto gp1;
	BN_set_word(e, RSA_F4);

	ret = RSA_generate_key_ex( myprsa, NBITSINKEY, e, NULL );
	BN_free( e );

	if( 0 == ret ) goto gp1;

  return( BB_RET_OK );
#endif	/* WITH_SSL */
  
  return( BB_RET_ERROR );
}


int f_handler_ssh( char **buffer, int *len, int *code ) {
  
  if( *code != MSG_NKEY )
    return( BB_RET_NONE );

#ifdef  WITH_SSL
  if( saversa( *buffer ) != BB_RET_OK )
    return( BB_RET_ERROR );
 
  if( genmyprsa() != BB_RET_OK )
    return( BB_RET_ERROR );

  if( sendrsa( MSG_NKEY ) != BB_RET_OK )
    return( BB_RET_ERROR );
#endif  /* WITH_SSL */

  return( BB_RET_OK );
}


int connectviassh() {
  int pin[2], pout[2];

  if( -1 == pipe( pin ) || -1 == pipe( pout ) )
    printmessage( stderr, CASE_FATAL_ERROR, 41, timestamp, "Pipe: %s\n", strerror( errno ) ); 

  sshchildpid = fork();
  switch( (int)(sshchildpid) ) {
    char *argv[4];
    int ret;

    case 0:
      argv[0] = "sh";
      argv[1] = "-c";
      argv[3] = NULL;
				
      close( pin[1] );
      if( 0 != pin[0] ) {
        dup2( pin[0], 0 );
	close( pin[0] );
      }
      close( pout[0] );
      if( 1 != pout[1] ) {
        dup2( pout[1], 1 );
	close( pout[1] );
      }
    
      argv[2] = printxb( &logmessage, "%s -x -a -o 'FallBackToRsh no' %s -i %s -l %s %s %s", 
		                      sshcmd, 
			              ( sshbatchmode ? "-o 'BatchMode yes' -o 'StrictHostKeyChecking no'" : "" ),
			              (char *)(expand( sshidentityfile )),
			              username,
			              hostname,
			              sshremotecmd
	               );
      
      if( debug )
        printmessage( stderr/*XXX*/, CASE_NORMAL, 0, timestamp, "sh -c %s\n", ptrxb( &logmessage ) );
       
      execvp( argv[0], argv );
      printmessage( stderr, CASE_FATAL_ERROR, 46, timestamp, "Execvp: %s\n", strerror( errno ) );
      /* game over */
    
    case -1: 
      close( pin[0] );
      close( pin[1] );
      close( pout[0] );
      close( pout[1] );
      printmessage( stderr, CASE_ERROR, 45, timestamp, "Fork: %s\n", strerror( errno ) );
      return( -1 );
    
    default:
      close( pin[0] );
      outcontrolsock = pin[1];
      close( pout[1] );
      incontrolsock = pout[0];

      ret = handlepackage( handler_ssh );
      switch( ret ) {
        case BB_RET_OK:
	  return( BB_RET_OK );
        case BB_RET_ERROR:
        case BB_RET_FT_NR:
	  bbftp_close_control();
	 /* fall through */
        default:
	  kill( sshchildpid, SIGTERM );
	  while( waitpid( sshchildpid, NULL, WNOHANG ) != sshchildpid ) /* do nothing */ ;
	  return( ret );
      }
  }
}


int f_handler_passwd( char **buffer, int *len, int *code ) {
  static struct mess_sec cryptbuffer;
  static struct mess_rsa rsabuffer;
  struct mess_rsa *msg_rsa = &rsabuffer;

start:
  if( *code != MSG_CRYPT )
    return( BB_RET_NONE );

  if( getcrypt( (struct mess_sec *)(*buffer) ) != BB_RET_OK )
    return( BB_RET_ERROR );

#ifdef	PRIVATE_AUTH
  if( bbftp_private_auth() < 0 ) {
    printmessage(stderr,CASE_ERROR,27,timestamp,"Error while private authentication : %s\n", ptrxb( &logmessage ) );
    return( BB_RET_ERROR );
  }
  goto end;
#endif	/* PRIVATE_AUTH */
  
#ifdef WITH_SSL
	if( crypton != 0 ) {
	  int i;
	  char *pp;
	  BIGNUM *challenge = NULL;
	  unsigned char *response;
	  
	  i = strlen( username );
	  if( writepackage( MSG_USER, i, username, i ) < 0 ) {
	    printmessage(stderr,CASE_ERROR,58,timestamp,"Error sending %s message\n", MESSAGE_NAMES[MSG_USER] );
	    return( BB_RET_ERROR );
	  }

	  *buffer = readpackage( len, code, NULL );
	  if( *code != MSG_OK ) {
	    if( verbose && *len > 0 ) printmessage( stderr, CASE_NORMAL, 0, timestamp, "<< %s\n", *buffer);
	    crypton = 0;
	    bbftp_close_control();
            incontrolsock = bbftp_connect( -1, htons( newcontrolport ), 0 );
            if( incontrolsock < 0 ) {
              printmessage(stderr,CASE_ERROR,52,timestamp,"bbftp_connect: %s\n", strerror( -incontrolsock ) );
              return( BB_RET_ERROR );
            }
            outcontrolsock = incontrolsock;

            *buffer = readpackage( len, code, NULL );
            goto start;
	  }

	  pp = "";
	  for( i = 0; i < 2; ++i ) {
	    char *rn = NULL;
	    
	    myprsa = key_load_private( sshidentityfile, pp, &rn );
	    if( myprsa != NULL) {
	      FREE( rn )
	      if( i != 0 )
		memset( pp, strlen( pp ), 0 );
	      break;
	    }
	    if( rn != NULL ) {
	      static char passphrase[512];
	      
	      FREE( rn )
	      if( i != 0 )
                break;
	      readpassphrase( "Passphrase: ", passphrase, sizeof( passphrase ), 0 );
	      if( 0 == strlen( passphrase ) )
		break;
	      pp = passphrase;
	      continue;
	    }
	    if( verbose ) printmessage(stderr,CASE_NORMAL,0,timestamp,"Id file %s unavailable\n", sshidentityfile );
	    break;
	  }
	 
	  if( NULL == myprsa ) {
		  
	    if( genmyprsa() != BB_RET_OK )
	      return( BB_RET_ERROR );
	    
	    if( sendrsa( MSG_NKEY ) != BB_RET_OK )
	      return( BB_RET_ERROR );
	    
	  pass:
	    password = askpassword();
	    i = strlen( password );
	    if( writepackage( MSG_PASS, i, password, i ) < 0 ) {
	      printmessage(stderr,CASE_ERROR,58,timestamp,"Error sending %s message\n", MESSAGE_NAMES[MSG_PASS] );
	      return( BB_RET_ERROR );
	    }
	    goto end;
	  }
	 
#if 0 
	  if( debug ) {
	    RSA *drsa;
	      
	    drsa = key_demote( myprsa );
	    RSA_print_fp( stderr, drsa, 8 );
	    RSA_free( drsa );
	  }
#endif

	  if( sendrsa( MSG_KEY ) != BB_RET_OK )
            return( BB_RET_ERROR );

	  *buffer = readpackage( len, code, NULL );
	  switch( *code ) {
	    case MSG_CHALLENGE:
	      break;      
	    case MSG_BAD:
	      if( verbose && *len > 0 ) printmessage(stderr,CASE_NORMAL,0,timestamp,"<< %s\n", *buffer);
	      goto pass;
	    default:
	      return( BB_RET_NONE );
	  }
	    
	  /*XXX*/
	  challenge = saveBN( *buffer );
	  if( NULL == challenge )
	    goto pass;

	  response = respond_to_challenge( challenge );
	  if( NULL == response )
	    goto pass;
	
	  i = writepackage( MSG_RESPONSE, 16, response, 16 );
	  memset( response, 0, 16 );
	    
	  if( i < 0 ) {
            printmessage(stderr,CASE_ERROR,58,timestamp,"Error sending %s message\n", MESSAGE_NAMES[MSG_RESPONSE] );
            return( BB_RET_ERROR );
          }
	   
	  goto end;
	}
										
	password = askpassword();

	msg_rsa->numuser = RSA_public_encrypt(strlen(username),(unsigned char *)username,msg_rsa->cryptuser,hisrsa,RSA_PKCS1_OAEP_PADDING);
	if( msg_rsa->numuser < 0 ) {
            printmessage(stderr,CASE_ERROR,56,timestamp,"RSA_public_encrypt(username): %s\n", (char *)(ERR_error_string( ERR_get_error(), NULL )) );
            return( BB_RET_ERROR );
        }

	msg_rsa->numpass = RSA_public_encrypt(strlen(password),(unsigned char *)password,msg_rsa->cryptpass,hisrsa,RSA_PKCS1_OAEP_PADDING);
	if( msg_rsa->numpass < 0 ) {
            printmessage(stderr,CASE_ERROR,56,timestamp,"RSA_public_encrypt(password): %s\n", (char *)(ERR_error_string( ERR_get_error(), NULL )) );
            return( BB_RET_ERROR );
        }

#ifndef WORDS_BIGENDIAN
	cryptbuffer.crtype = htonl( CRYPT_RSA_PKCS1_OAEP_PADDING );
#else
	cryptbuffer.crtype = CRYPT_RSA_PKCS1_OAEP_PADDING;
#endif

#else /*WITH_SSL*/

    /* Client does not support encryption */
    if( writepackage( MSG_CRYPT, 0 ) < 0 ) {
      printmessage(stderr,CASE_ERROR,58,timestamp,"Error sending %s message: %s\n",MESSAGE_NAMES[MSG_CRYPT],strerror(errno));
      return( BB_RET_ERROR );
    }

    *buffer = readpackage( len, code, NULL );
    if( *code != MSG_OK )
      return( BB_RET_NONE );
    
    if( verbose && *len > 0 ) printmessage(stderr,CASE_NORMAL,0,timestamp,"<< %s\n", *buffer);
  
    password = askpassword();
    
    cryptbuffer.crtype = NO_CRYPT;	/* Plain zero */

    msg_rsa->numuser = strlen(username) ;
    strcpy(msg_rsa->cryptuser, (unsigned char *)username) ;

    msg_rsa->numpass = strlen(password) ;
    strcpy(msg_rsa->cryptpass, (unsigned char *)password) ;
#endif /*WITH_SSL*/

    memset( password, 0, strlen( password ) );
    
#ifndef WORDS_BIGENDIAN
    msg_rsa->numuser = ntohl( msg_rsa->numuser );
    msg_rsa->numpass = ntohl( msg_rsa->numpass );
#endif

    if( writepackage( MSG_LOG, sizeof( struct mess_sec ) + sizeof( struct mess_rsa ), &cryptbuffer, sizeof( struct mess_sec ), msg_rsa, sizeof( struct mess_rsa ) ) < 0 ) {
      printmessage(stderr,CASE_ERROR,58,timestamp,"Error sending %s message: %s\n", MESSAGE_NAMES[MSG_LOG], strerror(errno)) ;
      return( BB_RET_ERROR );
    }
		
    if ( verbose) printmessage(stderr,CASE_NORMAL,0,timestamp,">> USER %s PASS\n",username) ;
    
    /*
    ** Now wait for the OK message
    */
end:
    *buffer = readpackage( len, code, NULL );
    if( *code != MSG_OK )
      return( BB_RET_NONE );
    
    if( verbose && *len > 0 ) printmessage(stderr,CASE_NORMAL,0,timestamp,"<< %s\n", *buffer);

    return( BB_RET_OK );
}

/****************************************************************************

 connectviapassword is the standart method to connect to bbftpd

*****************************************************************************/

int connectviapassword() {
    int ret;

    /*
    ** Get the socket
    */
    incontrolsock = bbftp_connect( -1, htons( newcontrolport ), 0 );
    if( incontrolsock < 0 ) {
      printmessage(stderr,CASE_ERROR,52,timestamp,"bbftp_connect: %s\n", strerror( -incontrolsock ) );
      return( BB_RET_ERROR );
    }

    /*
    ** Connection is correct so start the login procedure
    */
    outcontrolsock = incontrolsock;

    ret = handlepackage( handler_passwd );
    if( BB_RET_ERROR == ret )
      bbftp_close_control();
    if( BB_RET_FT_NR == ret )
      exit( errno = ENETDOWN );

    return( ret );
}


int todoafterconnection() {
    int     retcode ;
    /*
    ** Things to do are :
    **      chdir remote if remotedir != NULL
    **      chumask remote if remoteumask !=0
    **      remotecos
    */
    if ( remoteumask !=0 ) {
        if ( bbftp_setremoteumask( remoteumask ) != BB_RET_OK ) {
            return -1 ;
        }
    }
	if ( bbftp_setremotecos( remotecos ) != BB_RET_OK ) {
		return -1 ;
    }
    if ( remotedir != NULL ) {
        /*
        ** Care has to be taken NO MORE if we are under remoterfio
        ** We have first to change the option to noremoterfio,
        ** make the cd command and reset the option
        */
        retcode = bbftp_cd( remotedir );
        if ( retcode != BB_RET_OK ) return -1 ;
    }
    return 0 ;
}


void reconnecttoserver() {
    int     nbtrycon ;
    int     retcode ;

    for( nbtrycon = 0; nbtrycon < globaltrymax; ++nbtrycon ) {

#ifdef	WITH_SSL
	recrypt();	/* cipher off, RSA off */
#endif	/* WITH_SSL */
	    
#ifdef CERTIFICATE_AUTH
        if (usecert) {
            if ( debug ) printmessage(stderr,CASE_NORMAL,0,timestamp,"Certificate authentication...\n") ;
            retcode = bbftp_cert_connect() ;
            if( retcode != BB_RET_OK ) {
		/* Falling back to RSA authentication */
		usecert = 0;
		usessh = 0;
		nbtrycon = globaltrymax - 2;
		continue;
            }
        } else
#endif /*CERTIFICATE_AUTH*/
        if (usessh) {
            if ( debug ) printmessage(stderr,CASE_NORMAL,0,timestamp,"SSH connection...\n") ;
            retcode = connectviassh() ;
        } else {
            if ( debug ) printmessage(stderr,CASE_NORMAL,0,timestamp,"RSA authentication...\n") ;
            retcode = connectviapassword() ;
        }
	
        if( BB_RET_OK == retcode ) {
            if ( debug ) printmessage(stderr,CASE_NORMAL,0,timestamp,"Connection and authentication correct\n") ;

#ifdef	WITH_SSL
	    if( crypton ) crypton |= 2;	/* cipher on */
#endif	/* WITH_SSL */
	    
	    switch( sendproto() ) {
	      case BB_RET_OK:
		retcode = todoafterconnection() ;
		if( 0 == retcode )
		  return;
		bbftp_close_control();
		break;
	      case BB_RET_ERROR:
	      case BB_RET_FT_NR:
		bbftp_close_control();
		break;
	      default:	/* do nothing */
		break;
	    }
        }
        if( nbtrycon < globaltrymax - 1 ) {
            if (warning) printmessage(stderr,CASE_WARNING,22,timestamp,"Retrying connection waiting %d s\n",WAITRETRYTIME) ;
            sleep( WAITRETRYTIME );
        }
    }
    if( nbtrycon == globaltrymax ) {
        printmessage(stderr,CASE_FATAL_ERROR,33,timestamp,"Maximum try on connection reached (%d) aborting\n",globaltrymax) ;
    }
}
