/*
 * bbftpc/getdatasock.c
 * Copyright (C) 1999, 2000, 2001, 2002 IN2P3, CNRS
 * bbftp@in2p3.fr
 * http://doc.in2p3.fr/bbftp
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */ 

/****************************************************************************

  
  
 getdatasock.c  v 0.0.0 2000/01/11
                v 1.6.0 2000/03/24  - Set Receive buffer to TCPWINSIZE
                                      for the get option
                v 1.7.1 2000/04/04  - Variable cleaning (detected on IRIX)
                v 1.8.7 2000/05/28  - Change headers
                v 1.8.9 2000/08/08  - Add timestamp
                v 1.9.0 2000/08/21  - Use configure to help portage
                v 2.0.0 2000/03/19  - Use protocol version 2
                v 2.0.1 2001/04/19  - Correct indentation 
                                    - Port to IRIX

*****************************************************************************/
#include "bbftp.h"

#include <unistd.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <errno.h>

#include <fcntl.h>
#include <string.h>

#include "client.h"
#include "structures.h"
#include "client_proto.h"
#include "common.h"


extern int  pasvport_min ;
extern int  pasvport_max ;

extern int     debug ;
extern int     timestamp ;
extern unsigned short int     *myports ;
extern int     *mysockets ;
extern int      warning ;
extern int     	transferoption  ; 

extern int errcode;

extern int	*myips;
extern char	*localips;

extern	int     requestedstreamnumber;


extern int natip;


int *nips = NULL;
int sip = 0;

static int nextip() {
  static int no = 0;
  static int crt = 0;
  static int any = INADDR_ANY;
  int nip;

  if( 0 == no ) {
    char *ip;

    ip = strtok( localips, ", " );
    while( ip ) {
      int *ptr;

      ptr = (int *)(realloc( nips, (no + 1) * sizeof( int ) ));
      if( NULL == ptr ) {
	printmessage( stderr, CASE_ERROR, 35, timestamp, "Realloc: %s\n", strerror( errno ) );
        break;
      }
      nips = ptr;
      nips[no] = get_ip( ip, NULL );
      ++no;
      ip = strtok( NULL, ", " );
    }
    if( 0 == no ) {
      sip = 1;
      nips = &any;
      ++no;
    }
  }
  nip = nips[crt];
  crt = (crt + 1) % no;
  return( nip );
}


int getdatasock( int nbsock ) {
    int     i ;
    int     on = 1 ;
#if defined(SUNOS) || defined(_HPUX_SOURCE) || defined(IRIX) || defined(OSF1)
    int     addrlen ;
#else
    size_t  addrlen ;
#endif
    struct  linger li ;
    struct  sockaddr_in sck ;
    int     sock ;
    /* Declaring the variables need to change the value of the TOS */
    int qbss_value = 0x20;

    int port = pasvport_min;


    sck.sin_family = AF_INET;
    
    for( i = 0; i < nbsock; ++i ) {

	sck.sin_addr.s_addr = nextip();

        sock = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP ) ;
        if ( sock < 0 ) {
            printmessage(stderr,CASE_ERROR,91,timestamp,"Cannot create data socket: %s\n",strerror(errno));
            errcode = 91;
            break;
        }

        if ( setsockopt(sock,SOL_SOCKET, SO_REUSEADDR,(char *)&on,sizeof(on)) < 0 ) {
            printmessage(stderr,CASE_ERROR,92,timestamp,"Cannot set SO_REUSEADDR on data socket : %s\n",strerror(errno));
            errcode = 92;
            break;
        }

        if ( setsockopt(sock,SOL_SOCKET, SO_KEEPALIVE,(char *)&on,sizeof(on)) < 0 ) {
            printmessage(stderr,CASE_ERROR,92,timestamp,"Cannot set SO_KEEPALIVE on data socket : %s\n",strerror(errno));
            errcode = 92;
            break;
        }	

        if ( setsockopt(sock,IPPROTO_TCP, TCP_NODELAY,(char *)&on,sizeof(on)) < 0 ) {
            printmessage(stderr,CASE_ERROR,93,timestamp,"Cannot set TCP_NODELAY on data socket : : %s\n",strerror(errno));
            errcode = 93;
            break;
        }

		if( transferoption & TROPT_QBSS ) {
			/* Setting the value for IP_TOS to be 0x20 */
			if ( setsockopt(sock,IPPROTO_IP, IP_TOS, (char *)&qbss_value, sizeof(qbss_value)) < 0 ) {
				printmessage(stderr,CASE_ERROR,93,timestamp, "Cannot set IP_TOS on data socket : : %s\n", strerror(errno)); 
			}
		}

        li.l_onoff = 1 ;
        li.l_linger = 1 ;
        if ( setsockopt(sock,SOL_SOCKET,SO_LINGER,(char *)&li,sizeof(li)) < 0 ) {
            printmessage(stderr,CASE_ERROR,94,timestamp,"Cannot set SO_LINGER on data socket : %s\n",strerror(errno));
            errcode = 94;
            break;
        }


        if (pasvport_min) { /* Try to bind within a range */
                for( ; port <= pasvport_max; ++port ) {
                  sck.sin_port = htons(port);
                  if (bind(sock, (struct sockaddr *) &sck, sizeof(sck)) >= 0) break;
                }
                if( port > pasvport_max ) {
                  if( i > 0 ) {
                    int j;
		    int k;
			
		    /* round robin reuse of sockets, no more keyhole... */
		    for( j = i, k = 0; j < nbsock; ++j ) {
		      myips[j] = myips[k];
		      myports[j] = myports[k];
		      mysockets[j] = mysockets[k];
		      ++k;
		      if( k == i ) k = 0;
		    }
		    return( 0 );
	          }
		  printmessage(stderr,CASE_ERROR,95,timestamp,"Cannot bind on data socket : %s\n",strerror(errno));
		  errcode = 95;
		  break;
		}
		else ++port;
        } 
	else {
	    sck.sin_port = 0;
            if ( bind(sock, (struct sockaddr *) &sck ,sizeof(sck))  < 0) {
                printmessage(stderr,CASE_ERROR,95,timestamp,"Cannot bind on data socket : %s\n",strerror(errno));
                errcode = 95;
               break;
            }
        }

        addrlen = sizeof(sck) ;
        if (getsockname(sock,(struct sockaddr *)&sck, (socklen_t *)&addrlen) < 0) {
            printmessage(stderr,CASE_ERROR,96,timestamp,"Cannot getsockname on data socket : %s\n",strerror(errno));
            errcode = 96;
            break;
        }
        if (listen( sock, nbsock - i) < 0) {
	    printmessage(stderr,CASE_ERROR,97,timestamp,"Cannot listen on data socket : %s\n",strerror(errno));
            errcode = 97;
            break;
        }

	mysockets[i] = sock;
        myports[i]   = sck.sin_port ;
	myips[i] = natip ? natip : sck.sin_addr.s_addr;

        /*XXX*/if ( debug ) printmessage(stderr,CASE_NORMAL,0,timestamp,"Listen(%d) on %s:%d\n", i, inet_ntoa(sck.sin_addr), ntohs(sck.sin_port));
    }
    
    if( i < nbsock ) {
      int j;

      for( j = 0; j < i; ++j )
        close( mysockets[j] );
      if( sock >= 0 )
        close( sock );

      return( -1 );
    }
    
    return( 0 );
}


int makedatasock( int nbsock ) {
    int    i;
    int    sock ;
    struct sockaddr_in data_source ;
    int    on = 1 ;
#if defined(SUNOS) || defined(_HPUX_SOURCE) || defined(IRIX) || defined(OSF1)
    int    addrlen ;
#else
    size_t addrlen ;
#endif			  
    /* Declaring the variables need to change the value of the TOS */
    int qbss_value = 0x20;

    unsigned short int *sports;	/* List of active source ports */

    sports = (unsigned short int *)(malloc( nbsock * sizeof( unsigned short int ) ));
    if( NULL == malloc ) {
	    printmessage(stderr,CASE_ERROR,94,timestamp, "malloc : %s\n",strerror(errno));
	    errcode = 94; /* as good as any ? */
	    return( -1 );
    }

    for( i = 0; i < nbsock; ++i ) {
      sock = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP ) ;
      if ( sock < 0 ) {
        printmessage(stderr,CASE_ERROR,91,timestamp,"Cannot create data socket: %s\n",strerror(errno));
	errcode = 91;
	goto bye001;
      }

      /* make active and eventually reading sockets nonblock !! */
      if( -1 == fcntl( sock, F_SETFL, O_NONBLOCK ) ) {
        printmessage(stderr,CASE_ERROR,94,timestamp,"fcntl : %s\n",strerror(errno));
        errcode = 94; /* as good as any ? */
	goto bye001;
      }
     
      if ( setsockopt(sock,SOL_SOCKET, SO_KEEPALIVE,(char *)&on,sizeof(on)) < 0 ) {
        if (warning) printmessage(stderr,CASE_ERROR,92,timestamp,"Cannot set SO_KEEPALIVE on data socket : %s\n",strerror(errno));
        errcode = 92;
	goto bye001;
      }
     
      if ( setsockopt(sock,IPPROTO_TCP, TCP_NODELAY,(char *)&on,sizeof(on)) < 0 ) {
        printmessage(stderr,CASE_ERROR,94,timestamp,"Cannot set TCP_NODELAY on data socket : %s\n",strerror(errno));
	errcode = 94;
	goto bye001;
      }

      if( transferoption & TROPT_QBSS ) {
        /* Setting the value for IP_TOS to be 0x20 */
        if ( setsockopt(sock,IPPROTO_IP, IP_TOS, (char *)&qbss_value, sizeof(qbss_value)) < 0 ) {
          printmessage(stderr,CASE_ERROR,95,timestamp, "Cannot set IP_TOS on data socket : %s\n", strerror(errno));
	  errcode = 95;
	  goto bye001;
        }
      }
     
      data_source.sin_family = AF_INET;
      data_source.sin_addr.s_addr = nextip();
      data_source.sin_port = 0;

      if ( bind(sock, (struct sockaddr *) &data_source,sizeof(data_source)) < 0) {
        printmessage(stderr,CASE_ERROR,95,timestamp,"Cannot bind on data socket : %s\n",strerror(errno));
	errcode = 95;
	goto bye001;
      }
      addrlen = sizeof(data_source) ;
      if (getsockname(sock,(struct sockaddr *)&data_source, (socklen_t *)&addrlen) < 0) {
        printmessage(stderr,CASE_ERROR,96,timestamp,"Cannot getsockname on data socket : %s\n",strerror(errno));
	errcode = 96;
	goto bye001;
      }

      mysockets[i]  = sock;

      sports[i] = data_source.sin_port;
    }

    {
          int retcode;

          retcode = writepackage( MSG_POSITION, requestedstreamnumber * sizeof( unsigned short int ), sports, requestedstreamnumber * sizeof( unsigned short int ) );
          if( retcode < 0 ) {
		  printmessage(stderr,CASE_ERROR,64,timestamp,"Error sending %s message\n", MESSAGE_NAMES[MSG_POSITION] );
		  errcode = 64;
		  goto bye001;
	  }

	  if( debug ) printmessage(stderr,CASE_NORMAL,91,timestamp,"%s, sent!\n", MESSAGE_NAMES[MSG_POSITION] );

	  FREE( sports )
	  return( 0 );
    }

    {
      int j;

bye001:
      for( j = 0; j < i; ++j )
	close( mysockets[j] );
      if( sock >= 0 )
        close( sock );

      FREE( sports )      
      return( -1 );
    }
}
