/*
 * bbftpc/setsignals.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 

#include "bbftp.h"
#include <sys/types.h>

#include <errno.h>
#include <netinet/in.h>

#ifdef	__APPLE__
#define _POSIX_C_SOURCE 1L
#endif
#include <signal.h>

#include <stdio.h>

#include <unistd.h>
#include <string.h>

#include "client.h"
#include "structures.h"
#include "client_proto.h"
#include "common.h"


extern  int     timestamp;

extern	int	transferoption;

extern	volatile int *die;

extern	int debug;


static void bbftp_sigintfather( int sig ) {
	static int again = 0;

	++again;

	if( 1 == again ) { /* you only die once */
		if( debug ) printmessage( stderr, CASE_ERROR, EINTR, timestamp, "Killed by %s\n", (SIGTERM == sig ? "SIGTERM" : "SIGINT") );

		/* kill process group:
		 * the main bbftpPRO client, sh -c ssh if used, 
		 * the explicit piped command(s), and serialized transfers (-M 1) are in this group
		 */
		kill( 0, sig ); 
       	
		exit( errno = EINTR );
	}
}

static void bbftp_sigintson( int sig ) {
	static int again = 0;

	++again;

	if( 1 == again ) { /* you only die once */
		if( transferoption & TROPT_UNIX ) *die |= 64; /* signal TERM or INT error */
       	
		/* kill process group:
		 * the ringleader and all its children are in this group
		 */
		kill( 0, sig );
	}
}

static void bbftp_sigintchild( int sig ) {
      	if( transferoption & TROPT_UNIX ) *die |= 64; /* signal TERM or INT error */
 
       	exit( errno = EINTR ); /* does shmdt() */
}

int resetallsignals( void (*hndl)(int) /*__sighandler_t hndl*/ ) {
    struct sigaction sga;
    int ret = 0;

    sga.sa_handler= hndl ;
    sigemptyset(&(sga.sa_mask));
    sga.sa_flags = 0;

    ret += sigaction(SIGABRT,&sga,0);
    ret += sigaction(SIGALRM,&sga,0);
    ret += sigaction(SIGHUP,&sga,0);
    ret += sigaction(SIGINT,&sga,0);
    ret += sigaction(SIGPIPE,&sga,0);
    ret += sigaction(SIGQUIT,&sga,0);
    ret += sigaction(SIGTERM,&sga,0);
    ret += sigaction(SIGUSR1,&sga,0);
    ret += sigaction(SIGUSR2,&sga,0);
    ret += sigaction(SIGTSTP,&sga,0);
    ret += sigaction(SIGPOLL,&sga,0);
    ret += sigaction(SIGPROF,&sga,0);
    ret += sigaction(SIGURG,&sga,0);
    ret += sigaction(SIGVTALRM,&sga,0);
    ret += sigaction(SIGCHLD,&sga,0);

    if( ret < 0 )
      printmessage(stderr,CASE_ERROR,32,timestamp,"Sigaction: %s\n", strerror( errno ) ) ;

    return( ret );
}


int bbftp_setfathersignals() {
    struct sigaction sga;
    int ret;
    
    ret = resetallsignals( SIG_DFL );

    sga.sa_handler= bbftp_sigintfather;
    sigemptyset(&(sga.sa_mask));
    sga.sa_flags = 0;

    ret += sigaction(SIGINT,&sga,0);
    ret += sigaction(SIGTERM,&sga,0);

    sga.sa_handler = SIG_IGN;
    sigemptyset(&(sga.sa_mask));
    sga.sa_flags = 0;

    ret += sigaction(SIGPIPE,&sga,0);
    
    if( ret < 0 )
      printmessage(stderr,CASE_ERROR,32,timestamp,"Sigaction: %s\n", strerror( errno ) ) ;

    return( ret );
}


int bbftp_setsonsignals() {
  struct sigaction sga;
  int ret;

  ret = resetallsignals( SIG_IGN );	/* SIGPIPE too */

  sga.sa_handler= bbftp_sigintson;
  sigemptyset(&(sga.sa_mask));
  sga.sa_flags = 0;

  ret += sigaction(SIGINT,&sga,0);
  ret += sigaction(SIGTERM,&sga,0);

  sga.sa_handler = SIG_DFL;
  sigemptyset(&(sga.sa_mask));
  sga.sa_flags = 0;

  ret += sigaction(SIGCHLD,&sga,0);

  if( ret < 0 )
    printmessage(stderr,CASE_ERROR,32,timestamp,"Sigaction: %s\n", strerror( errno ) );

  return( ret );
}

int bbftp_setchildsignals() {
  struct sigaction sga;
  int ret;
	      
  ret = resetallsignals( SIG_IGN );     /* SIGPIPE too */
	        
  sga.sa_handler= bbftp_sigintchild;
  sigemptyset(&(sga.sa_mask));
  sga.sa_flags = 0;
		      
  ret += sigaction(SIGINT,&sga,0);
  ret += sigaction(SIGTERM,&sga,0);
			  
  if( ret < 0 )
    printmessage(stderr,CASE_ERROR,32,timestamp,"Sigaction: %s\n", strerror( errno ) );
			    
  return( ret );
}
