/*
 * bbftpc/treatcommand.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 

#include "bbftp.h"
#include <sys/types.h>

#include <errno.h>
#include <netinet/in.h>
#include <stdio.h>

#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "client.h"
#include "structures.h"
#include "client_proto.h"
#include "common.h"


extern  int     timestamp;
extern  int     globaltrymax;
extern  int     verbose;
extern  int     transferoption;
extern  int     sendwinsize;
extern  int     recvwinsize;
extern  int     localumask;
extern  int     remoteumask;
extern  int     remotecos;
extern	int		ackto;
extern	int		recvcontrolto;
extern	int		sendcontrolto;
extern	int		datato;
extern  char    *curfilename;
extern  char    *realfilename;
extern  int     nbport;
extern  int     buffersizeperstream;
extern	int	warning;
extern	int	statoutput;

extern xbuffer logmessage;
extern int errcode;

extern int maxfathers;


typedef int (*command)();

#define	COMMAND_LIST		\
	T(setbuffersize)	\
	T(setlocalumask)	\
	T(setnbstream)		\
	T(setoption)		\
	T(setremotecos)		\
	T(setremoteumask)	\
	T(setrecvwinsize)	\
	T(setsendwinsize)	\
	T(setrecvcontrolto)	\
	T(setsendcontrolto)	\
	T(setackto)		\
	T(setdatato)		\
	T(cd)			\
	T(get)			\
	T(lcd)			\
	T(mget)			\
	T(mkdir)		\
	T(mklnk)		\
	T(mput)			\
	T(put)			\
	T(delete)		\
	T(mdelete)		\
	T(rmdir)		\
	T(pipe)			\
	T(mlist)		\
	T(list)			\
	T(setmaxtransfers)


#define	T(x)	#x,
static const char * const COMMAND_NAMES[] = {
        "!",
        COMMAND_LIST
};
#undef  T
	
#define	T(x)	cds_##x,
enum {
	cds_comment,
	COMMAND_LIST
};
#undef	T

#define	T(x)	static int f_cds_##x();
static int f_cds_comment();
COMMAND_LIST
#undef	T

#define T(x)    f_cds_##x,
static command const commands[] = {
	f_cds_comment,
	COMMAND_LIST
};
#undef  T


static int choose( int size, const char * const names[], char *w ) {
  int i;
  int b = -1;
 
  for( i = 0; i < size; ++i )
    if( strstr( names[i], w ) == names[i] ) {
      if( b != -1 ) {
        b = -2;
        break;
      }
      b = i;
    }
 
  return b;
}


static int type;
static char *arg1;
static char *arg2;
static char *arg3;

static int shandle;
static int srhandle;

static int pipecnt = 1;

static const char * const NARG  = "number of arguments";
static const char * const NNARG = "requires a nonnegative argument";
static const char * const OARG  = "requires an octal argument";
static const char * const PARG  = "requires a positive argument";
static const char * const RARG  = "requires a positive argument or -1";
static const char * const DARG  = "requires a decimal argument";
static const char * const UARG  = "unknown option";
static const char * const AARG  = "ambiguous option";
static const char * const IARG  = " ";

#define	ARGS(x,y)							      \
  {									      \
    int dlt;								      \
									      \
    dlt = (arg1 ? (arg2 ? (arg3 ? ((x) - 3) : ((x) - 2)) : ((x) - 1)) : (x)); \
    if( dlt < 0 || dlt > (y) )						      \
      return( args( NARG ) );						      \
  }

#define IGNORED { return( ignored() ); }


static char *thecommand() {
  static char *rcd = NULL;
  int len;

  len = strlen( COMMAND_NAMES[type] ) + 1;
  if( arg1 ) {
    len += 1 + strlen( arg1 );
    if( arg2 ) {
      len += 1 + strlen( arg2 );
      if( arg3 )
	len += 1 + strlen( arg3 );
    }
  }
  rcd = (char *)(realloc( rcd, len ));
  if( rcd )
    sprintf( rcd, "%s%s%s%s%s%s%s", COMMAND_NAMES[type], (arg1 ? " " : ""), (arg1 ? arg1 : ""), (arg2 ? " " : ""), (arg2 ? arg2 : ""), (arg3 ? " " : ""), (arg3 ? arg3 : "") );

  return( rcd );
}

static int args( const char * const theerror ) {
  errcode = 26;
  printmessage( stderr, CASE_ERROR, 26, timestamp, "Incorrect(%s) command: %s\n", theerror, thecommand() );
  return( BB_RET_FT_NR );
}

static int ignored() {
  if( verbose )
    printmessage( stderr, CASE_NORMAL, 0, timestamp, "Command: %s safely ignored\n", COMMAND_NAMES[type] );
    
  return( BB_RET_OK );
}

static void printit( char *state ) {
  char *thecds;

  thecds = thecommand();
  
  if( verbose )
    printmessage( stderr, CASE_NORMAL, 0, timestamp, ">> COMMAND : %s <<%s", thecds, state );
  else if( 0 == statoutput )
    printmessage( stderr, CASE_NORMAL, 0, timestamp, "%s%s", thecds, state );

  if( srhandle != -1 ) {
    WRITE( srhandle, thecds, strlen( thecds ) );
    WRITE( srhandle, state, strlen( state ) );
  }
}

static int fivestar() {
  static int connectionisbroken = 0;
  int i;

 
  if( 1 == connectionisbroken ) {
    reconnecttoserver();
    connectionisbroken = 0;
  }
  
  for( i = 0; i < globaltrymax; ++i ) {
    int ret;
	             
    ret = commands[type]();         /* Goodies here ! */

    switch( ret ) {
      broken:
      case BB_RET_FT_NR_CONN_BROKEN:
        connectionisbroken = 1;
        /* fall through */
      fail:
      case BB_RET_FT_NR:
	printit( " FAILED\n" );
        return( ret );
      case BB_RET_OK:
	printit( " OK\n" );
        return( ret );
      case BB_RET_CONN_BROKEN:
        if( i < globaltrymax - 1 ) {
          sleep( WAITRETRYTIME );
          reconnecttoserver();
          continue;
        }
        goto broken;
      default:
        if( i < globaltrymax - 1 ) {
          if( warning ) printmessage( stderr, CASE_NORMAL, 0, timestamp, "Retrying command waiting %d s\n", WAITRETRYTIME );
          sleep( WAITRETRYTIME );
          continue;
        }
        goto fail;
    }
  }
}

static int f_cds_comment() {
  return( BB_RET_OK );
}

static int f_cds_cd() {
  ARGS( 1, 0 )
	  
  return( bbftp_cd( arg1 ) );  
}

static int f_cds_get() {
  int len;
  char *sl;
  
  ARGS( 2, 1 )

  len = (NULL == arg2 ? 0 : strlen( arg2 ));
  if( (0 == len || '/' == arg2[len - 1]) && (0 == (transferoption & TROPT_UNIX)) ) {
    sl = strrchr( arg1, '/' );
    sl = (sl ? sl + 1 : arg1);
    len += strlen( sl );
  }
  else
    sl = "";
  
  ++len;
  curfilename = (char *)(malloc( len ));
  if( transferoption & TROPT_TMP )
    len += strlen( ".bbftp.tmp" );
  realfilename = (char *)(malloc( len ));
  if( NULL == curfilename || NULL == realfilename ) {
    printmessage( stderr, CASE_ERROR, 35, timestamp, "Malloc: %s\n", strerror( ENOMEM ) );
    FREE( curfilename )
    FREE( realfilename )
    return( BB_RET_FT_NR );
  }
  sprintf( curfilename, "%s%s", (arg2 ? arg2 : ""), sl );
  sprintf( realfilename, "%s%s", curfilename, (transferoption & TROPT_TMP) ? ".bbftp.tmp" : "" );

  return( bbftp_get( arg1 ) ); 
}

static int f_cds_lcd() {
  ARGS( 1, 0 )
	  
  return( bbftp_lcd( arg1 ) );
}


static char *filelist;
static int filelistlen;

static xbuffer prefixdir = { NULL, 0, 0 };
static xbuffer path = { NULL, 0, 0 };

static int ctype; 

static int mrecs( char *pattern ) {
  int ret;

  if( cds_mput == ctype ) {
    ret = bbftp_retrlistdir( ptrxb( &prefixdir ), pattern, &filelist, &filelistlen );
    if( ret != 0 ) {
      printmessage( stderr, CASE_ERROR, errcode, timestamp, "%s\n", ptrxb( &logmessage ) );
      return( BB_RET_FT_NR );
    }
  }
  else {
    type = cds_list;
    arg1 = ptrxb( &prefixdir );
    arg2 = pattern;

    ret = fivestar();	/*bbftp_list()*/

    if( ret != BB_RET_OK )
      return( ret );
  }

  ret = 0;
  if( filelistlen != 0 ) {
    int hits = 0;
    char * const endlst = filelist + filelistlen;
    register char *rfile;
    register char *ratrb;
    xbuffer crtfile = { NULL, 0, 0 };
    xbuffer newpath = { NULL, 0, 0 };
    xbuffer dirs = { NULL, 0, 0 };

    if( NULL == appendxb( &crtfile, ptrxb( &prefixdir ), lenxb( &prefixdir ) - 1 ) ) {
      mal:
      errcode = 23;
      printmessage( stderr, CASE_ERROR, errcode, timestamp, "Malloc: %s\n", strerror( ENOMEM ) );

      freexb( &crtfile );
      freexb( &newpath );
      
      FREE( filelist )

      return( BB_RET_FT_NR );
    }

    for( rfile = filelist; rfile < endlst; rfile = ratrb + strlen( ratrb ) + 1 ) {
      int len;

      len = strlen( rfile );

      ratrb = rfile + len + 1;

      if( 0 == memcmp( rfile, ".", 2 ) || 0 == memcmp( rfile, "..", 3 ) )
        continue;

      shrinkxb( &crtfile, lenxb( &prefixdir ) - 1 );
      if( NULL == appendxb( &crtfile, rfile, len + 1 ) )
	goto mal;

      if( 'l' == *ratrb ) {

	if( cds_mdelete == ctype ) {
	  delit:
	  type = cds_delete;
	  arg1 = ptrxb( &crtfile );
	  arg2 = NULL;
          if( fivestar() != BB_RET_OK )
            ++ret;
	  continue;
	}
	      
	if( ptrxb( &newpath ) )
	  shrinkxb( &newpath, lenxb( &path ) - 1 );
	else if( NULL == appendxb( &newpath, ptrxb( &path ), lenxb( &path ) - 1 ) )
	  goto mal;
	if( NULL == appendxb( &newpath, rfile, len + 1 ) )
	  goto mal;
	      
	++hits;
	if( cds_mlist == ctype ) {
		printf( "%s\n", ptrxb( &newpath ) );
	}
	else if( cds_mget == ctype ) {
	  if( bbftp_storemklnk( ratrb + 2, ptrxb( &newpath ) ) != 0 )
	    ++ret;
	}
	else {
	  type = cds_mklnk;
	  arg1 = ratrb + 2;
	  arg2 = ptrxb( &newpath );
	  if( fivestar() != BB_RET_OK )	/* bbftp_mklnk() */
	    ++ret;
	}
	continue;
      }
      else {
        char kind;

        kind = *(ratrb + 1);
      
        switch( kind ) {
          case 'd':
	    if( 0 == (transferoption & TROPT_RECS) )
	      continue;
  
	    if( NULL == appendxb( &dirs, rfile, len + 1 ) )
	      goto mal;
	    
	    if( cds_mlist == ctype ) {
		    printf( "%s\n", ptrxb( &crtfile ) );
	    }

	    continue;
          case 'u':
            continue;	/*do nothing*/
          case 'f':
          default:
	    if( cds_mdelete == ctype )
	      goto delit;
	    ++hits;
	    if( cds_mlist == ctype ) {
		    printf( "%s\n", ptrxb( &crtfile ) );
	    }
	    else {
		    type = (cds_mput == ctype ? cds_put : cds_get);
		    arg1 = ptrxb( &crtfile );
		    arg2 = ptrxb( &path );
		    if( fivestar() != BB_RET_OK )
		      	    ++ret;
	    }
	    continue;
        }
      }
    }

    freexb( &crtfile );
    freexb( &newpath );

    FREE( filelist )

    if( lenxb( &dirs ) ) {
      char * const endlst = ptrxb( &dirs ) + lenxb( &dirs );
      register char *crt;
      int len;

      for( crt = ptrxb( &dirs ); crt < endlst; crt += len + 1 ) {
	len = strlen( crt );
	
	if( NULL == chainxb( &prefixdir, crt, len ) ||
	    NULL == appendxb( &prefixdir, "/", 2 ) ||
	    ( cds_mdelete != ctype && 
	      ( NULL == chainxb( &path, crt, len ) ||
	        NULL == appendxb( &path, "/", 2 )
	      )
	    )
	  ) {
	  errcode = 23;
	  printmessage( stderr, CASE_ERROR, errcode, timestamp, "Malloc: %s\n", strerror( ENOMEM ) );
		      
	  freexb( &dirs );
			            
	  return( BB_RET_FT_NR );
	}

	/*XXX Recursive */
	if( mrecs( "*" ) != BB_RET_OK )
	  ++ret;

	if( cds_mdelete == ctype ) {
	  chainxb( &prefixdir, "", 1 );
	  type = cds_rmdir;
	  arg1 = ptrxb( &prefixdir );
	  arg2 = NULL;
	  if( fivestar() != BB_RET_OK )
            ++ret;
	}

	shrinkxb( &prefixdir, lenxb(  &prefixdir ) - (len + 1) );
	shrinkxb( &path, lenxb( &path ) - (len + 1) );
      }

      freexb( &dirs );
    }
    else if( cds_mdelete != ctype && 0 == hits && (transferoption & TROPT_RECS) ) {
      switch( ctype ) {
	case cds_mget:
	  if( bbftp_storemkdir( ptrxb( &path ) ) != 0 )
	    ++ret;
	  break;
	case cds_mput:
	  type = cds_mkdir;
	  arg1 = ptrxb( &path );
	  arg2 = NULL;
	  if( fivestar() != BB_RET_OK )
	    ++ret;
          break;
	default:
	  break; /*do nothing*/
      }
    }
  }
  
  return( ret ? BB_RET_FT_NR : BB_RET_OK );
}

static int mgetputlist() {
  char *carg1;
  char *carg2;
  char *pattern;
  int ret;
  
  freexb( &prefixdir );
  freexb( &path );

  if( arg2 ) {
    ret = strlen( arg2 );
    if( NULL == appendxb( &path, arg2, ret + 1 ) )
      goto mal;
    if( arg2[ret -1] != '/' ) {
      if( NULL == chainxb( &path, "/", 2 ) )
        goto mal;
    }
  }
  else if( cds_mdelete != type ) if( NULL == appendxb( &path, "./", 3 ) )
    goto mal;

  pattern = strrchr( arg1, '/' );
  if( NULL == pattern ) {
    pattern = arg1;
    if( NULL == appendxb( &prefixdir, "./", 3 ) ) {
      mal:
      errcode = 23;
      printmessage( stderr, CASE_ERROR, errcode, timestamp, "Malloc: %s\n", strerror( ENOMEM ) );
      return( BB_RET_FT_NR );
    }
  }
  else {
    char ch;

    ++pattern;
    ch = *pattern;
    *pattern = 0;
    appendxb( &prefixdir, arg1, pattern - arg1 + 1 );
    *pattern = ch;
    if( NULL == ptrxb( &prefixdir ) )
      goto mal;
    if( 0 == ch )
      pattern = "*";
  }

  ctype = type;
  carg1 = arg1;
  carg2 = arg2;

  ret = mrecs( pattern );	/* Recursion here */

  type = ctype;
  arg1 = carg1;
  arg2 = carg2;

  return( ret );
}

static int f_cds_mlist() {
  ARGS( 2, 1 )

  return( mgetputlist() ); 
}

static int f_cds_mget() {
  ARGS( 2, 1 )

  return( mgetputlist() ); 
}

static int f_cds_mkdir() {
  ARGS( 1, 0 )
		  
  return( bbftp_mkdir( arg1 ) );
}

static int f_cds_mput() {
  ARGS( 2, 1 )

  return( mgetputlist() );
}

static int f_cds_mdelete() {
  ARGS( 1, 0 )

  return( mgetputlist() );
}
	  
static int f_cds_put() {
  int len;
  char *sl;
  char *rfile;
  
  ARGS( 2, 1 )

  len = (NULL == arg2 ? 0 : strlen( arg2 ));
  if( (0 == len || '/' == arg2[len - 1]) && (0 == (transferoption & TROPT_UNIX)) ) {
    sl = strrchr( arg1, '/' );
    sl = (sl ? sl + 1 : arg1);
    len += strlen( sl );
  }
  else
    sl = "";
      
  ++len;
  rfile = (char *)(malloc( len ));
  curfilename = strdup( arg1 );
  if( NULL == rfile || NULL == curfilename ) {
    printmessage( stderr, CASE_ERROR, 35, timestamp, "Malloc: %s\n", strerror( ENOMEM ) );
    FREE( rfile )
    FREE( curfilename )
    return( BB_RET_FT_NR );
  }
  sprintf( rfile, "%s%s", (arg2 ? arg2 : ""), sl );

  len = bbftp_put( rfile );
  FREE( rfile )
  return( len );
}

static int f_cds_setbuffersize() {
  int k;

  ARGS( 1, 0 )

  if( sscanf( arg1, "%d", &k ) != 1 || k <= 0 || k > 2097151 /* 2,147,483,647 / 1024 */ )
    return( args( PARG ) );
	  
  buffersizeperstream = k;
  return( BB_RET_OK );
}

static int f_cds_setmaxtransfers() {
  int k;

  ARGS( 1, 0 )

  if( sscanf( arg1, "%d", &k ) != 1 || k < -1 || 0 == k )
    return( args( RARG ) );
	  
  maxfathers = k;
  return( BB_RET_OK );
}

static int f_cds_setlocalumask() {
  int k;
  
  ARGS( 1, 0 )

  if( sscanf( arg1, "%o", &k ) != 1 || k < 0 )
    return( args( OARG ) );
 
  umask( k );
  localumask = k;
  return( BB_RET_OK );
}

static int f_cds_setnbstream() {
  int k;
	    
  ARGS( 1, 0 )
		      
  if( sscanf( arg1, "%d", &k ) != 1 || k <= 0 )
    return( args( PARG ) );
	     
  nbport = k; 
  return( BB_RET_OK );
}


#define	OPTION_LIST	\
	T(createdir)	\
	T(tmpfile)	\
	T(remoterfio)	\
	T(localrfio)	\
	T(keepmode)	\
	T(keepaccess)	\
	T(gzip)		\
	T(qbss)		\
	T(encrypt)	\
	T(recursive)	\
	T(exclude)	\
	T(unix)		\
	T(mpath)

#define	T(x)	#x,
static const char * const OPTION_NAMES[] = {
	OPTION_LIST
};
#undef	T

#define T(x)	opt_##x,
enum {
	OPTION_LIST
};
#undef  T
                                                            
#define	SETOPT(x)	(ohno ? (transferoption &= ~(x)) : (transferoption |= (x)));

static int f_cds_setoption() {
  char *opt;
  int ohno;
  int otype;
  
  ARGS( 1, 0 )

  if( 0 == strncmp( arg1, "no", 2 ) ) {
    ohno = 1;
    opt = arg1 + 2;
  }
  else {
    ohno = 0;
    opt = arg1;
  }

  otype = choose(  sizeof( OPTION_NAMES ) / sizeof( char * ), OPTION_NAMES, opt );

  switch( otype ) {
    case -1:
      return( args( UARG ) );
    case -2:
      return( args( AARG ) );
    case opt_createdir:
      SETOPT(TROPT_DIR)
      break;
    case opt_tmpfile:
      SETOPT(TROPT_TMP)
      break;
    case opt_remoterfio:
      SETOPT(TROPT_RFIO)
      break;
    case opt_localrfio:
      break;				/* backward compatibility */
    case opt_keepmode:
      SETOPT(TROPT_MODE)
      break;
    case opt_keepaccess:
      SETOPT(TROPT_ACC)
      break;
    case opt_gzip:
#ifndef	WITH_GZIP
      /* may work anyway in unix style ! */
#endif	/* WITH_GZIP */
      SETOPT(TROPT_GZIP)
      break;
    case opt_qbss:
      SETOPT(TROPT_QBSS)
      break;
    case opt_encrypt:
#ifndef	WITH_SSL
      if( 0 == ohno ) {
	printmessage( stderr, CASE_ERROR, 7, timestamp, "%s option is unavailable: bbftp was built without encryption\n", OPTION_NAMES[opt_encrypt] );
	return( BB_RET_FT_NR );
      }
#endif	/* WITH_SSL */
      SETOPT(TROPT_HIDDEN)
      break;
    case opt_recursive:
      SETOPT(TROPT_RECS)
      if( 0 == ohno ) SETOPT(TROPT_DIR)
      break;
    case opt_exclude:
      SETOPT(TROPT_EXCL)
      break;
    case opt_unix:
      SETOPT(TROPT_UNIX)
      break;
    case opt_mpath:
      SETOPT(TROPT_MPATH)
      break;
    default:
      return( BB_RET_FT_NR );		/* impossible */
  }

  return( BB_RET_OK );
}

static int f_cds_setremotecos() {
  int k;
	
  ARGS( 1, 0 )

  if( sscanf( arg1, "%d", &k ) != 1 )
    return( args( DARG ) );		                                                                                  
  return( bbftp_setremotecos( k ) );
}

static int f_cds_setremoteumask() {
  int k;
  int ret;
	    
  ARGS( 1, 0 )
		      
  if( sscanf( arg1, "%o", &k ) != 1 || k < 0 )
    return( args( OARG ) );                                                                                 
  ret = bbftp_setremoteumask( k );
  if( BB_RET_OK == ret ) 
    remoteumask = k;
  return( ret );
}

static int f_cds_setrecvwinsize() {
  int k;
	    
  ARGS( 1, 0 )
		      
  if( sscanf( arg1, "%d", &k ) != 1 || k <= 0 )
    return( args( PARG ) );
	      
  recvwinsize = k;
  return( BB_RET_OK );
}

static int f_cds_setsendwinsize() {
  int k;
	    
  ARGS( 1, 0 )
		      
  if( sscanf( arg1, "%d", &k ) != 1 || k <= 0 )
    return( args( PARG ) );
	      
  sendwinsize = k;
  return( BB_RET_OK );
}

static int f_cds_setrecvcontrolto() {
  int k;
  
  ARGS( 1, 0 )

  if( sscanf( arg1, "%d", &k ) != 1 || k < 0 )
    return( args( NNARG ) );
  
  recvcontrolto = k;
  return( BB_RET_OK );
}

static int f_cds_setsendcontrolto() {
  int k;
   
  ARGS( 1, 0 )

  if( sscanf( arg1, "%d", &k ) != 1 || k < 0 )
    return( args( NNARG ) );
    
  sendcontrolto = k;  
  return( BB_RET_OK );
}

static int f_cds_setackto() {
  IGNORED
}

static int f_cds_setdatato() {
  IGNORED
}

static int f_cds_list() {

  return( bbftp_list( arg1, arg2, &filelist, &filelistlen ) );
}

static int f_cds_mklnk() {
  ARGS( 2, 1 )
	
  return( bbftp_mklnk( arg1, arg2 ) );
}


static int f_cds_delete() {
  ARGS( 1, 0 )

  return( bbftp_delete( arg1 ) );
}

static int f_cds_rmdir() {
  ARGS( 1, 0 )

  return( bbftp_rmdir( arg1 ) );
}


extern char *incoming;
extern char *outgoing;

static int f_cds_pipe() {
  ARGS( 1, 0 )

  do {
    fSTAT st;
    int crtopt;
    int ret;
    
    pipecnt ^= 1;	  
    if( FSTAT( pipecnt, &st ) < 0 ) {
      printmessage( stderr, CASE_ERROR, 72, timestamp, "Fstat %d : %s\n", pipecnt, strerror( ERRNO ) );
      return( BB_RET_FT_NR );
    }

    if( 0 == S_ISFIFO( st.st_mode ) ) {
	   if( 0 == pipecnt ) {	
		    if( NULL == incoming )
			   continue;
	   }
   	   else {
		   if( NULL == outgoing )
			   continue;
	   }

	   /*spawn command*/
	   {
		   pid_t pid;
		   int p[2];

		   if( -1 == pipe( p ) ) {
			   printmessage( stderr, CASE_ERROR, 29, timestamp, "Pipe : %s\n", strerror( ERRNO ) );
			   return( BB_RET_FT_NR );
		   }

		   pid = fork();
		   switch( (int)(pid) ) {
			   char *argv[4];

			   case 0:
			   	argv[0] = "sh";
				argv[1] = "-c";
				argv[2] = (0 == pipecnt ? incoming : outgoing);
				argv[3] = NULL;

				close( p[pipecnt] );
				if( (pipecnt ^ 1) != p[pipecnt ^ 1] ) {
					dup2( p[pipecnt ^ 1], pipecnt ^ 1 );
					close( p[pipecnt ^ 1] );
				}

				execvp( argv[0], argv );

				/*Never get here*/
				printmessage( stderr, CASE_ERROR, 29, timestamp, "Execvp : %s\n", strerror( ERRNO ) );
			      	exit( errno );
		
			   case -1:
				printmessage( stderr, CASE_ERROR, 29, timestamp, "Fork : %s\n", strerror( ERRNO ) );
			       	return( BB_RET_FT_NR );

			   default:
				close( p[pipecnt ^ 1] );
				if( (pipecnt) != p[pipecnt] ) {
					dup2( p[pipecnt], pipecnt );
					close( p[pipecnt] );
				}
				break;
		   }
	   }
    }

    crtopt = transferoption;
    transferoption |= TROPT_UNIX | TROPT_PIPE;

    ret = ( 0 == pipecnt ? bbftp_put( arg1 ) : bbftp_get( arg1 ) );
    
    transferoption = crtopt;

    return( ret );
    
  }while( 0 == pipecnt );

  printmessage( stderr, CASE_ERROR, 29, timestamp, "No pipes !?\n" );
  return( BB_RET_FT_NR );
}

static void treatcommand( char *buffercmd ) {
  char *cds;
  int len;

  len = strlen( buffercmd );
  cds = strtok( buffercmd, " \t" );
  
  if( cds ) {
    
    type = choose( sizeof( COMMAND_NAMES ) / sizeof( char * ), COMMAND_NAMES, cds );

    switch( type ) {
      case -1:
        if( '!' == cds[0] ) {
	  type = cds_comment;
	  goto dflt;
        }
        printmessage( stderr, CASE_ERROR, 25, timestamp, "Unknown command: %s\n", cds );
        break;
      case -2:
        printmessage( stderr, CASE_ERROR, 25, timestamp, "Ambiguous command: %s\n", cds );
        break;
      dflt:
      default:
        if( shandle != -1 && -1 == srhandle ) {
	  if( type > cds_setdatato ) {
	    if( warning )
	      printmessage( stderr, CASE_WARNING, 13, timestamp, "Prohibited command: %s\n", COMMAND_NAMES[type] );
	    break;
	  }
        }
     
	if( cds_list == type )
		type = cds_mlist;
 
	arg1 = cds + strlen( cds ) + 1;
	while( *arg1 == ' ' || *arg1 == '\t' ) ++arg1; 
	if( type != cds_pipe ) {
		for( cds = arg1;; ) {
			cds = strpbrk( cds, "\\ \t" );
			if( NULL == cds ) {
				arg2 = NULL;
				break;
			}

			if( '\\' == *cds ) {
				memmove( cds, cds + 1, strlen( cds ) );
				++cds;
			}	
			else {
				*cds = 0;
				arg2 = cds + 1;
				while( *arg2 == ' ' || *arg2 == '\t' ) ++arg2;
				if( 0 == *arg2 ) arg2 = NULL;
				break;
			}
	
		}
	
		if( arg2 ) for( cds = arg2;; ) {
			cds = strpbrk( cds, "\\ \t" );
			if( NULL == cds ) {
				arg3 = NULL;
				break;
			}
	
			if( '\\' == *cds ) {
				memmove( cds, cds + 1, strlen( cds ) );
				++cds;
			}	
			else {
				*cds = 0;
				arg3 = cds + 1;
				while( *arg3 == ' ' || *arg3 == '\t' ) ++arg3; 
				if( 0 == *arg3 ) arg3 = NULL;
				break;
			}
		}
		else
			arg3 = NULL;
	}
	else {
		arg2 = arg3 = NULL;
      		if( arg1 ) {
	    		int larg;
	    
			larg = strlen( arg1 );
	    		if( arg1 + larg < buffercmd + len )
			      	*(arg1 + larg) = ' ';   
      		}
	}

        fivestar();

        break;   
    }
  }
}

static char *mygets( int handle ) {
  static xbuffer line = { NULL, 0, 0 };
  static char buffer[512];
  static int csize = 0;
  static int bsize = 0;

  emptyxb( &line );
  for(;;) { 
    int w1;
    char *b1;
    char *b2;
    
    if( csize >= bsize ) {
      csize = 0;
      bsize = 0;
      for(;;) {
        bsize = READ( handle, buffer, sizeof( buffer ) );
        if( -1 == bsize ) {
	  if( EINTR == ERRNO )
	    continue;
	  return( NULL );
	}
	break;
      }
    }

    if( 0 == bsize )
      return( appendxb( &line, "\0\0", 2 ) );
                  
    b2 = buffer + csize;
    w1 = bsize - csize;
    b1 = memchr( b2, '\n', w1 );
    if( b1 )
      w1 = b1 - b2;
    csize += w1 + 1;
    
    if( NULL == appendxb( &line, b2, w1 ) )
      return( NULL );
			              
    if( b1 )
      return( appendxb( &line, "\0\1", 2 ) );
  }
}

void interpreter( int handle, int rhandle, char *bbftpcmd ) {
  char *buffercmd;
 
  shandle = handle;
  srhandle = rhandle;
  
  for(;;) {
    if( handle != -1 ) {
      bbftpcmd = mygets( handle );
      if( NULL == bbftpcmd )
        printmessage( stderr, CASE_FATAL_ERROR, 24, timestamp, "READ: %s\n", strerror( errno ) );
      if( 0 == bbftpcmd[0] ) {
	if( 0 == bbftpcmd[1] )
          break;
	else
	  continue;
      }
    }
	      
    /*
    ** command may be under the form cmd ; cmd ; cmd....
    ** ; may be escaped with a backslash \;
    */
    do {
		    
      buffercmd = bbftpcmd;
      for(;;) {
        bbftpcmd = strchr( bbftpcmd, ';' );
	if ( NULL == bbftpcmd )
	  break;
	if( bbftpcmd == buffercmd || '\\' != *(bbftpcmd - 1) ) {
	  *bbftpcmd = 0;
	  ++bbftpcmd;
	  break;
	}
	memmove( bbftpcmd - 1, bbftpcmd, strlen( bbftpcmd ) + 1 );
      }
	      
      if( 0 == *buffercmd )
        continue;
	     
      treatcommand( buffercmd );
	      
    }while( bbftpcmd );

    if( -1 == handle )
      break;
  }
}
