/*
 * bbftpd/bbftpd.c
 * Copyright (C) 1999, 2000, 2001, 2002 IN2P3, CNRS
 * bbftp@in2p3.fr
 * http://doc.in2p3.fr/bbftp
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */ 

/****************************************************************************

  
  
 bbftpd.c v 0.0.0  1999/11/24
          v 1.3.0  2000/03/16   - Version in openlog
          v 1.4.0  2000/03/22   - Modify the connection protocole
                                  in order to send the crypt type
                                  after the connection.
          v 1.5.0  2000/03/24   - Change the wait timer in default
          v 1.6.0  2000/03/24   - Change the main loop to make it more
                                    modular
          v 1.6.1  2000/03/24   - Portage on OSF1
                                - Make the control socket non blocking
          v 1.8.0  2000/04/14   - Introduce RSA Cryptage
          v 1.8.6  2000/05/21   - Allow to run with inetd
          v 1.8.7  2000/05/24   - Introduce version.h and config.h
          v 1.8.10 2000/08/11   - Portage to Linux
          v 1.9.0  2000/08/18   - Use configure to help portage
                                - default time out set to 900 s
          v 1.9.3  2000/10/12   - Add -b and -w option in order to overwrite
                                  the fixed values
          v 1.9.4  2000/10/16   - Make all sockets blocking in order
                                  to prevent the error in case of lack 
                                  of memory
                                - Supress %m
          v 2.0.0  2001/03/26   - Protocol V2 implementation
          v 2.0.1  2001/04/23   - Correct indentation
                                - Port to IRIX
          v 2.0.2  2001/05/07   - Add debug option for RFIO
          v 2.1.0  2001/05/28   - Add private authentication
                                - Correct syslog level
                                - Add -l option
		  v 2.2.0  2001/10/03   - Add the certificate authentication process

*****************************************************************************/
#include "bbftpd.h"
#include <sys/types.h>

#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <syslog.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <unistd.h>

#include <setjmp.h>
#include <string.h>
#include <ctype.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <netdb.h>

#include "common.h"
#include "config.h"
#include "daemon.h"
#include "structures.h"
#include "daemon_proto.h"
#include "status.h"
#include "version.h"

#ifdef CERTIFICATE_AUTH
#include <gssapi.h>
#include <gfw.h>
#endif

#ifdef HAVE_BYTESWAP_H
#include <byteswap.h>
#endif

extern int getoptt (int argc, char *const *argv, const char *shortopts);

#ifdef CERTIFICATE_AUTH
#define OPTIONS    "bcd:e::fl:m:pR:suvw:j:n:"
#else
#define OPTIONS    "bd:e::fl:m:R:suvw:j:n:"
#endif
/*
** Common variables for BBFTP protocole version 1 and 2
*/
/*
** be_daemon :
**      That is the way the daemon is going to be run.
**      0 means run by inetd (that can be done trought 
**      a wrapper)                                      (run as bbftpd)
**      1 means that the control socket are stdin and 
**      stdout                                          (run as bbftpd -s)
**      2 means running in background                   (run as bbftpd -b)
**      Default 0
*/
int        be_daemon = 0 ;
/*
** fixeddataport :
**      If set to 1 then the server will use CONTROLPORT-1 
**      to use as local port number for the data socket. 
**      It is useful to set it to 1 if the server is 
**      running behing a firewall.                      (run as bbftpd)
**      If set to 0 then the server will not fixe the 
**      local data port                                 (run as bbftpd -f)
**      Default 1
*/
int     fixeddataport = 1 ;
/*
** state :
**      Define the state of the server. values differ depending
**      of the BBFTP protocol version
*/
int        state ;                         
/*
** newcontrolport :
**      Define the control port to listen on
*/
int        newcontrolport ; 
/*
** currentusername :
**      Define the local username
*/
char currentusername[MAXLEN] ;

int	crypton = 1;
RSA	*myprsa = NULL;
RSA	*hisrsa = NULL;
unsigned char ciphkey[16];
unsigned char ciphiv[8];

char *localips = "";

int natip = 0;

int ctrl_port;

/*
** bbftprc:
**      Where to store the bbftprc file
*/
char    *bbftprc = NULL ;

/*
** debug:
**      Set to one for RFIO debug. Valid only with the -b option
**      set to the value needed for RFIO_TRACE
*/
int     debug = 0 ;
/*
** End Common variables for BBFTP protocole version 1 and 2 ********************
*/
/*
** Variables for BBFTP protocole version 2
*/
/*
** incontrolsock :
**      Define the control socket for reading
** outcontrolsock :
**      Define the control socket for writing
**
**      For be_daemon equal to 0 or 2 incontrolsock = outcontrolsock
*/
int     incontrolsock ;
int     outcontrolsock ;
/*
** curfilename :
**      Define the pointer to the current file
*/
char    *curfilename = NULL ;
/*
** realfilename :
**      Define the pointer to the real file (= curfilename if TROPT_TMP not
**      set)
*/
char    *realfilename = NULL ;
/*
** myports :
**      Pointer to the first port
*/
unsigned short int     *myports = NULL ;
/*
** mysockets :
**      Pointer to the first socket
*/
int     *mysockets = NULL ;

/*
** myumask :
**      Umask for the bbftpd process, the default will be 022 which
**      means that group and other write permissions are denied
**      This can be changed by the bbftp umask command
**/
int     myumask = 022 ;
/*
**	Parameters describing the connection
*/
int	    recvcontrolto   = CONTROLSOCKTO;
int	    sendcontrolto   = SENDCONTROLTO;
int     force_encoding  = 1;
#ifdef CERTIFICATE_AUTH
int     accept_certs_only = 0;
int     accept_pass_only = 0;
#endif

struct sockaddr_in his_addr;

/*
** Parameters describing the transfer
*/
int     transferoption = 0;
int     filemode ;
char    lastaccess[9];
char    lastmodif[9];
int     sendwinsize ;        
int     recvwinsize ;        
int     buffersizeperstream ;
int     requestedstreamnumber ;
my64_t  filesize ;

jmp_buf env;

/*
** maxstreams :
**      That is to fix the maximum number of stream the deamon will accept.
**      (run as bbftpd -m 25)
**      Default 25
*/
int     maxstreams = MAXSTR;
/*
** End Variables for BBFTP protocole version 2 *********************************
*/
/*
 * Variables for protocol version 3 (PASV mode) ******************************
*/
/*
 * Range for the ephemeral ports for data connections
 */
int     pasvport_min = 0;
int     pasvport_max = 0;
/*
** End Variables for BBFTP protocole version 3 *********************************
*/
/*
**  credentials for certificate authentication
*/
#ifdef CERTIFICATE_AUTH
gss_cred_id_t   server_creds;
#endif

xbuffer logmessage = { NULL, 0, 0 };


int f_handler_login( char *buffer, int len, int code ) {
    int retval = 0;

    switch( code ) {
#ifdef	CERTIFICATE_AUTH
      case MSG_NKEY:
		retval = saversa( buffer );
		if( -1 == retval )
		  return( -1 );

		return( handlepackage( handler_login ) );	/* Expecting MSG_CERT_LOG */
      case MSG_CERT_LOG:
                if (accept_pass_only) {
                    printxb( &logmessage, "The server only accepts USER/PASS" );
                    syslog( BBFTPD_ERR, ptrxb( &logmessage ) );
                    reply( MSG_BAD_NO_RETRY, ptrxb( &logmessage ) );
                    return( -1 );
                }
                retval=bbftpd_cert_receive_connection();
                if ( retval < 0 ) {
                    /*
                    ** The login failed, do not wait for a new one
                    */
		    return( -1 );
                }
                /*
                ** If retval >0, means an MSG_WARN has been sent already
                */
		break;
#endif	/* CERTIFICATE_AUTH */
      case MSG_LOG:
      case MSG_PRIV_LOG:
#ifdef	CERTIFICATE_AUTH
                if (accept_certs_only) {
                    printxb( &logmessage, "The server only accepts certificates" );
                    syslog( BBFTPD_ERR, ptrxb( &logmessage ) );
                    reply( MSG_BAD_NO_RETRY, ptrxb( &logmessage ) );
		    return( -1 );
                }
#endif	/* CERTIFICATE_AUTH */
                /*
                ** It seems that it is the message we were waiting
                ** so lets decrypt 
                */
		retval = loginsequence( buffer );
                if( retval < 0 ) {
                    /*
                    ** The login failed, do not wait for a new one
                    */
                    return( -1 );
                }
                break;
      case MSG_CRYPT:
		if (force_encoding) {
                    /*
                    **  The client can't encode and ask if it can send uncrypted
                    **  login information
                    */
                    reply( MSG_BAD_NO_RETRY, printxb( &logmessage, "The server requires encrypted login" ) );
		    return( -1 );
                }
                reply( MSG_OK, printxb( &logmessage, "Uncrypted login dialog accepted" ) );
		
		return( handlepackage( handler_login ) );	/* Expecting MSG_LOG */
      case MSG_USER:
		retval = authsequence( buffer );
		if( retval < 0 )
		  return( -1 );
		break;
      default:
		if( VALID_MESSAGE( code ) )
		  printxb( &logmessage, "Unexpected message %s", MESSAGE_NAMES[code] );
		else
		  printxb( &logmessage, "Unlisted message %d", code );
		reply( MSG_BAD, ptrxb( &logmessage ) );
		return( RET_NONE );
    }
  
    if( 0 == retval ) 
      reply( MSG_OK, printxb( &logmessage, "bbftpdPRO version %s : OK", VERSION ) );
    return( 0 );
}


#define	COMMAND_LIST		\
	T(setackto)		\
	T(setrecvcontrolto)	\
	T(setsendcontrolto)	\
	T(setdatato)		\
	T(setcheckstdinto)

#define	T(x)	#x,
static const char * const COMMAND_NAMES[] = {
        "!",
        COMMAND_LIST
};
#undef	T
                                                                                                                            
#define	T(x)	cds_##x,
enum {
	cds_comment,
	COMMAND_LIST
};
#undef	T
                                                                                                                            

static int choose( int size, const char * const names[], char *w ) {
  int i;
  int b = -1;
 
  for( i = 0; i < size; ++i )
    if( strstr( names[i], w ) == names[i] ) {
      if( b != -1 ) {
        b = -2;
        break;
      }
      b = i;
    }
 
  return b;
}


static void treatcommand( char *buffercmd ) {
  char *cds;
  char *arg1;
  char *arg2;
  int type;

  cds = strtok( buffercmd, " \t" );
  
  if( cds ) {
    
    type = choose( sizeof( COMMAND_NAMES ) / sizeof( char * ), COMMAND_NAMES, cds );

    switch( type ) {
      case -1:
        if( '!' == cds[0] )
	  break;
        syslog( BBFTPD_ERR, "Unkown command: %s", cds );
        break;
      case -2:
        syslog( BBFTPD_ERR, "Ambiguous command: %s", cds );
        break;
      default:
	arg1 = strtok( NULL, " \t" );
	arg2 = strtok( NULL, " \t" );
	if( type > 0 && (NULL == arg1 || arg2) ) {
	  syslog( BBFTPD_ERR, "Incorrect(number of arguments) for %s", COMMAND_NAMES[type] );
	  break;
	}
	switch( type ) {
	  int k;
	  
	  case cds_comment:
	  case cds_setackto:
	  case cds_setdatato:
	  case cds_setcheckstdinto:
	    break;	/* backward compatibility */
	  case cds_setrecvcontrolto:
	  case cds_setsendcontrolto:
	    if( sscanf( arg1, "%d", &k ) != 1 || k < 0 ) {
	      syslog( BBFTPD_ERR, "Incorrect(negative argument) command: %s %s", COMMAND_NAMES[type], arg1 );
	      break;
	    }
	    if( cds_setrecvcontrolto == type )
	      recvcontrolto = k;
	    else
	      sendcontrolto = k;
	    break;
	  default:
	    break;	/* do nothing */
	}
        break;   
    }
  }
}

static char *mygets( int handle ) {
  static xbuffer line = { NULL, 0, 0 };
  static char buffer[512];
  static int csize = 0;
  static int bsize = 0;

  emptyxb( &line );
  for(;;) { 
    int w1;
    char *b1 = NULL;
    char *b2;
    
    if( csize >= bsize ) {
      csize = 0;
      bsize = 0;
      for(;;) {
        bsize = READ( handle, buffer, sizeof( buffer ) );
        if( -1 == bsize ) {
	  if( EINTR == ERRNO )
	    continue;
	  return( NULL );
	}
	break;
      }
    }
    b2 = buffer + csize;
    if( 0 == bsize ) {
      w1 = 1;
      b1 = b2;
    }
    else {
      w1 = bsize - csize;
      b1 = memchr( b2, '\n', w1 );
      if( b1 )
	w1 = b1 - b2 + 1;
    }

    if( NULL == appendxb( &line, b2, w1 ) )
      return( NULL );
    
    csize += w1;
    
    if( b1 )
      return( chainxb( &line, (bsize ? "\0\1" : "\0\0"), 2 ) );
  }
}

static void interpreter( int handle ) {
  char *buffercmd;
  char *bbftpcmd;
 
  for(;;) {
    bbftpcmd = mygets( handle );
    if( NULL == bbftpcmd ) {
      syslog( BBFTPD_ERR, "READ: %s", strerror( errno ) );
      break;
    }
    if( 0 == bbftpcmd[0] ) {
      if( 0 == bbftpcmd[1] )
        break;
      else
	continue;
    }
	      
    /*
    ** the command may be under the form cmd ; cmd ; cmd....
    */
    do {
		    
      buffercmd = bbftpcmd;
      bbftpcmd = strchr( bbftpcmd, ';' );
      if( bbftpcmd ) {
	*bbftpcmd = 0;
	++bbftpcmd;
      }
	      
      if( 0 == *buffercmd )
        continue;
	     
      treatcommand( buffercmd );
	      
    }while( bbftpcmd );
  }
}


int main (argc,argv,envp)
    int     argc ;
    char    **argv ;
    char    **envp ;
{
    extern char *optarg;
    extern int optind, opterr;

#if defined(SUNOS) || defined(_HPUX_SOURCE) || defined(IRIX) || defined(OSF1)
    int        addrlen ;
#else
    size_t     addrlen ;
#endif
    int        retcode ;
    int        i, j, k ;
    char    *bbftprcfile = NULL ;

    {
      static char ident[32];
      
      sprintf( ident, "bbftpPRO %s", VERSION );
      openlog( ident, LOG_PID | LOG_NDELAY, BBFTPD_FACILITY );
    }
    
    /*
    ** Set the log mask to BBFTPD_EMERG (0) 
    */
    setlogmask(LOG_UPTO(BBFTPD_EMERG));

    /*
    ** Initialize variables
    */
    newcontrolport = CONTROLPORT ;
    opterr = 0 ;
    while ((j = getoptt(argc, argv, OPTIONS)) != -1) {
        switch (j) {
            case 'v' :{
                printf("bbftpdPRO version %s\n",VERSION) ;
                printf("Compiled with  :   default port %d\n",CONTROLPORT) ;
                printf("                   default maximum number of streams = %d \n",maxstreams) ;

#ifdef PORT_RANGE
		printf("                   data ports range = %s \n", PORT_RANGE) ;
#endif

#ifdef WITH_GZIP
                printf("                   compression \n") ;
#endif
		printf("                   encryption \n");

#ifdef WITH_RFIO
                printf("                   RFIO interface \n") ;
#endif

#ifdef WITH_RFIO64
                printf("                   RFIO64 interface \n") ;
#endif

#ifdef CERTIFICATE_AUTH
                printf("                   certificate authentication \n") ;
#endif

#ifdef AFS
                printf("                   AFS authentication \n") ;
#endif

#if defined( HAVE_LSB2_SECURITY_PAM_APPL_H  ) || defined( HAVE_SECURITY_PAM_APPL_H )
		printf("                   PAM authentication \n") ;
#endif
		
#ifdef PRIVATE_AUTH
                printf("                   private authentication \n") ;
#endif	/* PRIVATE_AUTH */

                exit( errno = 0 );
            }
        }
    }
    /*
    ** Look at the loglevel option 
    */
    opterr = 0 ;
    optind = 1 ;
    while ((j = getoptt(argc, argv, OPTIONS)) != -1) {
        switch (j) {
            case 'l' :{
                for ( i=0 ; i< strlen(optarg) ; i++ ) {
                    optarg[i] = toupper(optarg[i]) ;
                }
                i = 0 ;
                if ( !strcmp(optarg,"EMERGENCY") ) {
                    i = BBFTPD_EMERG;
                } else if ( !strcmp(optarg,"ALERT") ) {
                    i = BBFTPD_ALERT;
                } else if ( !strcmp(optarg,"CRITICAL") ) {
                    i = BBFTPD_CRIT;
                } else if ( !strcmp(optarg,"ERROR") ) {
                    i = BBFTPD_ERR;
                } else if ( !strcmp(optarg,"WARNING") ) {
                    i = BBFTPD_WARNING;
                } else if ( !strcmp(optarg,"NOTICE") ) {
                    i = BBFTPD_NOTICE;
                } else if ( !strcmp(optarg,"INFORMATION") ) {
                    i = BBFTPD_INFO;
                } else if ( !strcmp(optarg,"DEBUG") ) {
                    i = BBFTPD_DEBUG;
                }
                if ( i > 0 ) {
                    setlogmask(LOG_UPTO(i));
                }
                break ;
            }
        }
    }
    syslog( BBFTPD_DEBUG, "Starting bbftpdPRO" );
    
    opterr = 0 ;
    optind = 1 ;
    while ((j = getoptt(argc, argv, OPTIONS)) != -1) {
        switch (j) {
            case 'b' :{
                if ( be_daemon != 0 ) {
                    syslog(BBFTPD_ERR,"-b and -s options are incompatible") ;
                    exit( errno = EPERM );
                }
                be_daemon = 2 ;
                break ;
            }
            case 'd' :{
                sscanf(optarg,"%d",&i) ;
                if ( i > 0 ) debug = i ;
                break ;
            }

	    case 'j':
		localips = optarg;
		break;

            case 'e' :{ if( optarg ) {
				if ((sscanf(optarg,"%d:%d",&i, &k) == 2) && (i <= k)) {
				  pasvport_min = i; pasvport_max = k;
				} else {
				  syslog(BBFTPD_ERR,"Invalid port range : %s",optarg) ;
				  fprintf(stderr,"Invalid port range : %s\n",optarg) ;
				  exit( errno = EPERM );
				}
			}
			else {
#ifdef PORT_RANGE
		          sscanf(PORT_RANGE,"%d:%d",&pasvport_min, &pasvport_max) ;
#endif
			  if( 0 == pasvport_max ) {	/*Data port equal to control port !!!*/
			    pasvport_min = 0;
			    pasvport_max = 1;
		          }		  
			}
                break ;
            }
            case 'f' :{
                fixeddataport = 0 ;
                break ;
            }
            case 'm' :{
                sscanf(optarg,"%d",&i) ;
                if ( i > 0 ) maxstreams = i;
                break ;
            }
            case 'R' :{
                bbftprcfile = optarg ;
                break ;
            }
            case 's' :{
                if ( be_daemon != 0 ) {
                    syslog(BBFTPD_ERR,"-b and -s options are incompatible") ;
                    exit( errno = EPERM );
                }

#ifdef PRIVATE_AUTH
                syslog(BBFTPD_ERR,"-s option cannot be used with private authentication") ;
                exit( errno = EPERM );                
#endif	/* PRIVATE_AUTH */
		
                be_daemon = 1 ;
                break ;
            }
            case 'u' :{
                force_encoding = 0 ;
                break ;
            }
            case 'w' :{
                sscanf(optarg,"%d",&newcontrolport) ;
                break ;
            }
#ifdef CERTIFICATE_AUTH
            case 'c' :{
                accept_certs_only = 1 ;
                break ;
            }
            case 'p' :{
                accept_pass_only = 1 ;
                break ;
            }
#endif
            default : {
                break ;
            }
        }
    }


    opterr = 0 ;
    optind = 1 ;
    while ((j = getoptt(argc, argv, OPTIONS)) != -1) {
        switch (j) {
            case 'n' :{
		natip = get_ip( optarg, NULL );
		break;
            }
	}
    }
    
    if( NULL == bbftprcfile ) {
      int id;

      id = getuid(); 
        /*
        ** look for the local user in order to find the .bbftprc file
	** use /etc/bbftpd.conf if root
        */
      if( 0 == id )
	  bbftprcfile = printxb( &logmessage, "/etc/bbftpd.conf" );
      else {
	struct passwd *mypasswd;
	  
	mypasswd = getpwuid( id );
	if( NULL == mypasswd )
	  syslog( BBFTPD_WARNING, "Unable to get passwd entry, .bbftprc will not be used" );
	else if( NULL == mypasswd->pw_dir )
	  syslog( BBFTPD_WARNING, "No home directory, .bbftprc will not be used" );
	else
	  bbftprcfile = printxb( &logmessage, "%s/.bbftprc", mypasswd->pw_dir );
      }
    }
    
    if( bbftprcfile ) {
      if( strncmp( bbftprcfile, "none", 4) != 0 ) {
        int rcfd;
	
	rcfd  = OPEN( bbftprcfile, O_RDONLY, 0666 );
	
	if( rcfd < 0 )
	  syslog( BBFTPD_WARNING, "Open(%s):  %s", bbftprcfile, strerror( errno ) );
        else {
          interpreter( rcfd );                                                                                                                             
          CLOSE( rcfd );
	  
        }
      }
    }

    /*
    ** Reset debug to zero if -b option is not present
    */
    if ( (be_daemon != 2) ) debug = 0 ;
    /*
    ** Check if be_daemon = 0 and in this case reset the 
    ** control port to CONTROLPORT
    */
    if ( be_daemon == 0 ) newcontrolport = CONTROLPORT ;

#ifdef CERTIFICATE_AUTH                
    if (be_daemon != 1 && !accept_pass_only) {
        OM_uint32 min_stat, maj_stat;
	maj_stat = gfw_acquire_cred(&min_stat, NULL, &server_creds);
        if (maj_stat != GSS_S_COMPLETE) {
            gfw_msgs_list *messages = NULL;
            gfw_status_to_strings(maj_stat, min_stat, &messages) ;
            while (messages != NULL) {
                syslog(BBFTPD_ERR,"gfw_acquire_cred: %s", messages->msg) ;
                if (be_daemon == 2) fprintf(stderr,"Acquire credentials: %s\n", messages->msg) ;
                messages = messages->next;
            }
            exit( errno = EPERM );
        }
    }
#endif

    stir();	/* OpenSSL */
   
    switch( be_daemon ) {
      case 0:
	incontrolsock = outcontrolsock = 0;
	break;
      case 2:
        /*
        ** Run as a daemon 
        */
        do_daemon( argc, argv, envp );

	setjmp( env );

	break;
      case 1:
        /*
        ** Run by a remote user
        */
        /*
        ** Get the username 
        */
	{
          struct passwd *result ;
	
	  result = getpwuid( getuid() );
          if( NULL == result ) {
            syslog( BBFTPD_WARNING, "Error getting username" );
            sprintf( currentusername, "UID %d", getuid() );
          } 
	  else
            strcpy( currentusername, result->pw_name );
	}
        /*
        ** Set the control sock to stdin and stdout
        */
        incontrolsock = 0;
        outcontrolsock = 1;

	/*close( 2 );*//*do not close stderr*/
	
	break;
      default:
	break;	/* do nothing */
    }
   
    switch( be_daemon ) {
      case 0:
      case 2:
        /* Get the remote address */
        addrlen = sizeof(his_addr);
        if (getpeername(incontrolsock, (struct sockaddr *)&his_addr, (socklen_t *)&addrlen) < 0) {
            syslog(BBFTPD_ERR, "getpeername: %s", strerror(errno));
            exit( errno );
        }
	{
	  struct sockaddr_in laddr;
	  
          addrlen = sizeof( laddr );
          if (getsockname(incontrolsock, (struct sockaddr *)&laddr, (socklen_t *)&addrlen) < 0) {
            syslog(BBFTPD_ERR, "getsockname: %s", strerror(errno));
            exit( errno );
          }
	  ctrl_port = laddr.sin_port;
	}

        if( pasvport_min == pasvport_max &&
            pasvport_min == ntohs( ctrl_port )
          ) {
          pasvport_min = 0;
          pasvport_max = 1;         /* just in case of a too clever user */
        }
	
        syslog( BBFTPD_INFO,"Getting new bbftpPRO connexion from : %s", inet_ntoa(his_addr.sin_addr) );

	break;
      default:
	break;	/* do nothing */
    }

    switch( be_daemon ) {
      case 0:
      case 2:
	sendcrypt();
	retcode = handlepackage( handler_login );
        if( -1 == retcode )
          exit( errno = EPERM );
        break;
      case 1:
	checkfromwhere();

	his_addr.sin_addr.s_addr = INADDR_ANY;
        
	syslog( BBFTPD_INFO, "bbftpdPRO started by : %s", currentusername );
        break;
      default:
	break;	/* do nothing */
    }
    
    state = S_LOGGED ;
   
    crypton |= 2;	/* cipher on */
    
/*
** Loop for the v2 protocol (also available for v1, v3, v4, v5...)
*/
    /*
    ** Set up v2 handlers
    */
    if( bbftpd_setfathersignals() < 0 ) {
        exit( errno = EPERM );
    }
    
    /*
    ** Set the umask ; first unset it and then reset to the default value
    */
    umask(0) ;
    umask(myumask) ;

    for(;;) {
      retcode = handlepackage( handler_control );

      for( ; (int)(waitpid( -1, NULL, WNOHANG )) > 0; )
	      ;/* do nothing */
      
      if( -1 == retcode ) {
        exit_clean();
        exit( errno = 0 );
      }
    }
    
}

void exit_clean() 
{
    
    switch (state) {
        case S_LOGGED :
        case S_WAITING_FILENAME_STORE :
        case S_WAITING_STORE_START :
            syslog( BBFTPD_INFO, "User %s disconnected", currentusername );
            return ;
        default :{
            return ;
        }
    }
}

my64_t convertlong(my64_t v) {
    struct bb {
        int    fb ;
        int sb ;
    } ;
    struct bb *bbpt ;
    int     tmp ;
    my64_t    tmp64 ;
    
    tmp64 = v ;
    bbpt = (struct bb *) &tmp64 ;
    tmp = bbpt->fb ;
    bbpt->fb = ntohl(bbpt->sb) ;
    bbpt->sb = ntohl(tmp) ;    
    return tmp64 ;
}

#ifndef HAVE_NTOHLL
my64_t ntohll(my64_t v) {
#ifdef HAVE_BYTESWAP_H
    return bswap_64(v);
#else
    long lo = v & 0xffffffff;
    long hi = v >> 32U;
    lo = ntohl(lo);
    hi = ntohl(hi);
    return ((my64_t) lo) << 32U | hi;
#endif
}
#define htonll ntohll
#endif
