/*
 * bbftpd/bbftpd_cert.c
 * Copyright (C) 1999, 2000, 2001, 2002 IN2P3, CNRS
 * bbftp@in2p3.fr
 * http://doc.in2p3.fr/bbftp
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */ 

/****************************************************************************

  
  
 bbftpd_cert.c   v 2.2.0 2001/10/03  - Routines creation

*****************************************************************************/
#include "bbftpd.h"
#include <sys/types.h>

#include <unistd.h>

#include <errno.h>
#include <netinet/in.h>
#include <pwd.h>
#include <stdio.h>
#include <syslog.h>
#include <sys/socket.h>

#include "config.h"
#include "common.h"
#include "structures.h"
#include "daemon_proto.h"
#include "daemon.h"
#include "version.h"
#include <gssapi.h>
#include <gfw.h>

extern  int             incontrolsock ;
extern  int             outcontrolsock ;
extern  char            currentusername[MAXLEN] ;
extern  gss_cred_id_t   server_creds;

extern xbuffer logmessage;


/*******************************************************************************
** bbftpd_cert_receive_connection :                        	                   *
**                                                                             *
**      This routine is called when a connection occurs in certificate         *
**      authentication mode. It receives the plublic key of the client and     *
**      calls the GSS Framework (GFW) api to validate the connection           * 
**      routine returns 0 to set up correctly the global parameters            *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          hisrsa              MODIFIED                                       *
**                                                                             *
**                                                                             *
**      RETURN:                                                                *
**          -1  Unrecoverable error                                            *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/
int bbftpd_cert_receive_connection() {
    char          *username ;
    struct passwd *uspass ;
    gss_buffer_desc client_name;
    OM_uint32 min_stat, maj_stat;

    printxb( &logmessage, "bbftpdPRO version %s", VERSION );

    maj_stat = gfw_accept_sec_context(&min_stat, incontrolsock, outcontrolsock, server_creds, &client_name);
    if (maj_stat != GSS_S_COMPLETE) {
        gfw_msgs_list *messages = NULL;
        gfw_status_to_strings(maj_stat, min_stat, &messages) ;
        catxb( &logmessage, " : " );
        catxb( &logmessage, messages->msg );
        while (messages != NULL) {
            syslog(BBFTPD_ERR,"gfw_accept_sec_context failed: %s", messages->msg) ;
            messages = messages->next;
        }
        reply( MSG_BAD_NO_RETRY, ptrxb( &logmessage ) );
        return -1;
    }
	
    syslog(BBFTPD_INFO,"Checked certificate : \"%s\"",(char *)client_name.value) ;
    
	/*
	** Map cert with local user
	*/
    if (globus_gss_assist_gridmap((char *)client_name.value, &username) != 0) {
        syslog(BBFTPD_ERR,"mapping failed for: %s",(char *)client_name.value) ;
	catxb( &logmessage, " : grid mapping failed" );
        reply( MSG_BAD_NO_RETRY, ptrxb( &logmessage ) );
        return -1 ;
    }
    
    syslog(BBFTPD_INFO, "Mapfile user is: %s", username);
	
    /*
    ** Here we check the username and pass and set the default dir
    */
    uspass = getpwnam(username);
    if( NULL == uspass ) {
        syslog(BBFTPD_ERR,"%s is not a local user",username) ;
        catxb( &logmessage, " : You need an account on the server" );
        reply( MSG_BAD_NO_RETRY, ptrxb( &logmessage ) );
        return -1 ;
    }

    return( su( uspass, username ) ); 
}
