/*
 * bbftpd/bbftpd_check.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 
#include "bbftpd.h"
#include <sys/types.h>

#include <errno.h>
#include <netinet/in.h>
#include <syslog.h>
#include <sys/socket.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#include "common.h"
#include "daemon.h"
#include "structures.h"
#include "daemon_proto.h"
#include "version.h"

extern  int     incontrolsock ;
extern  int     outcontrolsock ;
extern  int     newcontrolport ; 
extern  int     fixeddataport;
extern  char    currentusername[MAXLEN] ;

extern  int     ctrl_port;

extern	RSA	*hisrsa;
extern	RSA	*myprsa;
extern	unsigned char ciphkey[16];
extern	unsigned char ciphiv[8];
extern	int	crypton;


int sendrsa( int code ) {
  int lene;
  int lenn;
  unsigned char *to;
  int tolen;
  int retcode;

  const BIGNUM *n, *e;

  RSA_get0_key( myprsa, &n, &e, NULL );

  lenn = BN_bn2mpi( n, NULL );
  lene = BN_bn2mpi( e, NULL );
  tolen = sizeof( int ) + lenn + sizeof( int ) + lene;
  to = (unsigned char *)(malloc( tolen ) );
  if( NULL == to ) {
    syslog( BBFTPD_ERR, "Malloc: %s", strerror( errno ) );
    exit( errno );
  }
  BN_bn2mpi( n, to + sizeof( int ) );
  BN_bn2mpi( e, to + sizeof( int ) + lenn + sizeof( int ) );

#ifndef WORDS_BIGENDIAN
  retcode = htonl( lenn );
  memcpy( to, (char *)(&retcode), sizeof( int ) );
  retcode = htonl( lene );
  memcpy( to + sizeof( int ) + lenn, (char *)(&retcode), sizeof( int ) );
#else
  memcpy( to, (char *)(&lenn), sizeof( int ) );
  memcpy( to + sizeof( int ) + lenn, (char *)(&lene), sizeof( int ) );
#endif

  retcode = writepackage( code, tolen, to, tolen );	/* Either MSG_KEY or MSG_NKEY */
  FREE( to )
  if( retcode < 0 ) {
    syslog( BBFTPD_ERR, "Error sending %s message", MESSAGE_NAMES[code] );
    exit( errno = -retcode );
  }

  return( 0 );
}


int saversa( char *buffer ) {
  int bnlen;

  BIGNUM *n, *e;

  hisrsa = key_new();
  if( NULL == hisrsa ) {
    syslog( BBFTPD_ERR, "key_new: %s", ERR_error_string( ERR_get_error(), NULL ) );
    return( -1 );
  }
	
  memcpy( (char *)(&bnlen), buffer, sizeof( int ) );  
#ifndef WORDS_BIGENDIAN
  bnlen = ntohl( bnlen );
#endif

  n = BN_mpi2bn( buffer + sizeof( int ), bnlen, NULL );
  if( NULL == n ) {
    syslog( BBFTPD_ERR, "(n)BN_mpi2bn: %s", ERR_error_string( ERR_get_error(), NULL ) );
    return( -1 );
  }

  buffer += sizeof( int ) + bnlen;
     
  memcpy( (char *)(&bnlen), buffer, sizeof( int ) ); 
#ifndef WORDS_BIGENDIAN
  bnlen = ntohl( bnlen );
#endif

  e = BN_mpi2bn( buffer + sizeof( int ), bnlen, NULL );
  if( NULL == e ) {
    syslog( BBFTPD_ERR, "(e)BN_mpi2bn: %s", ERR_error_string( ERR_get_error(), NULL ) );
    return( -1 );
  }

  RSA_set0_key( hisrsa, n, e, NULL );

  /* Cipher too */
  buffer += sizeof( int ) + bnlen;
 
  memcpy( ciphkey, buffer, 16 );
  memcpy( ciphiv, buffer + 16, 8 );

  return( 0 );
}


int f_handler_sshd( char *buffer, int len, int code ) {
  static char logstr[64];
  int ret;

  switch( code ) {
    case MSG_NKEY:
      ret = saversa( buffer );
      if( -1 == ret )
	return( -1 );

      return( 0 );
    default:
      if( VALID_MESSAGE( code ) )
        sprintf( logstr, "Unexpected message %s", MESSAGE_NAMES[code] );
      else
        sprintf( logstr, "Unlisted message %d", code );
      reply( MSG_BAD, logstr );
      return( RET_NONE );
  }
}

/*******************************************************************************
** checkfromwhere :                                                            *
**                                                                             *
**      This routine get a socket on a port, send a MSG_LOGGED_STDIN and       *
**      wait for a connection on that port.                                    *
**                                                                             *
**      RETURN:                                                                *
**          No return but the routine exit in case of error.                   *
*******************************************************************************/

void checkfromwhere() 
{
    if( 1 == fixeddataport )
      ctrl_port = htons( newcontrolport + 1 );
    else
      ctrl_port = 0;

    genmyprsa();
    
    sendrsa( MSG_NKEY );
    
    if( -1 == handlepackage( handler_sshd ) ) {
      syslog(BBFTPD_INFO,"User %s disconnected",currentusername) ;
      exit( errno = EPERM );
    }
}
