/*
 * bbftpd/bbftpd_daemon.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 
#include "bbftpd.h"
#include <sys/types.h>

#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <stdio.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>

#include <string.h>
#include <errno.h>

#include "common.h"
#include "daemon.h"
#include "structures.h"
#include "daemon_proto.h"
#include "version.h"
/*
** Common variables for BBFTP protocole version 1 and 2
*/
extern int  newcontrolport ;
/*
** Variables for BBFTP protocole version 2
*/
extern int incontrolsock ;
extern int outcontrolsock ;


pid_t    daemonpid;

static int  controlsock = -1;

static volatile int  opencnt = 0;

int do_closedoors() {
    struct sockaddr_in server;
    int on;

    controlsock = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP ) ;
    if ( controlsock < 0 ) {
      syslog(BBFTPD_ERR, "Cannot create control socket : %s", strerror(errno));
      return( 1 );      
    }
  
    on = 1;
    if ( setsockopt(controlsock,SOL_SOCKET, SO_REUSEADDR,(char *)&on,sizeof(on)) < 0 ) {
      syslog(BBFTPD_ERR,"Cannot set SO_REUSEADDR on control socket : %s",strerror(errno)) ;
      close( controlsock );
      controlsock = -1;
      return( 1 );
    }
    
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_port = htons( newcontrolport );
    if( bind( controlsock, (struct sockaddr *)(&server), sizeof(server) ) < 0 ) {
      syslog(BBFTPD_ERR,"Bind on control socket : %s",strerror(errno)) ;
      close( controlsock );
      controlsock = -1;
      return( 1 );
    }
	
    if( listen( controlsock, 100 ) < 0 ) {
      syslog(BBFTPD_ERR,"Listen on control socket : %s",strerror(errno)) ;
      close( controlsock );
      controlsock = -1;
      return( 1 );
    }
				
    syslog( BBFTPD_DEBUG, "Doors are closed" ) ;

  return( 0 );
}

void opendoors() {
  syslog( BBFTPD_DEBUG, "Open doors request" );
  ++opencnt;
}

int slamdoors() {
  for(;;) {
    int as;

    if( controlsock != -1 ) {
      int ret;
      fd_set mask;

      FD_ZERO( &mask );
      FD_SET( controlsock, &mask );
			
      ret = select( controlsock + 1, &mask, 0, 0, NULL );
      if( ret <= 0 ) {
	if( ret < 0 && errno != EINTR )
	  syslog( BBFTPD_ERR, "Select on control socket : %s", strerror( errno ) );
	mal:
	opencnt = 0;
	close( controlsock );
        controlsock = -1;
        syslog( BBFTPD_DEBUG, "Doors are opened" );
        continue;
      }
      as = accept( controlsock, 0, 0 );
      if( as < 0 ) {
	syslog( BBFTPD_ERR, "Accept on control socket : %s", strerror( errno ) );
	goto mal;
      }
      return( as );
    }

    {
      int ret;
      struct timeval timer;
    
      timer.tv_sec  = 0;
      timer.tv_usec = 600000;	/* 0.60 seconds */
      ret = select( 0, 0, 0, 0, &timer );
      if( ret <= 0 ) {
        if( 0 == ret ) {
	  if( opencnt ) {
            opencnt = 0;
	    continue;
	  }
	  do_closedoors();
	  continue;
	}
        if( EINTR == errno ) {
          opencnt = 0;
          continue;
	}
        syslog( BBFTPD_DEBUG, "Select(poll): %s ", strerror( errno ) );
        continue;
      }
      continue;
    }
  }

}


void do_daemon(int argc,char **argv,char **envp)
{

    int        prpg ;
    int     retcode ;
    int     pid;
    
    
    if( do_closedoors() != 0 ) {
      fprintf(stderr,"Cannot use control socket : %s\n ",strerror(errno)) ;
      exit( errno = EPERM );
    }
    
    /* Fork - so I'm not the owner of the process group any more */
    retcode = fork();
    if( retcode < 0 ) {
      syslog( BBFTPD_ERR, "Fork: %s", strerror( errno ) );
      exit( errno );
    }
    /* No need for the parent any more */
    if( retcode > 0 ) exit( errno = 0 );

    daemonpid = getpid();
    
    prpg = setsid();     /* dissociate from control terminal */
    if( prpg < 0 ) {
      syslog( BBFTPD_ERR, "Setsid: %s", strerror( errno ) );
      exit( errno );
    }
    
    /* Close off all file descriptors */
    close( 0 ); close( 1 ); close( 2 );

    chdir( "/" );
    
    syslog( BBFTPD_DEBUG, "Starting bbftpdPRO in background mode" );
    
    if( bbftpd_setdaemonsignals() < 0 )
      exit( errno = EPERM );
    
    for(;;) {
      int csock;

	csock = slamdoors();
	
        /* Fork off a handler */
        pid = fork();
        if( pid < 0 ) {
            syslog( BBFTPD_ERR, "Fork: %s", strerror( errno ) );
            continue;
        }
	
        if( 0 == pid ) {
            /* I am that forked off child */
            
	    close( controlsock );
	    
	    incontrolsock = outcontrolsock = csock;
	    
            return;
        }

        /* I am the parent */
        close( csock );
    }
}

