/*
 * bbftpd/bbftpd_login.c
 * Copyright (C) 1999, 2000, 2001, 2002 IN2P3, CNRS
 * bbftp@in2p3.fr
 * http://doc.in2p3.fr/bbftp
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */ 

/****************************************************************************

  
 RETURN:
         0  login OK
        -1 login failed
 
 bbftpd_login.c  v 1.6.0  2000/03/24    - Creation of the routine. This part of
                                          code was contained in readcontrol.c
                 v 1.8.0  2000/04/14    - Introduce RSA Cryptage
                 v 1.8.3  2000/04/18    - Change the reply message in order
                                          to stop retry in case of unkown user
                                          or incorrect password
                 v 1.8.7  2000/05/24    - Modify headers
                 v 1.8.11 2000/05/24    - Add Pam for Linux (and only pam...)
                 v 1.9.0  2000/08/18    - Use configure to help portage
                                        - Change the reply message (for future
                                          use)
                 v 1.9.0  2000/09/19    - Correct bug for pam
                 v 1.9.4  2000/10/16    - Supress %m
                 v 2.0.0  2000/12/18    - Use incontrolsock instead of msgsock
                 v 2.0.1  2001/04/23    - Correct indentation
                 v 2.0.2  2001/05/04    - Correct return code treatment
                                        - Change use of tmpchar
                 v 2.1.0  2001/06/11    - Change file name
                                     
 *****************************************************************************/
#include "bbftpd.h"
#include <sys/types.h>

#ifndef	__APPLE__
#include <crypt.h>
#endif

#include <errno.h>
#include <netinet/in.h>
#include <pwd.h>
#include <syslog.h>
#include <unistd.h>

#include <string.h>
#include <stdio.h>

#include "config.h"
#include "common.h"
#include "daemon.h"
#include "structures.h"
#include "daemon_proto.h"
#include "version.h"

#ifdef HAVE_INITGROUPS
#include <grp.h>
#endif

#ifdef SHADOW_PASSWORD
#include <shadow.h>
#endif

#ifdef	HAVE_LSB2_SECURITY_PAM_APPL_H
#include <lsb2/security/pam_appl.h>
#endif

#ifdef	HAVE_SECURITY_PAM_APPL_H
#include <security/pam_appl.h>
#endif

#ifdef HAVE_SECURITY_PASS
#include <userpw.h>
#endif


extern	RSA *hisrsa;

extern char currentusername[MAXLEN] ;

extern xbuffer logmessage;

static char username[MAXLEN];
static char password[MAXLEN];

#if defined( HAVE_LSB2_SECURITY_PAM_APPL_H  ) || defined( HAVE_SECURITY_PAM_APPL_H )
/* PAM conversation function
 * We assume that echo on means login name, and echo off means password.
 */
static int PAM_conv(int num_msg, const struct pam_message **msg, struct pam_response **resp, void *appdata_ptr)
{
    int replies;
    struct pam_response *reply;

    reply = (struct pam_response *) malloc(sizeof(struct pam_response) * num_msg);
    if( NULL == reply ) 
      return PAM_CONV_ERR;

    for( replies = 0; replies < num_msg; ++replies ) {
        switch (msg[replies]->msg_style) {
        case PAM_PROMPT_ECHO_ON:
	    reply[replies].resp_retcode = 0;
	    reply[replies].resp = strdup( username );	/* PAM does free !! */
            break;
        case PAM_PROMPT_ECHO_OFF:
            reply[replies].resp_retcode = 0;
            reply[replies].resp = strdup( password );	/* PAM does free !! */
            break;
        case PAM_TEXT_INFO:
        case PAM_ERROR_MSG:
            /* ignore it... */
            reply[replies].resp_retcode = 0;
            reply[replies].resp = NULL;
            break;
        default:
           /* Must be an error of some sort... */
            return PAM_CONV_ERR;
        }
    }
    *resp = reply;
    return PAM_SUCCESS;
}

static struct pam_conv PAM_conversation =
{
    &PAM_conv,
    NULL
};
#endif /* PAM */


static void seteuser() {
  struct passwd *uspass;

  uspass = getpwnam( username );
  if( NULL == uspass )
    return;

  seteuid( uspass->pw_uid );
  setegid( uspass->pw_gid );
}

static void setetoruser() {
  seteuid( getuid() );
  setegid( getgid() );
}


int su( struct passwd *uspass, char *username ) {

#ifdef HAVE_INITGROUPS
    initgroups( username, uspass->pw_gid );
#endif	/*HAVE_INITGROUPS*/

    /*
    ** Set the euid and egid of the process [NOT uid and gid!!!]
    */
    if ( setegid(uspass->pw_gid) < 0 ) {
        syslog(BBFTPD_ERR,"Setegid(%s): %s",username,strerror(errno)) ;
        catxb( &logmessage, " : Cannot set gid: " );
	catxb( &logmessage, strerror( errno ) );
        reply( MSG_BAD, ptrxb( &logmessage ) );
        return -1 ;
    }
    if ( seteuid(uspass->pw_uid) < 0 ) {
        syslog(BBFTPD_ERR,"Seteuid(%s): %s",username,strerror(errno)) ;
        catxb( &logmessage, " : Cannot set uid: " );
	catxb( &logmessage, strerror( errno ) );
        reply( MSG_BAD, ptrxb( &logmessage ) );
        return -1 ;
    }
    if ( uspass->pw_dir == NULL ) {
        syslog(BBFTPD_ERR,"No home directory for user %s : %s",username,strerror(errno)) ;
        catxb( &logmessage, " : You need a home directory on the server" );
        reply( MSG_BAD, ptrxb( &logmessage ) );
        return -1 ;
    }
    
    /*
    ** Try to cd into home directory. If permission denied (ie no AFS token)
    ** try to cd into "/tmp"
    */

    if ( CHDIR(uspass->pw_dir) < 0) {
	int savederrno = ERRNO;
	   
        if( EACCES == savederrno ) {
            syslog( BBFTPD_WARNING, "Chdir: %s", strerror( savederrno ) );
	    
            if( CHDIR( "/tmp" ) < 0) {
		    
                syslog( BBFTPD_ERR, "Chdir: %s", strerror( ERRNO ) );
                catxb( &logmessage, " : Cannot access home directory nor /tmp" );
                reply( MSG_BAD, ptrxb( &logmessage ) );
                return -1 ;
            }
            catxb( &logmessage, " : Home directory not accessible, /tmp used instead" );
            syslog( BBFTPD_INFO, "User %s connected", username );
            strcpy(currentusername,username) ;
            reply( MSG_WARN, ptrxb( &logmessage ) );
            return 1 ;
        }
        syslog( BBFTPD_ERR, "Chdir: %s", strerror(savederrno) );
        catxb( &logmessage, " : Cannot access home directory: " );
	catxb( &logmessage, strerror( errno ) );
        reply( MSG_BAD, ptrxb( &logmessage ) );
        return -1 ;
    }
                
    syslog(BBFTPD_INFO,"User %s connected",username) ;
    strcpy(currentusername,username) ;

    return( 0 );
}


int checkuser() {
  struct passwd *uspass;
  int retcode;
  
#ifdef SHADOW_PASSWORD
  struct spwd *sunpass ;
#endif

#ifdef HAVE_SECURITY_PASS
  struct userpw *secpass ;
#endif

  char *calcpass = NULL;

#ifdef AFS
  char *inst = "";
#endif

#if defined( HAVE_LSB2_SECURITY_PAM_APPL_H  ) || defined( HAVE_SECURITY_PAM_APPL_H )
  pam_handle_t *pamh=NULL;
#endif
		      
  /*
  ** Here we check the username and pass and set the default dir
  */
  uspass = getpwnam(username);
  if( NULL == uspass ) {
    syslog(BBFTPD_ERR,"Unknown user %s",username) ;
    catxb( &logmessage, " : Unknown user (" );
    catxb( &logmessage, username );
    catxb( &logmessage, ")" );
    reply( MSG_BAD_NO_RETRY, ptrxb( &logmessage ) );
    return( -1 );
  }

/*
** If the AFS flag is set it take precedence on all other mechanism
*/
#ifdef	AFS
  setpag();
  retcode = ka_UserAuthenticate(username,inst,0,password,0,&calcpass) ;
  if( retcode != 0 ) {
    syslog(BBFTPD_ERR,"ka_UserAuthenticate message : %s ",calcpass) ;
#endif	/*AFS*/

    /*
    ** Check local user 
    */

#if defined( HAVE_LSB2_SECURITY_PAM_APPL_H  ) || defined( HAVE_SECURITY_PAM_APPL_H )
    PAM_SUCCESS == (retcode = pam_start( "bbftp", username, &PAM_conversation, &pamh )) &&
    PAM_SUCCESS == (retcode = pam_authenticate( pamh, PAM_SILENT )) &&
    PAM_SUCCESS == (retcode = pam_acct_mgmt( pamh, 0 )) &&
    PAM_SUCCESS == (retcode = pam_setcred( pamh,

#if defined(PAM_ESTABLISH_CRED) || defined(__APPLE__)
                                           PAM_ESTABLISH_CRED
#else
                                           PAM_CRED_ESTABLISH
#endif
                                         ));
    pam_end( pamh, 0 );

    if( PAM_SUCCESS != retcode ) {
      syslog( BBFTPD_ERR, "PAM error (%d) user %s", retcode, username );

#ifdef	AFS
      catxb( &logmessage, " : " );
      catxb( &logmessage, calcpass );
#else
      catxb( &logmessage, " : PAM Error" );
#endif	/*AFS*/

      reply( MSG_BAD_NO_RETRY, ptrxb( &logmessage ) );
      return( -1 );
    }

#else	/*PAM*/

    if( (strcmp(uspass->pw_passwd,"x") == 0) || 
        (strcmp(uspass->pw_passwd,"X") == 0) ) {
#if defined( SHADOW_PASSWORD )
      if( (sunpass = getspnam(username)) == NULL ) {
        syslog(BBFTPD_ERR,"Unknown user %s",username) ;

#ifdef	AFS
        catxb( &logmessage, " : " );
        catxb( &logmessage, calcpass );
#else
	catxb( &logmessage, " : Unknown user" );
#endif	/*AFS*/

        reply( MSG_BAD_NO_RETRY, ptrxb( &logmessage ) );
        return( -1 );
      }
      calcpass = (char *) crypt(password,sunpass->sp_pwdp) ;
      if( strcmp(calcpass,sunpass->sp_pwdp) != 0 ) {
        syslog(BBFTPD_ERR,"Incorrect password user %s",username) ;
        catxb( &logmessage, " : Incorrect password" );
        reply( MSG_BAD_NO_RETRY, ptrxb( &logmessage ) );
        return( -1 );
      }
#elif defined( HAVE_SECURITY_PASS )
      if( (secpass = getuserpw(username)) == NULL ) {
        syslog(BBFTPD_ERR,"Unknown user %s",username) ;

#ifdef	AFS
        catxb( &logmessage, " : " );
        catxb( &logmessage, calcpass );
#else
	catxb( &logmessage, " : Unknown user" );
#endif	/*AFS*/

        reply( MSG_BAD_NO_RETRY, ptrxb( &logmessage ) );
        return( -1 );
      }
      calcpass = (char *) crypt(password,secpass->upw_passwd) ;
      if( strcmp(calcpass,secpass->upw_passwd) != 0 ) {
        syslog(BBFTPD_ERR,"Incorrect password user %s",username) ;
        catxb( &logmessage, " : Incorrect password" );
        reply( MSG_BAD_NO_RETRY, ptrxb( &logmessage ) );
        return( -1 );
      }
#else
      syslog(BBFTPD_ERR,"No Password user %s",username) ;
      catxb( &logmessage, " : No password user" );
      reply( MSG_BAD_NO_RETRY, ptrxb( &logmessage ) );
      return( -1 );
#endif
    } 
    else {
      calcpass = (char *) crypt(password,uspass->pw_passwd) ;
      if( strcmp(calcpass,uspass->pw_passwd) != 0 ) {
        syslog(BBFTPD_ERR,"Incorrect password user %s",username) ;
        catxb( &logmessage, " : Incorrect password" );
        reply( MSG_BAD_NO_RETRY, ptrxb( &logmessage ) );
        return( -1 );
      }
    }
#endif	/*PAM*/

#ifdef	AFS
  }
#endif	/*AFS*/

  return( su( uspass, username ) );
}


int loginsequence( char *buffer ) {
    int    crtype;
    struct mess_sec *msg_sec ;
    struct mess_rsa *msg_rsa ;

    printxb( &logmessage, "bbftpdPRO version %s", VERSION );
    
    /*
    ** Read the crypt type
    */
    msg_sec = (struct mess_sec *)(buffer);
#ifndef WORDS_BIGENDIAN
    crtype = ntohl( msg_sec->crtype );
#else
    crtype = msg_sec->crtype;
#endif

    msg_rsa = (struct mess_rsa *)(buffer + sizeof( struct mess_sec ));
    
    switch( crtype) {
        case CRYPT_RSA_PKCS1_OAEP_PADDING :

#ifdef	PRIVATE_AUTH
	    if ( bbftpd_private_auth() < 0 ) {
		syslog( BBFTPD_ERR, "bbftpd_private_auth failed : %s", ptrxb( &logmessage ) );
		reply( MSG_BAD, ptrxb( &logmessage ) );
		return -1 ;
	    }
	    break;
#endif	/* PRIVATE_AUTH */
		
            /*
            ** RSA
            */
            if( decodersapass( msg_rsa, username, password ) < 0 ) {
                syslog(BBFTPD_ERR,"Decode user/password error") ;
                catxb( &logmessage, " : Decode user/password error" );
                reply( MSG_BAD, ptrxb( &logmessage ) );
                return -1 ;
            }
            break ;
        case NO_CRYPT :
            strcpy( username, (char *)(msg_rsa->cryptuser) );
            strcpy( password, (char *)(msg_rsa->cryptpass) );
            break ;
        default :
            syslog( BBFTPD_ERR, "Unkwown encryption %d", crtype );
            catxb( &logmessage, " : Unknwon encryption" );
            reply( MSG_BAD, ptrxb( &logmessage ) );
            return -1 ;
    }

    return( checkuser() );
}


static int sendBN( BIGNUM *bn ) {
  int len;
  int tolen;
  unsigned char *to;
  int retcode;

  len = BN_bn2mpi( bn, NULL );
  tolen = sizeof( int ) + len;
  to = (unsigned char *)(malloc( tolen ) );
  if( NULL == to ) {
    syslog( BBFTPD_ERR, "Malloc: %s", strerror( errno ) );
    return( -1 );
  }

  BN_bn2mpi( bn, to + sizeof( int ) );

#ifndef WORDS_BIGENDIAN
  retcode = htonl( len );
  memcpy( to, (char *)(&retcode), sizeof( int ) );
#else
  memcpy( to, (char *)(&len), sizeof( int ) );
#endif

  retcode = writepackage( MSG_CHALLENGE, tolen, to, tolen );
  FREE( to )
	  
  if( retcode < 0 ) {
    syslog( BBFTPD_ERR, "Error sending %s message", MESSAGE_NAMES[MSG_CHALLENGE] );
    return( -1 );
  }

  return( 0 );
}

int f_handler_auth( char *buffer, int len, int code ) {
  int ret;
  rsal *rl = NULL;
  char *rn = NULL;
  static char fn[512];
  static BIGNUM *challenge = NULL;

  switch( code ) {
    case MSG_NKEY:
    case MSG_KEY:
      ret = saversa( buffer );
      if( -1 == ret )
	return( -1 );

      if( MSG_NKEY == code )
	return( handlepackage( handler_auth ) );	/* expecting MSG_PASS */

      sprintf( fn, "~%s/.ssh/authorized_keys", username );
      
      seteuser();
      ret = key_try_load_public( &rl, fn, &rn );
      setetoruser();
      
      if( 0 == ret ) {
        if( NULL == rn )
          printxb( &logmessage, "Couldn't open file %s", fn );
        else
          printxb( &logmessage, "No keys found" );
        FREE( rn )
        
	reply( MSG_BAD, ptrxb( &logmessage ) );

	return( handlepackage( handler_auth ) );	/* expecting MSG_PASS */
      }

      syslog( BBFTPD_DEBUG, "Public keys loaded from %s", rn );

      ret = 0;
      while( rl != NULL ) {
        rsal *l;
	
	if( key_equal( hisrsa, rl->rsa ) )
	  ret = 1;
        RSA_free( rl->rsa );
        l = rl;
        rl = rl->next;
        FREE( l )
      }

      if( 0 == ret ) {
	printxb( &logmessage, "No matching key" );
	syslog( BBFTPD_DEBUG, ptrxb( &logmessage ) );
	
	reply( MSG_BAD, ptrxb( &logmessage ) );

	return( handlepackage( handler_auth ) );	/* expecting MSG_PASS */
      }

      syslog( BBFTPD_DEBUG, "Challenging key" );
      
      challenge = auth_generate_challenge( hisrsa );
      if( NULL == challenge ) {
	printxb( &logmessage, "auth_generate_challenge: failed" );
	syslog( BBFTPD_ERR, ptrxb( &logmessage ) );

	reply( MSG_BAD, ptrxb( &logmessage ) );

	return( handlepackage( handler_auth ) );        /* expecting MSG_PASS */
      }

      ret = sendBN( challenge );
      if( ret < 0 )
	return( -1 );
    
      return( handlepackage( handler_auth ) );		/* expecting MSG_RESPONSE */
    
    case MSG_RESPONSE:
      if( len != 16 ) {
	BN_free( challenge );
	reply( MSG_BAD_NO_RETRY, "Permission denied" );
	return( -1 );
      }
      
      ret = auth_verify_response( hisrsa, challenge, buffer );
      BN_free( challenge );
      
      if( ret ) {
	struct passwd *uspass;

	printxb( &logmessage, "bbftpdPRO version %s", VERSION );
	
	uspass = getpwnam(username);
        if( NULL == uspass ) {
          syslog(BBFTPD_ERR,"Unknown user %s",username) ;
          catxb( &logmessage, " : Unknown user (" );
          catxb( &logmessage, username );
          catxb( &logmessage, ")" );
          reply( MSG_BAD_NO_RETRY, ptrxb( &logmessage ) );
          return( -1 );
        }
	
	return( su( uspass, username ) );	/* success */
      }
      
      reply( MSG_BAD_NO_RETRY, "Permission denied" );
      return( -1 );
      
    case MSG_PASS:
      printxb( &logmessage, "bbftpdPRO version %s", VERSION );
      
      strncpy( password, buffer, sizeof( password ) );
      password[sizeof( password ) - 1] = 0;
      return( checkuser() );
    default:
      if( challenge != NULL )
	BN_free( challenge );
      return( RET_NONE );
  }
}

int authsequence( char *buffer ) {
  int retcode;
  
  strncpy( username, buffer, sizeof( username) );
  username[sizeof( username) - 1] = 0;

  syslog( BBFTPD_DEBUG, "Request to authenticate user %s", username );
  
  retcode = writepackage( MSG_OK, 0 );
  if( retcode < 0 ) {
    syslog(BBFTPD_ERR, "Error sending %s message", MESSAGE_NAMES[MSG_OK]) ;
    return( -1 );
  }

  return( handlepackage( handler_auth ) );
}
