/*
 * bbftpd/bbftpd_mklnk.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */
#include "bbftpd.h"
#include <sys/types.h>

#include <stdio.h>
#include <syslog.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>

#include "common.h"
#include "daemon.h"
#include "structures.h"
#include "daemon_proto.h"


extern	int	transferoption;

extern xbuffer logmessage;


static int bbftpd_storemklnk( char *oldpath, char *newpath );

void bbftpd_mklnk( char *buffer ) {
    char *oldpath;
    char *newpath;
    int  ret;

    memcpy( (char *)(&transferoption), buffer, sizeof( int ) );
#ifndef	WORDS_BIGENDIAN
    transferoption = ntohl( transferoption );
#endif
    
    oldpath = buffer + sizeof( int );
    newpath = buffer + sizeof( int ) + strlen( oldpath ) + 1;
    
    ret = bbftpd_storemklnk( oldpath, newpath );
    if( ret != 0 ) {
      reply( (ret < 0 ? MSG_BAD_NO_RETRY : MSG_BAD), ptrxb( &logmessage ) );
      return;
    }

    printxb( &logmessage, "Symbolic link %s -> %s created", newpath, oldpath );
    reply( MSG_OK, ptrxb( &logmessage ) );
}

static int bbftpd_storemklnk( char *oldpath, char *newpath ) {
  for(;;) {
    int ret;
    int err;    
  
    if( 0 == SYMLINK( oldpath, newpath ) ) 
	    
      return( 0 );

    err = ERRNO;
    switch( err ) {
      char *end;

      case ENOMEM:
      case EBUSY:
        ret = 1;
        break;
      case EEXIST:
	ret = 0;
	
	if( -1 == UNLINK( newpath ) ) {
		
	  err = ERRNO;
	  ret = (ENOMEM == err || EBUSY == err ? 1 : -1);
	}
	break;
      case ENOENT:
        end = strrchr( newpath, '/' );
        if( transferoption & TROPT_DIR ) {
          *end = 0;
          ret = bbftpd_storemkdir( newpath );
          *end = '/';
          if( ret != 0 )
            return( ret );
          break;
        }
	/* fall through */
      default:
        ret = -1;
        break;
    }
    
    if( ret != 0 ) {
      printxb( &logmessage, "Symlink(%s): %s", newpath, strerror( err ) );
      return( ret );
    }
  }
}
