/*
 * bbftpd/bbftpd_readcontrol.c
 * Copyright (C) Dan Scrager
 * Dan.Schrager@weizmann.ac.il
 */ 
#include "bbftpd.h"
#include <sys/types.h>

#include <errno.h>
#include <stdio.h>
#include <syslog.h>

#include <string.h>
#include <sys/stat.h>

//#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "common.h"
#include "daemon.h"
#include "structures.h"
#include "daemon_proto.h"
#include "status.h"
#include "version.h"


extern int  state ;
extern int  incontrolsock ;
extern int  outcontrolsock ;
extern	int	recvcontrolto ;
extern char *curfilename ;
extern char *realfilename ;
extern int  transferoption ;
extern int  filemode ;
extern char lastaccess[9];
extern char lastmodif[9];
extern int  sendwinsize ;        
extern int  recvwinsize ;        
extern int  buffersizeperstream ;
extern int  requestedstreamnumber ;
extern my64_t  filesize ;
extern unsigned short int  *myports ;
extern int  *mysockets ;

extern int  myumask ;
extern char *readbuffer ;
extern char *compbuffer ;

extern xbuffer logmessage;

int  protocolversion ;

int store; /*STORE or RETRIVE*/

#ifndef WORDS_BIGENDIAN
#ifndef HAVE_NTOHLL
my64_t    ntohll(my64_t v) ;
#define htonll ntohll
#endif
#endif


#define	T(x)	#x
const char * const STATUS_NAMES[] = {
	STATUS_LIST
};
#undef	T


int  *myips = NULL;

static int takefmem( int skip ) {
  if( 0 == skip ) {
    mysockets = (int *)(malloc( requestedstreamnumber * sizeof( int ) ));
    myports = (unsigned short int *)(malloc( requestedstreamnumber * sizeof( unsigned short int ) ));
    myips = (int *)(malloc( requestedstreamnumber * sizeof( int ) ));

    if( NULL == mysockets || NULL == myports || NULL == myips ) {
      printxb( &logmessage, "Malloc: %s", strerror( errno ) );
      syslog( BBFTPD_ERR, ptrxb( &logmessage ) );
      reply( MSG_BAD, ptrxb( &logmessage ) );
      free_f_var();

      state = S_LOGGED;
      return( 0 );
    }
  }
  
  return( 1 );
}


static int goahead() {
  unsigned char *buffer;
  int len;
  int i;
  int code;
  int retcode;

  retcode = takefmem( 0 == filesize && 0 == (transferoption & TROPT_UNIX) );
  if( 0 == retcode )
    return( 0 );
  
  switch( protocolversion ) {
    case 12:
      buffer = readpackage( &len, &code, NULL );
      if( code != MSG_PORTS ) {
	syslog( BBFTPD_ERR, "Error, NOT getting %s message", MESSAGE_NAMES[MSG_PORTS] );
	free_f_var() ;
	return( -1 );
      }
      memcpy( myports, buffer, requestedstreamnumber * sizeof( unsigned short int ) );
      memcpy( myips, (unsigned short int *)(buffer) + requestedstreamnumber, requestedstreamnumber * sizeof( int ) );
      
      if( bbftpd_makedatasock( requestedstreamnumber/*, (unsigned short int *)(buffer)*/ ) < 0 ) {
	syslog(BBFTPD_ERR,"Error creating data sockets") ;
	free_f_var() ;
	return( -1 );
      }

      break;

    case 13:
      if( bbftpd_getdatasock(requestedstreamnumber) < 0 ) {
        syslog(BBFTPD_ERR,"Error creating data sockets") ;
        free_f_var() ;
        return( -1 );
      }
      
      retcode = writepackage( MSG_PORTS, requestedstreamnumber * (sizeof( unsigned short int ) + sizeof( int )), myports, requestedstreamnumber * sizeof( unsigned short int ), myips, requestedstreamnumber * sizeof( int ) );
      if( retcode < 0 ) {
	  syslog(BBFTPD_ERR,"Writemessage: %s", strerror( -retcode ) );
          free_f_var() ;
          return( -1 );
      }
      
      break;

    default:
      break;	/* do nothing */
  }

  return( start_pumps() );
}

static int msgfilename( char *buffer, int len ) {
  static struct mess_store_v2 messy; 
  struct mess_store_v2 *msg_store_v2 = &messy; 
  int code;
  int retcode;
	
  FREE( realfilename )
  
  curfilename = strdup( buffer );
  if( NULL == curfilename ) {
    syslog(BBFTPD_ERR,"Error allocating memory for filename : %s",strerror(errno)) ;
    return -1 ;
  }

  if( store ) {
    realfilename = (char *) malloc( len + 11 );
    if( NULL == realfilename ) {
      syslog(BBFTPD_ERR,"Malloc: %s", strerror( errno ) );
      free_f_var();
      return -1 ;
    }
		    
    syslog(BBFTPD_DEBUG,"Request to store file %s",curfilename) ;

    bbftpd_storecheckoptions();
    if( transferoption & TROPT_TMP ) {
      /*
      ** Caculate tempfile name if TROPT_TMP is set
      */
      sprintf(realfilename,"%s.bbftp.tmp",curfilename) ;
    } else {
      sprintf(realfilename,"%s",curfilename) ;
    }

    if( 0 == (transferoption & TROPT_UNIX) ) { 
      if( 0 == filesize ) {
        retcode = bbftpd_storecreatefile( realfilename );
        if( retcode != 0 ) {
          syslog( BBFTPD_ERR, ptrxb( &logmessage ) );
          free_f_var() ;
          reply( (retcode < 0 ? MSG_BAD_NO_RETRY : MSG_BAD), ptrxb( &logmessage ) );
          state = S_LOGGED ;
          return( 0 );
        }
	    
        retcode = end_file();
        if( retcode >= 0 )
          reply( MSG_OK, "" );
        free_f_var() ;
	  
        state = S_LOGGED;
        return( 0 );
      }
    
      if( transferoption & TROPT_EXCL ) {
        if( filelength( curfilename ) == filesize ) {

          reply( MSG_OK, printxb( &logmessage, "File %s excluded", curfilename ) );
	
	  free_f_var();
	
	  state = S_LOGGED;
	  return( 0 );
        }
      }
    }
    
  }
  else {
	  
    syslog(BBFTPD_DEBUG,"Request to retrieve file: %s",curfilename) ;

    /*
    ** Check the file
    */
    retcode = bbftpd_retrcheckfile( curfilename );
    if( retcode != 0 ) {
      syslog( BBFTPD_ERR, ptrxb( &logmessage ) );
      free_f_var() ;
      reply( (retcode <= 0 ? MSG_BAD_NO_RETRY : MSG_BAD), ptrxb( &logmessage ) );

      state = S_LOGGED ;
      return 0 ;
    }
  }

#ifndef WORDS_BIGENDIAN
  msg_store_v2->transferoption       = htonl(transferoption);
  msg_store_v2->filemode             = htonl(filemode) ;
  msg_store_v2->sendwinsize          = htonl(sendwinsize) ;
  msg_store_v2->recvwinsize          = htonl(recvwinsize) ;
  msg_store_v2->buffersizeperstream  = htonl(buffersizeperstream) ;
  msg_store_v2->nbstream             = htonl(requestedstreamnumber) ;
  msg_store_v2->filesize             = htonll(filesize) ;
#else
  msg_store_v2->transferoption       = transferoption  ;       
  msg_store_v2->filemode             =  filemode;
  msg_store_v2->sendwinsize          = sendwinsize ;
  msg_store_v2->recvwinsize          = recvwinsize ;
  msg_store_v2->buffersizeperstream  = buffersizeperstream ;
  msg_store_v2->nbstream             = requestedstreamnumber;
  msg_store_v2->filesize             = filesize ;
#endif
  strncpy( msg_store_v2->lastaccess, lastaccess, 8 );
  strncpy( msg_store_v2->lastmodif, lastmodif, 8 );
  
  switch( protocolversion ) {
    case 12:
      code = MSG_TRANS_OK_V2;
      break;
    case 13:
      code = MSG_TRANS_OK_V3;
      break;
    default:
      break;	/* do nothing */
  }
  retcode = writepackage( code, sizeof( struct mess_store_v2 ), msg_store_v2, sizeof( struct mess_store_v2 ) );
  if( retcode < 0 ) {
    syslog( BBFTPD_ERR, "Error sending %s message", MESSAGE_NAMES[code] );
    free_f_var() ;
    return( -1 );
  }

  if( 0 == (transferoption & TROPT_UNIX) ) {
    if( 0 == filesize ) {
      free_f_var();

      state = S_WAITING_CREATE_ZERO;
      return( 0 );
    }

    if( 0 == store && (transferoption & TROPT_EXCL) ) {
      state = S_WAITING_CREATE_ZERO;
      return( 0 );
    }
  }
  
  return( goahead() );
}

static int msgtranstart( char *buffer, int code ) {
  return( start_pumps() );
}

static int msgretrstore( struct mess_store_v2 *msg_store_v2 ) {
		    
#ifndef WORDS_BIGENDIAN
  transferoption          = ntohl(msg_store_v2->transferoption) ;
  sendwinsize             = ntohl(msg_store_v2->sendwinsize) ;
  recvwinsize             = ntohl(msg_store_v2->recvwinsize) ;
  buffersizeperstream     = ntohl(msg_store_v2->buffersizeperstream) ;
  requestedstreamnumber   = ntohl(msg_store_v2->nbstream) ;
  if( store ) {	  
    filemode              = ntohl(msg_store_v2->filemode) ;
    filesize              = ntohll(msg_store_v2->filesize) ;
  }
#else
  transferoption          = msg_store_v2->transferoption ;
  sendwinsize             = msg_store_v2->sendwinsize ;
  recvwinsize             = msg_store_v2->recvwinsize ;
  buffersizeperstream     = msg_store_v2->buffersizeperstream ;
  requestedstreamnumber   = msg_store_v2->nbstream ;
  if( store ) {	
    filemode              = msg_store_v2->filemode ;
    filesize              = msg_store_v2->filesize ;
  }
#endif
  if( store ) {
    strncpy( lastaccess, msg_store_v2->lastaccess, 8 );
    lastaccess[8] = '\0';
    strncpy( lastmodif, msg_store_v2->lastmodif, 8 );
    lastmodif[8] = '\0';
  }
  
  syslog(BBFTPD_DEBUG,"transferoption         = %d ",transferoption) ;
  syslog(BBFTPD_DEBUG,"sendwinsize            = %d ",sendwinsize) ;
  syslog(BBFTPD_DEBUG,"recvwinsize            = %d ",recvwinsize) ;
  syslog(BBFTPD_DEBUG,"buffersizeperstream    = %d ",buffersizeperstream) ;
  syslog(BBFTPD_DEBUG,"requestedstreamnumber  = %d ",requestedstreamnumber) ;
  if( store ) {
    syslog(BBFTPD_DEBUG,"filemode               = %d ",filemode) ;
    syslog(BBFTPD_DEBUG,"lastaccess             = %s ",lastaccess) ;
    syslog(BBFTPD_DEBUG,"lastmodif              = %s ",lastmodif) ;
    syslog(BBFTPD_DEBUG,"filesize               = %" LONG_LONG_FORMAT " ",filesize) ;
  }


  transferoption &= (2 * TROPT_MPATH - 1);     /*XXX TROPT_PIPE is the last implemented option, for now */
  
  state = (store ? S_WAITING_FILENAME_STORE : S_WAITING_FILENAME_RETR);
  return 0 ;
}


int f_handler_control( char *buffer, int len, int code ) {
    int     retcode ;

    /*
    ** Depending on the state the accepted message may be different
    */
hell:
    switch (state) {
        /*
        ** Case S_LOGGED :
        **      Nothing is going on, so message accepted are :
        **
        **         MSG_CHCOS
        **         MSG_CHDIR_V2
        **         MSG_CHUMASK
        **         MSG_CLOSE_CONN
        **         MSG_LIST_V2
        **         MSG_MKDIR_V2
        **         MSG_RETR_V2
        **         MSG_STORE_V2
        **
        */
           case S_LOGGED : {
            switch( code ) {
        
		case MSG_PROT: {

		    if( len != 4 ) {
          		syslog(BBFTPD_ERR,"Unexpected length(%d) for %s message", len, MESSAGE_NAMES[code] );
          		return( -1 );
		    }

		    memcpy( (char *)(&protocolversion), buffer, sizeof( int ) );
#ifndef WORDS_BIGENDIAN
		    protocolversion = ntohl( protocolversion );
#endif

	            syslog(BBFTPD_INFO,"Using bbftpPRO protocol version %d",protocolversion) ;
		    return( 0 );
		}
		    
                case MSG_CHCOS: {
                    syslog( BBFTPD_DEBUG, "Ignoring %s", MESSAGE_NAMES[code] );
		    reply( MSG_OK, "COS ignored" );
                    return 0 ;
                }

                case MSG_CHDIR_V2: 
		case MSG_CHDIR: {
                    bbftpd_cd( buffer );
                    return( 0 );
                }
                
                case MSG_CHUMASK: {
		    int oldumask, newumask;
		    static char amask[64];

		    memcpy( (char *)(&myumask), buffer, sizeof( int ) );
#ifndef WORDS_BIGENDIAN
                    myumask = ntohl( myumask );
#endif
							    
                    /*
                    ** We reset the umask first then call twice the umask
                    ** in order to see the real umask set
                    */
                    oldumask = umask(0) ;
                    umask(myumask) ;
                    newumask = umask(myumask) ;
                    sprintf( amask, "Umask changed from %04o to %04o", oldumask, newumask );
                    reply( MSG_OK, amask );
                    return 0 ;
                }
 
                case MSG_CLOSE_CONN:
                    return( -1 );
                
                case MSG_LIST_V2:
		case MSG_LIST:
                   retcode = bbftpd_list( buffer );
                   return( retcode );
                
                case MSG_MKDIR_V2:
		case MSG_MKDIR:
                    bbftpd_mkdir( buffer );
                    return( 0 );
               
		case MSG_MKLNK:
		    bbftpd_mklnk( buffer );
		    return( 0 );

		case MSG_DELETE:
		    bbftpd_delete( buffer );
		    return( 0 );

		case MSG_RMDIR:
		    bbftpd_rmdir( buffer );
		    return( 0 );
			       	
                case MSG_RETR_V2:
		    store = 0;
		    return( msgretrstore( (struct mess_store_v2 *)(buffer) ) );
	
                case MSG_STORE_V2 :
		    store = 1;
		    return( msgretrstore( (struct mess_store_v2 *)(buffer) ) );

                default :{
                    reply(MSG_BAD_NO_RETRY,"Unexpected message in S_LOGGED state") ;
                    return( RET_NONE );
                }
            }
        }
        /*
        ** Case S_WAITING_FILENAME_STORE
        **
        **         MSG_FILENAME
        */
           case S_WAITING_FILENAME_STORE : {
            switch( code) {
                case MSG_FILENAME :
		    store = 1;
		    retcode = msgfilename( buffer, len );
		    if( retcode != 1 )
		      return( retcode );
		    return( 0 );
                default :{
                    return( RET_NONE );
		}
            }
        }
        /*
        ** Case S_WAITING_FILENAME_RETR
        **
        **         MSG_FILENAME
        */
           case S_WAITING_FILENAME_RETR : {
            switch( code ) {
                case MSG_FILENAME :
		    store = 0;
		    retcode = msgfilename( buffer, len );
		    if( retcode != 1 )
		      return( retcode );
		    return( 0 );
                default :{
                    return( RET_NONE );
                }
            }
        }
        
        /*
        ** Case S_WAITING_STORE_START
        **
        **         MSG_TRANS_START_V2
        **         MSG_ABORT
        */
        case S_WAITING_STORE_START : {
             switch( code ) {
                case MSG_TRANS_START_V2 :
                case MSG_TRANS_START_V3 :
			store = 1;
			return( msgtranstart( buffer, code ) );
                case MSG_ABORT :
			syslog( BBFTPD_ERR, "Received ABORT message" );
			free_f_var() ;
                    
			state = S_LOGGED ;
			return( 0 );
                default :
			free_f_var() ;
		    	return( RET_NONE );
            }
        }
        /*
        ** Case S_WAITING_CREATE_ZERO
        **
        **         MSG_CREATE_ZERO
        **         MSG_ABORT
        */
        case S_WAITING_CREATE_ZERO : {
             switch( code ) {
		case MSG_GOAHEAD:
		  store = 0;
		  retcode = goahead();
		  if( retcode != 1 )
		    return( retcode );
		  return( 0 );
		case MSG_ABORT:
		case MSG_ABR:
		  free_f_var();
		  /* fall through */
		case MSG_CREATE_ZERO:
		  state = S_LOGGED;
		  return( 0 );
                default:
		  free_f_var();
                  return( RET_NONE );
             }
        }
        /*
        ** Case S_WAITING_RETR_START
        **
        **         MSG_TRANS_START_V2
        **         MSG_ABORT
        */
        case S_WAITING_RETR_START : {
             switch( code ) {
                case MSG_TRANS_START_V2:
                case MSG_TRANS_START_V3:
		case MSG_RETR_START:
		    store = 0;
		    return( msgtranstart( buffer, code ) );
                case MSG_ABORT: 
		case MSG_ABR: {
                    free_f_var() ;
		    
                    state = S_LOGGED ;
                    return 0 ;
                }
                default :{
                    return( RET_NONE );
                }
            }
        }
        /*
        ** Any other state
        */
        default : {
            syslog( BBFTPD_ERR, "Got message %s in state %s", MESSAGE_NAMES[code], STATUS_NAMES[state] );
            return( RET_NONE );
        }
    }
}
