/*
 * bbftpd/bbftpd_retr.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 
#include "bbftpd.h"
#include <sys/types.h>

#if HAVE_DIRENT_H
# include <dirent.h>
#else
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# elif HAVE_SYS_DIR_H
#  include <sys/dir.h>
# elif HAVE_NDIR_H
#  include <ndir.h>
# endif
# define dirent direct
# define dirent64 direct64
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <string.h>
#include <stdio.h>

#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <fnmatch.h>
#include <signal.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <glob.h>
#include <sys/wait.h>

#include <setjmp.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#include <math.h>

#include "common.h"
#include "daemon.h"
#include "structures.h"
#include "daemon_proto.h"
#include "status.h"

#ifdef WITH_GZIP
#include <zlib.h>
#endif


extern int  transferoption ;
extern my64_t  filesize ;
extern int  requestedstreamnumber ;
extern int  buffersizeperstream ;
extern int  maxstreams ;
extern int  filemode ;
extern unsigned short int  *myports ;
extern int  *mysockets ;

extern int  *myips;

extern char *realfilename;
extern char *curfilename;
extern char currentusername[MAXLEN];

extern int incontrolsock;
extern int outcontrolsock;

extern int sendwinsize;
extern int recvwinsize;

extern int  state;
extern char lastaccess[9];
extern char lastmodif[9];
extern int  protocolversion;

extern int     be_daemon;
extern int     pasvport_min;
extern int     pasvport_max;
extern pid_t   daemonpid;
extern jmp_buf env;


extern struct sockaddr_in his_addr;


extern	int	crypton;

extern xbuffer logmessage;

extern xbuffer thecd;

extern int store;


/*******************************************************************************
** bbftpd_retrlisdir :                                                         *
**                                                                             *
**      Routine to list a directory                                            *
**                                                                             *
**      OUPUT variable :                                                       *
**          filelist   :  Where to store the file list (to be freed by caller  *
**          filelistlen:  Length of filelist buffer                            *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**                                                                             *
**      RETURN:                                                                *
**          -1  Unrecoverable error                                            *
**           0  OK                                                             *
**           1  recoverable error                                              *
**                                                                             *
*******************************************************************************/
int bbftpd_retrlistdir( char *pattern, char **filelist, int *filelistlen ) {
    DIR     *curdir;
   
    sSTAT  statbuf;

    sDIRENT *dp;
    
    int     err;
    int     prefixlen;

    xbuffer crtfile = { NULL, 0, 0 };    
    xbuffer list = { NULL, 0, 0 };
    
    char *prefixdir;

    
    *filelist = NULL;   /* play safe for subsequent FREE */
    *filelistlen = 0;
    
    {
      char *slash;
      int len;

      slash = strrchr( pattern, '/' );

      len = (NULL == slash ? 2 : slash - pattern + 1);

      prefixdir = (char *)(malloc( len + 1 ));
      if( NULL == prefixdir ) {
	printxb( &logmessage, "Malloc: %s", strerror( ENOMEM ) );
	return( 1 );
      }

      memcpy( prefixdir, (NULL == slash ? "./" : pattern), len ); 
      *(prefixdir + len) = 0;

      if( slash )
	pattern = slash + 1;
    }
    
    curdir = OPENDIR( prefixdir );
    
    if( NULL == curdir ) {
      err = ERRNO;
      printxb( &logmessage, "Opendir(%s): %s ", prefixdir, strerror( err ) );

      FREE( prefixdir )
      
      return( EACCES == err ? -1 : 1 );
    }
   
    prefixlen = strlen( prefixdir );
    if( NULL == appendxb( &crtfile, prefixdir, prefixlen ) ) {
      mal:
      printxb( &logmessage, "Malloc: %s", strerror( ENOMEM ) );
					              
      CLOSEDIR( curdir );

      FREE( prefixdir )      
      
      return( 1 );
    }
    
    do {
        sERRNO = 0;

	dp = READDIR( curdir );

	if( NULL == dp )
	  break;

        if( 0 == fnmatch( pattern, dp->d_name, 0 ) ) {
	    static char mode[3] = { '\0', '\0', '\0' };
	    int namelen;

	    namelen =  strlen( dp->d_name ) + 1;
	    if( NULL == appendxb( &list, dp->d_name, namelen ) ) {
	      freexb( &crtfile );
	      goto mal;
	    }

	    shrinkxb( &crtfile,  prefixlen );
	    if( NULL == appendxb( &crtfile, dp->d_name, namelen ) ) {
	      freexb( &list );
	      goto mal;
	    }
	    
	    if( LSTAT( ptrxb( &crtfile ), &statbuf ) < 0 ) {
		    
                printxb( &logmessage, "Lstat(%s): %s ", ptrxb( &crtfile ), strerror( ERRNO ) );

		cmal:
		
		CLOSEDIR( curdir );

		freexb( &crtfile );
		freexb( &list );

		FREE( prefixdir )
		
                return( 1 );
            }
	    
            if( S_IFLNK == (statbuf.st_mode & S_IFLNK) ) {
	        mode[0] = 'l';
	       
		if( STAT( ptrxb( &crtfile ), &statbuf ) < 0 ) {

                    /*
                    ** That means that the link refer to an unexisting file
                    */
                    mode[1] = 'u';
                } else {
                    if( S_IFDIR == (statbuf.st_mode & S_IFDIR) )
                      mode[1] = 'd' ;
                    else
                      mode[1] = 'f' ;
                }
            } else {
                mode[0] = ' ';
                if( S_IFDIR == (statbuf.st_mode & S_IFDIR) ) {
                    mode[1] = 'd';
                } else {
                    mode[1] = 'f';
                }
            }
	    
	    if( NULL == appendxb( &list, mode, 3 ) ) {
	      fmal:
	      freexb( &crtfile );
	      goto mal;
	    }

	    if( 'l' == mode[0] ) {
	      int llen;
	      static int buflen;
              static char *buf = NULL;	/* It's a leak of memory ! */

	      if( buf )
		goto inside;	      
	      
	      for( buflen = 128; ; buflen += 128 ) {
		     
		buf = (char *)(realloc( buf, buflen ));
                if( NULL == buf ) {
                  freexb( &crtfile );
                  freexb( &list );
                  goto mal;
                }
	
		inside:	
		
		llen = READLINK( ptrxb( &crtfile ), buf, buflen );

		if( -1 == llen ) {
		  printxb( &logmessage, "Readlink(%s): %s ", ptrxb( &crtfile ), strerror( ERRNO ) );
		  goto cmal;
		}
		
		if( llen < buflen ) {
		  *(buf + llen) = 0;
		  break;
		}
	      }

	      if( NULL == chainxb( &list, buf, llen + 1 ) )
		goto fmal;
	    }
	    
        }
    }while( 1 );
		
    err = ERRNO;

    CLOSEDIR( curdir );

    freexb( &crtfile );
    
    if( err != 0 ) {
        printxb( &logmessage, "Readdir(%s): %s ", prefixdir, strerror( err ) );

	freexb( &list );

	FREE( prefixdir )
	
        return( 1 );
    }

    FREE( prefixdir )

    *filelistlen = lenxb( &list );
    *filelist = ptrxb( &list );
   
    return( 0 );   
}


int sendlist( char *pattern, char **filelist, int *filelistlen ) {
    glob_t pg ;
    int    i ; 
    int    retcode ;
    char   *tmpfile ;
    
    *filelist = NULL;
    *filelistlen = 0;
    /*
    ** We get the filelist
    */
    retcode = glob(pattern,0,0,&pg);
    if( retcode < 0 ) {
	globfree(&pg) ;
        if ( GLOB_NOMATCH == retcode ) {
            /*
            ** Just reply with a message MESS_LIST_REPL with msglen equal to zero
            */
	    return( 0 );
        }
        printxb( &logmessage, "Error during glob call %d", retcode );
        return 1 ;
    }
    /*
    ** The glob call was successfull
    */
    /*
    ** Calculate the total length of filename
    */
    for (i = 0 ; i < pg.gl_pathc ; i++ )
        *filelistlen += strlen( pg.gl_pathv[i] )+1 ;
    *filelist = (char *)(malloc( *filelistlen ));
    if( NULL == *filelist ) {
        printxb( &logmessage, "Malloc: %s", strerror( errno ) );
	return( 1 );
    }
    /*
    ** Now prepare the file names separated by \0
    */
    tmpfile = *filelist;
    for (i = 0 ; i < pg.gl_pathc ; i++ ) {
        retcode = strlen( pg.gl_pathv[i] );
        strcpy( tmpfile, pg.gl_pathv[i] );
        tmpfile += retcode + 1;
    }
    return 0 ;
}


my64_t filelength( char *filename ) {
	
  sSTAT  statbuf;

  if( STAT( filename, &statbuf ) < 0 )
    return( -1 );

  if( S_IFDIR == (statbuf.st_mode & S_IFDIR) )
    return( -1 );

  return( statbuf.st_size );
}

/*******************************************************************************
** bbftpd_retrcheckfile :                                                      *
**                                                                             *
**      Routine to check a file and set the global parameters                  *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                      NOT MODIFIED                   * 
**          filemode                            MODIFIED                       *
**          lastaccess                          MODIFIED                       *
**          lastmodif                           MODIFIED                       *
**          filesize                            MODIFIED                       *
**          requestedstreamnumber               POSSIBLY MODIFIED              *
**                                                                             *
**      RETURN:                                                                *
**          -1  Incorrect check unrecoverable error                            *
**           0  OK                                                             *
**          >0  Incorrect check recoverable error                              *
**                                                                             *
*******************************************************************************/

int bbftpd_retrcheckfile( char *filename ) {
    sSTAT  statbuf;
		
    int	err;

    if( transferoption & TROPT_UNIX ) {
       if( requestedstreamnumber > maxstreams )
         requestedstreamnumber = maxstreams;
      return( 0 );
    }
    
    if( STAT( filename, &statbuf ) < 0 ) {

	err = ERRNO;
	printxb( &logmessage, "Stat(%s): %s ", filename, strerror( err) );
	
        /*
        ** It may be normal to get an error if the file
        ** does not exist but some error code must lead
        ** to the interruption of the transfer:
        **        EACCES        : Search permission denied
        **        ELOOP        : To many symbolic links on path
        **        ENAMETOOLONG: Path argument too long
        **        ENOENT        : The file does not exists
        **        ENOTDIR        : A component in path is not a directory
        */
	return( ENOMEM == err || EBUSY == err ? 1 : -1 );
    }
    /*
    ** The file exists so check if it is a directory
    */
    if( S_IFDIR == (statbuf.st_mode & S_IFDIR) ) {
        printxb( &logmessage, "File %s is a directory", filename );
        return -1 ;
    }

    snprintf( lastaccess, 9, "%08x", (unsigned int)(statbuf.st_atime) );
    snprintf( lastmodif, 9, "%08x", (unsigned int)(statbuf.st_mtime) );
    if (S_ISREG(statbuf.st_mode)) {
        filemode = statbuf.st_mode & ~S_IFREG;
    } else {
        filemode = statbuf.st_mode;
    }
    
    filesize = statbuf.st_size ;

    {	/* adjust requestedstreamnumber */
      int tmpnbport;
      
      tmpnbport = 1 + filesize / (4*1024*1024);	/* 4MB/stream */
      if( tmpnbport < requestedstreamnumber )
        requestedstreamnumber = tmpnbport;
      if( requestedstreamnumber > maxstreams ) 
	requestedstreamnumber = maxstreams;
    }

    return( 0 );   
}


static void slam( int as ) {
  char *sm[] = { "I'm spacing out, try typing louder, I might snap out of it.\r\n",
                 "~*~ I was on ~*~ but now I'm gone I ~*~ left my name ~*~ To turn you on!!! ~*~\r\n",
                 "404 error: Cannot locate (your SN), please try again in 5 minutes, or 10 minutes, or 30 sorry for whatever inconvienence this may have caused.\r\n",
                 "Did you know that when someone isn't talking for a while and then their SN goes grey that means they arn't around the computer? Seems like you just found that out!! (=\r\n",
                 "Oops...It seems as though I've wandered away from my computer...either that or I really don't want to talk to you.\r\n",
                 "Am I just here ignoring you, or am I really idle? While you are figuring that out you may aswell leave a message!\r\n",
                 "Right now I'm sitting here looking at the screen trying to see things from your point of view but I can't get my head that far up my ass.\r\n",
                 "Everyone is entitled to be stupid, but %n abuses the privilege.\r\n",
                 "A pretty girl can kiss a guy* a bird can kiss a butterfly* the rising sun can kiss the grass* but you my friend!! yes you!! YOU CAN KISS MY ASS*******\r\n",
                 "Don't piss me off today, I'm running out of places to hide to bodies\r\n",
                 "Roses are red violets are blue, God made me pretty, what the hell happen to you?\r\n",
                 "I have always woundered why people bang their heads against brick walls..... then I met you.Don't bother leaving a message.\r\n",
                 "If you didn't have feet you wouldn't wear shoes.....then why do you wear a bra\?\?!\r\n",
                 "Mirrors don't talk but lucky for you %n they don't laugh\r\n",
                 "I'm out like you in a beauty contest.\r\n",
               };
  static int idx = 0;

  write( as, sm[idx], strlen( sm[idx] ) );
  close( as );
  idx = (idx + 1) % (sizeof( sm ) / sizeof( char * ));
}


void childcloseall( int i ) {
  int j;

  for( j = i + 1; j < requestedstreamnumber; ++j )
    close( mysockets[j] );
}


static int myconnect( int i ) {
  struct sockaddr_in data_source ;
#if defined(SUNOS) || defined(_HPUX_SOURCE) || defined(IRIX) || defined(OSF1)
  int     addrlen ;
  int     erlen;
#else
  size_t  addrlen ;
  size_t  erlen;
#endif
  int retcode;
	  
  
  data_source.sin_family = AF_INET;
  data_source.sin_port = myports[i];
  data_source.sin_addr.s_addr = INADDR_ANY == (in_addr_t)(myips[i]) ? his_addr.sin_addr.s_addr : (in_addr_t)(myips[i]);

  /* Fat chance if both are INADDR_ANY... */
  if( INADDR_ANY == data_source.sin_addr.s_addr ) {
	  errno = ECONNREFUSED;
	  goto mc002;
  }

  addrlen = sizeof(data_source) ;

  retcode = connect( mysockets[i], (struct sockaddr *)&data_source, addrlen );
  if( retcode < 0 ) {
    if( EINPROGRESS == errno ) {
      fd_set mask;
      struct timeval timer;
      int er;

      for(;;) {
        FD_ZERO( &mask );
        FD_SET( mysockets[i], &mask );
        timer.tv_sec = 30;        /* thirty seconds */
        timer.tv_usec = 0;
        retcode = select( mysockets[i] + 1, 0, &mask, 0, &timer );
        if( retcode <= 0 ) {
	  if( 0 == retcode )
	    errno = ETIMEDOUT;
	  else if( EINTR == errno )
	    continue;
          printxb( &logmessage, "Select on data socket: %s ", strerror( errno ) );
	  syslog( BBFTPD_ERR, ptrxb( &logmessage ) );
          return( -1 );
        }
	break;
      }
      erlen = sizeof( er );
      if( getsockopt( mysockets[i], SOL_SOCKET, SO_ERROR, (char *)&er, (socklen_t *)&erlen ) < 0 ) {
        printxb( &logmessage, "Getsockopt on data socket: %s ", strerror( errno ) );
	syslog( BBFTPD_ERR, ptrxb( &logmessage ) );
        return( -1 );
      }
      if( 0 == er )
        goto mc001;
      errno = er;
    }

mc002:
    printxb( &logmessage, "Couldn't connect to %s:%d : %s ", inet_ntoa( data_source.sin_addr ), ntohs( data_source.sin_port ), strerror( errno ) );
    syslog( BBFTPD_ERR, ptrxb( &logmessage ) );
    return( -1 );
  }

mc001:
  {/*XXX*/
	  static char mes[64];
	  int len;

	  len = sprintf( mes, "Connection(%s:%d <-- ", inet_ntoa( data_source.sin_addr ), ntohs( data_source.sin_port ) );
	  addrlen = sizeof( data_source ); /* reused */
	  if( getsockname( mysockets[i], (struct sockaddr *)&data_source, (socklen_t *)&addrlen ) < 0 ) {
		  printxb( &logmessage, "Getsockname on data socket: %s ", strerror( errno ) );
		  syslog( BBFTPD_ERR, ptrxb( &logmessage ) );
		  return( -1 );
	  }
	  sprintf( &mes[len], "%s:%d)", inet_ntoa( data_source.sin_addr ), ntohs( data_source.sin_port ) );
	  syslog( BBFTPD_DEBUG, mes );
  }

  return( mysockets[i] );
}	  

static void cloner( int ns ) {
      pid_t pid;
      int i;

      pid = fork();
      if( 0 == (int)(pid) ) {

        for( i = 0; i < requestedstreamnumber; ++i ) {
	      if( mysockets[i] < 0 ) mysockets[i] = -mysockets[i] - 1; /* restore out of order sockets, then close them all */
	      close( mysockets[i] );
        }
					          
	close( incontrolsock );
        if( incontrolsock != outcontrolsock )
	  close( outcontrolsock );
							          
	free_f_var();
	
	incontrolsock = outcontrolsock = ns;
	
	recrypt();      /* cipher off */
	
	seteuid( getuid() );
        setegid( getgid() );
                                                                                
        longjmp( env, 3 );
      }

      close( ns );
      if( -1 == (int)(pid) )
        syslog( BBFTPD_DEBUG, "fork: %s", strerror( errno ) );
      else
        syslog( BBFTPD_DEBUG, "Successfully cloned to pid %d", (int)(pid) );
}

static int expect( unsigned short int *sp_ptr, unsigned short int hisport, int current ) {
  int i;

  for( i = current; i < requestedstreamnumber; ++i )
    if( hisport == *(sp_ptr + i) )
      return( i );
  return( -1 );
}

int myaccept() {
  struct sockaddr_in peer_addr;
#if defined(SUNOS) || defined(_HPUX_SOURCE) || defined(IRIX) || defined(OSF1)
  int     addrlen ;
#else
  size_t  addrlen ;
#endif
  int ns;
  int i;

  int keyhole;

  unsigned short int *sports;

  /* Get the source ports */
  {
	  int len, code;

	  sports = (unsigned short int *)(readpackage( &len, &code, NULL));
	  if( MSG_POSITION != code ) {
		  printxb( &logmessage, "Expecting %s, got %s! ", MESSAGE_NAMES[MSG_POSITION], MESSAGE_NAMES[code] );
		  goto maerr;
	  }
	  syslog( BBFTPD_DEBUG, "Got %s!", MESSAGE_NAMES[MSG_POSITION] );
  }

  if( 2 == be_daemon && 0 == pasvport_min && pasvport_max != 0 )
	  keyhole = mysockets[0];

  for( i = 0; i < requestedstreamnumber; ++i ) {
    int retcode;
    fd_set mask;
    struct timeval timer;
    int as;
	  
    as = mysockets[i];

    if( as < 0 ) continue; /* skip out of order sockets */

    FD_ZERO( &mask );
    FD_SET( as, &mask );
    timer.tv_sec = 30;        /* THIRTY seconds */
    timer.tv_usec = 0;
    retcode = select( as + 1, &mask, 0, 0, &timer );
    if( retcode <= 0 ) {
      if( 0 == retcode )
	errno = ETIMEDOUT;
      else if( EINTR == errno ) {
	--i;
	continue;
      }
      printxb( &logmessage, "Select on data socket: %s ", strerror( errno ) );
    maerr:
      for( i = 0; i < requestedstreamnumber; ++i ) {
	      if( mysockets[i] < 0 ) mysockets[i] = -mysockets[i] - 1; /* restore out of order sockets, then close them all */
	      close( mysockets[i] );
      }
      return( -1 );
    }
    ns = accept( as, 0, 0 );
    if( ns < 0 ) {
      printxb( &logmessage, "Accept on data socket: %s ", strerror( errno ) );
      goto maerr;
    }
    addrlen = sizeof( peer_addr );
    if( getpeername( ns, (struct sockaddr *)&peer_addr, (socklen_t *)&addrlen ) < 0 ) {
      close( ns );
      printxb( &logmessage, "Getpeername on data port(%d): %s ", i, strerror( errno ) );
      goto maerr;
    }

    /*if( peer_addr.sin_addr.s_addr == his_addr.sin_addr.s_addr )*/ {
      int place;

      place = expect( sports, peer_addr.sin_port, i );
      if( place != -1 ) {

	{/*XXX*/
		static char mes[64];
		int len;

		len = sprintf( mes, "Connection(%s:%d --> ", inet_ntoa( peer_addr.sin_addr ), ntohs( peer_addr.sin_port ) );
		addrlen = sizeof( peer_addr ); /* peer_addr is reused */
		if( getsockname( ns, (struct sockaddr *)&peer_addr, (socklen_t *)&addrlen ) < 0 ) {
			close( ns );
			printxb( &logmessage, "Getsockname on accepted socket(%d) : %s ", i, strerror(errno));
			goto maerr;
		}
		sprintf( &mes[len], "%s:%d)%s", inet_ntoa( peer_addr.sin_addr ), ntohs( peer_addr.sin_port ), (place == i ? "" : " -- out of order, restored!") );
		syslog( BBFTPD_DEBUG, mes );
	}

	if( place == i ) { /* close listening socket if no longer needed */
                int j;

                for( j = i + 1; j < requestedstreamnumber; ++j )
                        if( as == mysockets[j] ) break;
                if( j == requestedstreamnumber && !(2 == be_daemon && 0 == pasvport_min && pasvport_max != 0) ) {
                        close( as );

                        /*XXX*/syslog( BBFTPD_DEBUG, "Socket(%s, %d) closed", inet_ntoa( peer_addr.sin_addr ), ntohs( peer_addr.sin_port ) );
                }
        }

	mysockets[place] = ( i == place ? ns : (--i, -(ns + 1)) ); /* make it negative if out of order and stay on same i */

	/* make eventually reading sockets nonblock !! */
	if ( -1 == fcntl( ns, F_SETFL, O_NONBLOCK ) ) {
	  syslog(BBFTPD_ERR, "fcntl : %s", strerror(errno) );
	  goto maerr;
	}

	continue;
      }
    }   

    if( 2 == be_daemon && 0 == pasvport_min && pasvport_max != 0 ) {
      cloner( ns );
      --i;
      continue;
    }

    syslog(BBFTPD_DEBUG, "Door slammed to %s:%d", inet_ntoa(peer_addr.sin_addr), ntohs(peer_addr.sin_port) );
    slam( ns ); /* does close too */
    --i;
    continue;
  }

  if( 2 == be_daemon && 0 == pasvport_min && pasvport_max != 0 ) {
    if( 0 == fcntl( keyhole, F_SETFL, O_NONBLOCK ) ) {
      for(;;) {
	ns = accept( keyhole, 0, 0 );
	if( -1 == ns )
	  break;
	cloner( ns );
      }   
    }
    else
      syslog( BBFTPD_ERR, "fcntl : %s", strerror(errno) );

    close( keyhole );
  }
	  
  for( i = 0; i < requestedstreamnumber; ++i )
          if( mysockets[i] < 0 ) mysockets[i] = -mysockets[i] - 1; /* restore out of order sockets */

  return( 0 );
}

int end_file() {
  int retcode;
  
  do {
    retcode = closecastfd();
    if( retcode < 0 ) {
      printxb( &logmessage, "Close %s : %s ", realfilename, strerror( -retcode ) );
      break;
    }
    if( transferoption & TROPT_ACC ) {
      struct utimbuf ftime;

      unsigned int am;

      sscanf( lastaccess, "%08x", &am );
      ftime.actime = am;
      sscanf( lastmodif, "%08x", &am );
      ftime.modtime = am;
      /*
      sscanf( lastaccess, "%08x", (unsigned int *)(&ftime.actime) );
      sscanf( lastmodif, "%08x", (unsigned int *)(&ftime.modtime) );
      */

      retcode = bbftpd_storeutime( realfilename, &ftime );
      if( retcode < 0 )
        break;
    }
    if( transferoption & TROPT_MODE ) {
      retcode = bbftpd_storechmod( realfilename, filemode );
      if( retcode < 0 )
	break;
    }
    if( transferoption & TROPT_TMP ) {
      retcode = bbftpd_storerename( realfilename, curfilename );
    }
  }while( 0 );

  if( retcode < 0 ) {
    syslog( BBFTPD_ERR, ptrxb( &logmessage ) );
    
    syslog( BBFTPD_DEBUG, "File %s is deleted!", realfilename );
    bbftpd_storeunlink( realfilename );
    
    reply( MSG_BAD, ptrxb( &logmessage ) );
  }

  return( retcode );
}

static int getstatus( pid_t pid ) {
  int status;
  int ret;

  waitpid( pid, &status, 0 );

  if( WIFEXITED( status ) )
    ret = WEXITSTATUS( status );
  else
    ret = EINTR;

  return( ret );
}

/* Stats PUT user file bytes_transfered seconds kbytes/s Mbits/s*/
static struct timespec tstart;
void printstats( my64_t size ) {
  struct timespec tend; 
  struct timeval tdiff;
  float s, bs;
  static char details[512];

  clock_gettime( CLOCK_MONOTONIC, &tend );
  tdiff.tv_sec = tend.tv_sec - tstart.tv_sec ;
  tdiff.tv_usec = (tend.tv_nsec - tstart.tv_nsec) / 1000;
  if( tdiff.tv_usec < 0 ) tdiff.tv_sec--, tdiff.tv_usec += 1000000;
  s = tdiff.tv_sec + (tdiff.tv_usec / 1000000.);
  #define nz(x)   ((x) == 0 ? 1 : (x))
  bs = size / nz(s);
  printxb( &logmessage, "%s %s: %s: ", (store ? "GOT from" : "SENT to"), currentusername, curfilename );
  sprintf( details, "%" LONG_LONG_FORMAT" %.3g %.3g %.3g", size, s, bs / 1024.0, (8.0*bs) / (1024.0 * 1024.0) );
  catxb( &logmessage, details );
  syslog( BBFTPD_NOTICE, ptrxb( &logmessage ) );
}


static int waitforchildprocess( int *counter ) {
        int status;
        int ecode;
        pid_t pid;

        for(;;) {
                pid = wait( &status );	/* wait for any child process */
                if( -1 == (int)(pid) ) {
                        int err;

                        err = errno;
                        if( EINTR == err ) continue;
			syslog( BBFTPD_ERR, "Wait: %s", strerror( errno ) );
                        if( ECHILD == err ) return( 0 ); /* no child */
                        continue;
                }
                break;
        }

        --(*counter);

        if( WIFEXITED( status ) ) ecode = WEXITSTATUS( status ); /* normal termination */
        else ecode = EINTR; /* killed by a signal? */

        /*XXX*/ /*if( ecode ) syslog(BBFTPD_DEBUG, "Process %d ends with error: %s", pid, strerror( ecode ) );*/

        return( ecode );
}


char *buckets;
spadetype *spades;
volatile int *die;
my64_t *speeds;
int shmid;
int shaping;


static int waitforallchildren( int cnt ) {
        int childerr = 0;

        while( cnt > 0 ) {
                int tc;

                tc = waitforchildprocess( &cnt );
                if( tc && 0 == childerr && (((transferoption & TROPT_UNIX) && 0 == (*die & 1)) || 0 == (transferoption & TROPT_UNIX)) ) {
                        childerr = tc;
			if( cnt ) kill( 0, SIGTERM ); /* swiftly */
                }

		/*syslog(BBFTPD_DEBUG,"Children left: %d", cnt );*/
        }

        return( childerr );
}


void clean_child() {
	closecastfd();	/* effectless at get */
	if( store && 0 == (transferoption & TROPT_UNIX) ) {
		syslog( BBFTPD_DEBUG, "File %s is deleted!", realfilename );
		bbftpd_storeunlink( realfilename );
	}
}


static int bbftpd_end() {
    int osend = 0;

    clock_gettime( CLOCK_MONOTONIC, &tstart );
    
    if( transferoption & TROPT_UNIX ) {
      int ptar;		/* pipe from/to tar */
      int p[2];		/* main pipe */
      spadetype * const endspades = spades + requestedstreamnumber;

      pid_t tarpid;

      int lmd;
      int mlmd;
      
      my64_t tsize = 0; /* transferred size in Unix mode */

      if( incontrolsock != 0 )
        dup2( incontrolsock, 0 );

      if( outcontrolsock != 1 )
        dup2( outcontrolsock, 1 );

      if( -1 == pipe( p ) ) {
	syslog( BBFTPD_ERR, "Pipe: %s", strerror( errno ) );
	mal:
	*die |= 32; /* pipe error */
	goto end2;
      }

      tarpid = fork();
      switch( (int)(tarpid) ) {
        char *argv[4];
	char *dir;
	char *file;
	
	case 0:	
          argv[0] = "sh";
          argv[1] = "-c";
          argv[3] = NULL;

	  if( 0 == store ) {
	    close( p[0] );
            if( 1 != p[1] ) {
              dup2( p[1], 1 );
              close( p[1] );
            }

	    if( transferoption & TROPT_PIPE )
	      goto exe;

            dir = strrchr( curfilename, '/' );
            if( dir ) {
	      file = dir + 1;
	      if( dir == curfilename )
		dir = "/";
	      else {
		*dir = 0;
		dir = curfilename;
	      }
            }
            else {
              dir = ( lenxb( &thecd ) ? ptrxb( &thecd ) : "." );
              file = curfilename;
            }

#if defined( WITH_RFIO ) || defined( WITH_RFIO64 )
            {
              char *w = strpbrk( file, "[]{}*?+^$," );
#endif
					              
	      printxb( &logmessage,

#if defined( WITH_RFIO ) || defined( WITH_RFIO64 )
                       ( NULL == w ? "rftar %s -c -f - --totals -C %s %s%s %s" :
			             "rftar %s -c -f - --totals -C %s `rfdir %s | awk '{print $9}' | grep -E %s` %s"
                       ),
#else
		                     "tar %s -c -f - --totals -C %s `cd %s && ls -d %s` %s",
#endif
		       
		       ( transferoption & TROPT_GZIP ? "-z" : "" ), dir,
		       (
		       
#if defined( WITH_RFIO ) || defined( WITH_RFIO64 )
                         NULL == w ? "" :
#endif
			 
			 dir
                       ),
                       file,
		       (1 != be_daemon ? "2>/dev/null" : "" ) 
                     );

#if defined( WITH_RFIO ) || defined( WITH_RFIO64 )
            }
#endif

	  }
	  else {
	    close( p[1] );
	    if( 0 != p[0] ) {
	      dup2( p[0], 0 );
	      close( p[0] );
	    }

	    if( transferoption & TROPT_PIPE )
	      goto exe;

            if( '/' == *curfilename )
              dir = curfilename;
            else {
              static xbuffer xdir = { NULL, 0, 0 };

	      dir = printxb( &xdir, "%s/%s", (lenxb( &thecd ) ? ptrxb( &thecd ) : "."), curfilename );
	    }
				
            printxb( &logmessage, "%smkdir -p %s %s %star %s %s %s -x -f - --totals -C %s %s",

#if defined( WITH_RFIO ) || defined( WITH_RFIO64 )
				  "rf",
#else
				  "",
#endif
				  
				  dir,

#if defined( WITH_RFIO ) || defined( WITH_RFIO64 )
				  ";", "rf",
#else
				  "&&", "",
#endif

				  ( transferoption & TROPT_ACC ? "--atime-preserve" : "" ), 
				  ( transferoption & TROPT_MODE ? "-p" : "" ), 
				  ( transferoption & TROPT_GZIP ? "-z" : "" ), 
				  dir,
				  (1 != be_daemon ? "2>/dev/null" : "")
		   );
	  }
          argv[2] = ptrxb( &logmessage );
		       
	  exe:
	  if( transferoption & TROPT_PIPE)
	    argv[2] = curfilename;
	  
	  syslog( BBFTPD_DEBUG, "sh -c '%s'", (transferoption & TROPT_PIPE) ? curfilename : ptrxb( &logmessage ) );

	  execvp( argv[0], argv );
	  syslog( BBFTPD_DEBUG, "Execvp: %s", strerror( errno ) );
	  exit( errno );

	case -1:
	  syslog( BBFTPD_ERR, "Fork: %s", strerror( errno ) );
	  close( p[0] );
	  close( p[1] );
	  goto mal;

	default:
	  if( 0 == store ) {
	    close( p[1] );
	    ptar = p[0];
	  }	   
	  else {
	    close( p[0] );
	    ptar = p[1];
	  }
	  break;
      }

      if( shaping ) {
	lmd = 6 * 4 / requestedstreamnumber;
	if( lmd < 2 ) lmd = mlmd = 2;
	else mlmd = lmd;
      }

      for(;;) {
	register char *lbucket;
	register spadetype *lspade;
	my64_t mint;
	my64_t st;
	my64_t stsq;

	for( lspade = spades, lbucket = buckets + 2 * sizeof( int ); lspade < endspades; ++lspade, lbucket += (buffersizeperstream * 1024 + 3 * sizeof( int )) ) {
	
	      	for( ; 0 == *die && FILLED == lspade->state; ) {
			struct timeval timer;
                        
	  		timer.tv_sec  = 0;
			timer.tv_usec = 1111;
			select( 0, 0, 0, 0, &timer );
		}
                
  		if( FILLED == lspade->state ) {
			for( ; (*die & 1) && FILLED == lspade->state; ) {
				struct timeval timer;
	
				timer.tv_sec  = 0;
				timer.tv_usec = 1111;
				select( 0, 0, 0, 0, &timer );
			}
			if( FILLED == lspade->state ) goto inside; /* Bail out now */
		}

	  if( 0 == store ) {
	    if( shaping ) {
	      int k = lspade - spades;
	      register my64_t *lspeed = speeds + k;

	      if( -1 == lmd ) {
		if( -1 == mint )
		  lspade->length = buffersizeperstream * 1024;
		else {
		  lspade->length = (int)(buffersizeperstream * 1024.0 * mint / (float)(*lspeed) * 1.3);
		  if( lspade->length < 1024 ) lspade->length = 1024;
		  else if( lspade->length > buffersizeperstream * 1024 )
		    lspade->length = buffersizeperstream * 1024;
		}
	      }
	      else if( lmd >= 0 )
		lspade->length = buffersizeperstream * 1024;

	      if( 0 == lmd )
		if( 0 == k )
		  mint = *lspeed;
		else if( *lspeed < mint )
		  mint = *lspeed;

	      if( 0 == lmd || -12 == lmd )
		if( 0 == k ) {
		  st = *lspeed;
		  stsq = st * st;
		}
		else {
		  st += (*lspeed);
		  stsq += (*lspeed) * (*lspeed);
		}
	
	      if( lmd > 0 || lmd <= -9 ) *lspeed = -mlmd;
	      else if( -1 == lmd ) *lspeed = -2;
	    }
	    lspade->length = readpipe( ptar, lbucket, shaping ? lspade->length : buffersizeperstream * 1024 );
	    if( lspade->length <= 0 ) {
		
	      inside:
	      lspade->length = 0;
	      lspade->state = FILLED;

              goto end;
            }
	    tsize += lspade->length;
	  }
	  else {
	    int len = 0;
	 
            if( lspade->length > 0 )
              len = writepipe( ptar, lbucket, lspade->length );
		              
	    if( lspade->length <= 0 || len != lspade->length ) {
	      goto end;
	    }
	    tsize += len;
	  }
	  lspade->state = FILLED;
	}
	
	if( shaping ) {
	  static my64_t psig6 = -1;
	  my64_t sig6;
	  static my64_t pop;
	  static my64_t msig6;
	  static my64_t msd;

	  if( 0 == lmd ) msig6 = st / requestedstreamnumber, msd = sqrt( (stsq - (st / requestedstreamnumber * st)) / (const int)(requestedstreamnumber - 1) );

	  if( -8 == lmd ) mlmd = 3;

	  if( -12 == lmd ) {
	    static int clk = 0;

	    sig6 = st / requestedstreamnumber;
	    //syslog( BBFTPD_INFO, "AVRG=%d", (int)(sig6/1000) );

	    if( psig6 < 0 ) {
	      if( -1 == psig6 ) {
		my64_t sd;

		sd = sqrt( (stsq - (st / requestedstreamnumber * st)) / (const int)(requestedstreamnumber - 1) );
		if( sig6 > msig6 || sd > msd ) {
		  if( sig6 > msig6 && sd > msd ) {
		    //syslog( BBFTPD_INFO, "Reshaping to measured" );
		    psig6 = -2;
		    lmd = -1;
		    mint = -1;
		    continue;
		  }
		  //syslog( BBFTPD_INFO, "Reshaping" );
		  lmd = -13;
		  mlmd = 1;
		  continue;
		}
	      }
	      psig6 = sig6;
	      lmd = -10;
	      pop = 1;
	      continue;
	    }

	    clk = (clk + 1) % (psig6 > 20000 ? 32 : 64);
	    if( 3 * sig6 > 4 * psig6 )
              /*syslog( BBFTPD_INFO, "Reshaping down" )*/;
	    else if( 0 == clk )
	      /*syslog( BBFTPD_INFO, "Reshaping clocked" )*/;
            else { 
	      psig6 = (pop * psig6 + sig6) / (pop + 1);
              pop = (pop + 1) % 3; if( 0 == pop ) ++pop;

	      lmd = -10;
	      continue;
	    }

	    clk = 0;
	    mlmd = 1;
	  }

	  if( -13 == lmd ) {
	    lmd = 6 * 4 / requestedstreamnumber;
	    if( lmd < 2 ) lmd = mlmd = 2;
	    else mlmd = lmd;
            psig6 = -1;
            continue;
	  }

	  --lmd; continue;
	}

      }
      
      end:
      close( ptar );

      { 
        int tarend = getstatus( tarpid );

	tarend &= ~(1 == be_daemon ? 2 : 0); /* quick hack for ad hoc ssh server in un/tar mode */

	osend = tarend;

        syslog( BBFTPD_DEBUG, "Piped process ends, status: %s", strerror( tarend ) );
      }
        
      end2:
      osend |= waitforallchildren( requestedstreamnumber );

      printstats( tsize );
      
      syslog( osend ? BBFTPD_ERR : BBFTPD_DEBUG, osend ? "Piped transfer lost" : "Piped transfer successful" );
     
      /*shmdt( buckets );*/
      shmctl( shmid, IPC_RMID, NULL );

      return( osend );	
    }


    osend = waitforallchildren( requestedstreamnumber );
    if( osend ) {
	    clean_child();
	    syslog( BBFTPD_ERR, "Failed: %s: %s", curfilename, strerror( osend ) );
	    return( osend );
    }

    if( store ) osend = end_file();	/* can alter final result */

    if( osend >= 0 ) printstats( filesize );

    /*XXX*/ /*if( store ) sync();*/

    return( osend >= 0 ? 0 : -osend  );
}


/*******************************************************************************
** bbftpd_retrtransferfile :                                                   *
**                                                                             *
**      Routine to transfer a file                                             *
**                                                                             *
**      INPUT variable :                                                       *
**          filename    :  file to send    NOT MODIFIED                        *
**                                                                             *
**      OUTPUT variable :                                                      *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                              
**                                                                             *
**      RETURN:                                                                *
**          -1  transfer failed unrecoverable error                            *
**           0  Keep the connection open (does not mean that the file has been *
**                successfully transfered)                                     *
**          >0  Recoverable error but calling has the cleaning to do           *
**                                                                             *
*******************************************************************************/

static int crt;
static char *phpath;

static int bbftpd_transferfile() {

  OFF_T	nbperchild;
  OFF_T	nbtosend;
  OFF_T	startpoint;
  OFF_T	nbread;
  OFF_T	numberread;

  int	retcode;
  int	fd;
  int	ns;
  
  char *bucket;
  spadetype *spade;
  my64_t *speed;

  int en;


  if( transferoption & TROPT_UNIX ) {
    bucket = buckets + crt * (buffersizeperstream * 1024 + 3 * sizeof( int )) + 2 * sizeof( int );
    spade = spades + crt;
    if( shaping )
      speed = speeds + crt;
      
    nbtosend = buffersizeperstream * 1024;
  }
  else {
    nbperchild = filesize / requestedstreamnumber;
            
    startpoint = crt * nbperchild;
    if( crt == requestedstreamnumber - 1 )
      nbtosend = filesize - startpoint;
    else
      nbtosend = nbperchild;
  }
    
  if( bbftpd_setchildsignals() < 0 ) {
    if( transferoption & TROPT_UNIX ) {
      *die |= 2; /* signal handling error */
      spade->state = EMPTY;
    }
    exit( errno );
  }

  /*syslog( BBFTPD_DEBUG, "Child %d starting", getpid() );*/
	
  childcloseall( crt );
  ns = mysockets[crt];
	 
  retcode = 1024 * (store ? recvwinsize : sendwinsize);
  setsockopt( ns, SOL_SOCKET, (store ? SO_RCVBUF : SO_SNDBUF), (char *)(&retcode), sizeof( retcode ) );
	
  close( incontrolsock );
  if( incontrolsock != outcontrolsock )
    close( outcontrolsock );

  incontrolsock = outcontrolsock = ns;	/* for read&writemessage&package */

  if( 0 == (transferoption & TROPT_HIDDEN) ) crypton = 0;
	
  if( 0 == (transferoption & TROPT_UNIX) ) {   
    if( store ) {
      sSTAT statbuf;
	    
      if( STAT( phpath, &statbuf ) < 0 ) {
		    
        en = ERRNO;

        syslog(BBFTPD_ERR, "Stat(%s): %s", phpath, strerror( en ) );
        close( ns );
        exit( errno = en );
      }
    }

    fd = OPEN( phpath, (store ? O_WRONLY : O_RDONLY), 0666 );

    if( fd < 0 ) {
      en = ERRNO;

      if( store )
	      
        UNLINK( phpath );

      syslog( BBFTPD_ERR, "Open(%s): %s", phpath, strerror( en ) );
      close( ns );
            
      exit( errno = en );
    }

    /* Is this is the last writing process? */
    if( store && (en == requestedstreamnumber - 1) ) {
	    /* Seek to the end of file, in advance! */
	    if( LSEEK( fd, filesize, SEEK_SET ) < 0 )
		    goto seek001;
	    /* Allocate the entire file now!? */
	    if( WRITE( fd, "", 0 ) < 0 )
		    goto seek001;
    }

    if( LSEEK( fd, startpoint, SEEK_SET ) < 0 ) {

      seek001:
      en = ERRNO;
      
      syslog( BBFTPD_ERR, "LSEEK/WRITE(%s): %s", phpath, strerror( en ) );
		
      CLOSE( fd );

      if( store )
	      
        UNLINK( phpath );

      close( ns );
      exit( errno = en );
    }
  }

	nbread = 0;
        while( nbread < nbtosend ) {
	  static char *readbuffer = NULL;
	      
	  if( transferoption & TROPT_UNIX ) {
	    for( ; 0 == *die && EMPTY == spade->state; ) {
	      struct timeval timer;
					                
	      timer.tv_sec  = 0;
	      timer.tv_usec = 1111;
	      select( 0, 0, 0, 0, &timer );
	    }
	    if( EMPTY == spade->state ) {
	      close( ns );
	      exit( errno = 0 );
	    }
	    if( 0 == spade->length )
	      break;
	  }

	  if( 0 == store ) {
	    char *rb;
	    
	    retcode = 1;
	    if( transferoption & TROPT_UNIX ) {	/* read from pipe */
	      numberread = spade->length;
	      rb = bucket;
	    }
	    else {	/* read from file */

	      if( NULL == readbuffer ) {
	        readbuffer = (char *)(malloc( buffersizeperstream * 1024 ));
	        if( NULL == readbuffer ) {
		  en = errno;
		  syslog(BBFTPD_ERR, "Malloc: %s", strerror( errno ) );

		  CLOSE( fd );

		  close( ns );
		  exit( errno = en );
	        }
	      }

	      numberread = READ( fd, readbuffer, (buffersizeperstream*1024 <= nbtosend - nbread ? buffersizeperstream*1024 : nbtosend-nbread) );

	      if( numberread <= 0 ) {
	        en = ERRNO;
	        syslog(BBFTPD_ERR,"Read: %s", strerror( en ) ) ;

	        CLOSE( fd );

	        close( ns );
	        exit( errno = en );
	      }

              nbread += numberread;
	      rb = readbuffer;
	    }

            if( transferoption & TROPT_GZIP && ((0 == (transferoption & TROPT_UNIX)) || (transferoption & TROPT_PIPE)) ) {
			
#ifdef WITH_GZIP        
	      uLong bufcomplen;
	      static Bytef *compbuffer = NULL;

	      if( NULL == compbuffer ) {
	        compbuffer = (Bytef *)(malloc( buffersizeperstream * 1024 ));
	        if( NULL == compbuffer ) {
		  en = errno;
		  syslog( BBFTPD_ERR, "Malloc: %s", strerror( errno ) );

		  CLOSE( fd );

		  close( ns );
		  exit( errno = en );
	        }
	      }
			
              bufcomplen = buffersizeperstream * 1024;

              retcode = compress( compbuffer, &bufcomplen, (Bytef *)(rb), (uLong)(numberread) );
	      if( 0 == retcode ) {
		numberread = (OFF_T)(bufcomplen);
		rb = (char *)(compbuffer);
		goto wtn001;
	      }
	      else
		retcode = 1;
#endif	/* WITH_GZIP */
	      
	    }

	    /* write to network */
	    wtn001:
	    if( shaping && 0 > (*speed) ) stopwatch( START );
	    retcode = writepackage( (0 == retcode ? DATA_COMPRESS : DATA_NOCOMPRESS), (int)(numberread), rb, (int)(numberread) );
	    if( shaping && 0 > (*speed) ) {
	      static int avrg = 0;
	      static my64_t as = 0;
	      my64_t s;
	      int mod;

	      s = stopwatch( STOP );
	      mod = -(int)(*speed);
	      as = (avrg * as + s) / (avrg + 1);
	      if( avrg == mod - 1 ) {
		*speed = as;
		if( 0 == as ) ++(*speed);
	      }
	      avrg = (avrg + 1) % mod;
            }
	    if( retcode < 0 ) {
              if( transferoption & TROPT_UNIX ) {
                *die |= 4; /* write to network error */
                spade->state = EMPTY;
              }
              else
		      
                CLOSE( fd );
	      
              close( ns );
              exit( errno = -retcode );
            }

            if( transferoption & TROPT_UNIX ) {
              spade->state = EMPTY;
              continue;
            }
	  }
	  else {	/*XXX*/	
	    char *buffer;
            int lentowrite;
            int tmplen;
	       	
	    /* read from network */

	    if( transferoption & TROPT_GZIP ) {
	      if( (0 == (transferoption & TROPT_UNIX)) || (transferoption & TROPT_PIPE) )
	        goto rfn001;
	    }

            if( transferoption & TROPT_UNIX )
              buffer = (char *)readpackage( &lentowrite, &retcode, (unsigned char *)(bucket - 2 * sizeof( int )) );
            else {
              rfn001:
              buffer = (char *)readpackage( &lentowrite, &retcode, NULL );
            }
            if( MSG_NONE == retcode )
              retcode = -( buffer ? ECONNRESET : EMSGSIZE );

            if( retcode < 0 ) {
              if( transferoption & TROPT_UNIX ) {
                spade->state = EMPTY;
	        spade->length = lentowrite;
		*die |= 8; /* read from network error */
              }
              else {

                CLOSE( fd );

                UNLINK( phpath );

              }
              close( ns );
              exit( errno = -retcode );
            }

#ifdef	WITH_GZIP
	    if( (transferoption & TROPT_GZIP) && ( (0 == (transferoption & TROPT_UNIX)) || (transferoption & TROPT_PIPE) ) )
	      if( DATA_COMPRESS == retcode ) {
	        int rc;
	        uLong bufcomplen;
	        static Bytef *compbuffer = NULL;

	        if( transferoption & TROPT_PIPE )
	          compbuffer = bucket;
	        else if( NULL == compbuffer ) {
	          compbuffer = (Bytef *)(malloc( buffersizeperstream * 1024 ));
                  if( NULL == compbuffer ) {
                    en = errno;
		    syslog( BBFTPD_ERR, "Malloc: %s", strerror( errno ) );

                    CLOSE( fd );

                    UNLINK( phpath );

                    close( ns );
                    exit( errno = en );
                  }
                }

                bufcomplen = buffersizeperstream * 1024;

                rc = uncompress( compbuffer, &bufcomplen, (Bytef *)(buffer), (uLong)(lentowrite) );
                if( rc != 0 ) {
	          syslog(BBFTPD_ERR, "Uncompress: %s", strerror( EILSEQ ) );

                  CLOSE( fd );

                  UNLINK( phpath );

                  close( ns );
                  exit( errno = EILSEQ );
                }

                lentowrite = bufcomplen;
                buffer = (char *)(compbuffer);
              }
	      else if( transferoption & TROPT_PIPE )
	        memcpy( bucket, buffer, lentowrite );
#endif	/* WITH_GZIP */
	  
            /* write to pipe */
            if( transferoption & TROPT_UNIX ) {
              spade->length = ( MSG_ACK == retcode ? 0 : lentowrite );
              spade->state = EMPTY;
	      if( MSG_ACK == retcode ) {
		      /*XXX*/syslog( BBFTPD_DEBUG, "Transfer is over, got MSG_ACK!" );
		      /*if( 0 == spade->length )*/
		      break;
	      }
              continue;
            }
		    
	    /* write to file */
            tmplen = 0;
            while ( tmplen < lentowrite ) {

	      retcode = WRITE( fd, buffer + tmplen, lentowrite - tmplen );

              if( retcode < 0 ) {
                en = ERRNO;

	        syslog( BBFTPD_ERR, "WRITE(%s): %s", phpath, strerror( en ) );

                CLOSE( fd );

                UNLINK( phpath );

                close( ns );
                exit( errno = en );
              }
	    
              tmplen += retcode;
            }

            nbread += lentowrite;
          }
        }

        /* close the file */
        if( 0 == (transferoption & TROPT_UNIX ) ) {

          retcode = CLOSE( fd );

          if( retcode < 0 && store ) {
            en = ERRNO;

            UNLINK( phpath );

            close( ns );
            exit( errno = en );
          }
        }

	/*XXX*/	  
        if( store || (transferoption & TROPT_UNIX) ) {
		
		if( (transferoption & TROPT_UNIX) && store ) *die |= 1;	/* early! */ 
          
		retcode = writepackage( MSG_ACK, 0 );
      		if( retcode < 0 ) {
	    		/*XXX*/syslog( BBFTPD_DEBUG, "Failed to send MSG_ACK!" );
	    		if( transferoption & TROPT_UNIX ) {
		  		*die |= 4;	/* write to network error */
				if( store ) *die &= ~1;	/* transfer may be over though */

			    bad001:
		  		spade->state = EMPTY;
	    		} else if( store ) UNLINK( phpath );

		    bad002:
	    		close( ns );
			exit( errno = -retcode );
      		}
	    	/*XXX*/syslog( BBFTPD_DEBUG, "Sent MSG_ACK!" );
	}

	if( 0 == store ) {
            int len, code;

            readpackage( &len, &code, NULL );
	    syslog( BBFTPD_DEBUG, MSG_ACK == code ? "Got %s!!" : "Did not get %s!!", MESSAGE_NAMES[MSG_ACK] );
	    if( MSG_ACK != code ) {
		    retcode = -ENETRESET;
		    if( transferoption & TROPT_UNIX ) {
			    *die |= 8; /* read from network error */
			    goto bad001;
		    }
		    goto bad002;
	    }
	}

        if( transferoption & TROPT_UNIX ) {

		if( 0 == store ) *die |= 1;	/* late! */
		
		spade->state = EMPTY;
	}
        else {
          /*syslog( BBFTPD_DEBUG, "Child has %s %" LONG_LONG_FORMAT " bytes, ends correctly", (store ? "received" : "sent"), (my64_t)(nbread) );*/
	}

        /* close network */
        close( ns );

        exit( errno = 0 );
}
	

int start_pumps() {
  int ns;
  int i;
  int retcode;

  /* Connect/accept all */
  switch( protocolversion ) {
    case 12:
      /* Active mode */
      for(  i = 0; i < requestedstreamnumber; ++i ) {
        ns = myconnect( i );
        if( -1 == ns ) {
          childcloseall( -1 );
          break;
        }
      }
      break;
    case 13:
      /* Passive mode */
      ns = myaccept();
      break;
    default:
      break; /* Do nothing */
  }

  if( -1 == ns ) {
    syslog( BBFTPD_ERR, ptrxb( &logmessage ) );
    free_f_var();
    state = S_LOGGED;
    return( 0 );
  }
  
  retcode = fork();
  switch( retcode ) {
    case 0:
      break;
    case -1:
      syslog( BBFTPD_ERR, "Fork: %s", strerror( errno ) );
      /* fall through */
    default:
      childcloseall( -1 );
      free_f_var();
      state = S_LOGGED;
      return( 0 );
  }

  /* Here ! */
  setreuid( geteuid(), -1 );
  setregid( getegid(), -1 );

  if( bbftpd_setsonsignals() < 0 )
    exit( errno );

  if( transferoption & TROPT_UNIX ) {

    shaping = (0 == store) && (transferoption & TROPT_MPATH) && (requestedstreamnumber > 1);

    shmid = shmget( IPC_PRIVATE, requestedstreamnumber * ( buffersizeperstream * 1024 + 3 * sizeof( int ) + sizeof( spadetype ) + (shaping ? sizeof( my64_t ) : 0) ) +  sizeof( int ), IPC_CREAT | IPC_EXCL | 0770 );
    if( -1 == shmid ) {
      shmid = errno;
      syslog( BBFTPD_ERR, "Shmget: %s", strerror( errno ) );
      exit( errno = shmid );
    }
    buckets = shmat( shmid, NULL, 0 );
    if( (char *)(-1) == buckets ) {
      retcode = errno;
      syslog( BBFTPD_DEBUG, "Shmat: %s", strerror( errno ) );
      shmctl( shmid, IPC_RMID, NULL );
      exit( errno = retcode );
    }
    spades = (spadetype *)(buckets + requestedstreamnumber * (buffersizeperstream * 1024 + 3 * sizeof( int )));
    memset( spades, store ? FILLED : EMPTY, requestedstreamnumber * sizeof( spadetype ) );
    die = (int *)(spades + requestedstreamnumber);
    *die = 0;
    if( shaping )
      speeds = (my64_t *)(die + 1);
  }
  else {
    if( store ) {
      retcode = bbftpd_storecreatefile( realfilename );
      if( retcode != 0 ) {
        syslog( BBFTPD_ERR, ptrxb( &logmessage ) );
        exit( errno = ENOENT );
      }

#if defined( WITH_RFIO ) || defined( WITH_RFIO64 )
      if( 1 == is_rfio( realfilename ) ) {
        if( 0 == rfio_HsmIf_FindPhysicalPath( realfilename, &phpath ) )
          exit( errno = ERRNO );
      }
      else
#endif
	phpath = realfilename;
    }
    else
      phpath = curfilename; /*XXX*/
  }

  setsid(); /* always!! get a new process group */

  /*
  ** Now start all our children
  */
  for( i = 0; i < requestedstreamnumber; ++i ) {
    retcode = fork();
    switch( retcode ) {
      case 0:
	crt = i;
        bbftpd_transferfile(); /* new process, no return */
        break;
      case -1:
	retcode = errno;
        syslog( BBFTPD_ERR, "Fork: %s", strerror( errno ) );
        if( transferoption & TROPT_UNIX ) {
          *die |= 16; /* fork error */
          shmctl( shmid, IPC_RMID, NULL );
        }
	if( i ) {
		kill( 0, SIGTERM );
		while( i > 0 ) waitforchildprocess( &i );
	}
        clean_child();
        exit( errno = retcode );
      default:
        close( mysockets[i] );
        continue;
    }
  }
    
  free_n_var();
  {
	  int status = bbftpd_end();
  
	  syslog( BBFTPD_DEBUG, "Leader ends with status: %s", strerror( status ) );
	  exit( errno = status );
  }
}
