/*
 * bbftpd/bbftpd_signals.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 

#ifdef	__APPLE__
#define _POSIX_C_SOURCE 1L
#endif

#include "bbftpd.h"
#include <sys/types.h>

#include <errno.h>
#include <signal.h>
#include <syslog.h>

#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include "common.h"
#include "daemon.h"
#include "structures.h"
#include "daemon_proto.h"
#include "status.h"


extern	int	transferoption;

extern	volatile int *die;


static void bbftpd_sigusr1( int sig ) {
	opendoors();
}

static void bbftpd_sigtermfather( int sig ) {
	/* you only die once */
	syslog( BBFTPD_DEBUG, "Killed by %s", (SIGTERM == sig ? "SIGTERM" : "SIGINT") );

	/* the main bbftpPRO server is in this group */
	exit( errno = EINTR );
}

static void bbftpd_sigtermson( int sig ) {
	static int again = 0;

        ++again;

        if( 1 == again ) { /* you only die once */
                if( transferoption & TROPT_UNIX ) *die |= 64; /* signal TERM or INT error */

                /* kill process group:
                 * the ringleader and all its children are in this group
		 */
                kill( 0, sig );
        }
}

static void bbftpd_sigtermchild( int sig ) {
      	if( transferoption & TROPT_UNIX ) *die |= 64; /* signal TERM or INT error */

      	exit( errno = EINTR ); /* does shmdt() */
}


int bbftpd_resetallsignals( void (*hndl)(int)/*__sighandler_t hndl*/ ) {
    struct sigaction sga;
    int ret = 0;

    sga.sa_handler= hndl;   
    sigemptyset(&(sga.sa_mask));
    sga.sa_flags = 0;

    ret += sigaction(SIGUSR1,&sga,0);
    ret += sigaction(SIGUSR2,&sga,0);
    ret += sigaction(SIGALRM,&sga,0);
    ret += sigaction(SIGHUP,&sga,0);
    ret += sigaction(SIGABRT,&sga,0);
    ret += sigaction(SIGINT,&sga,0);
    ret += sigaction(SIGPIPE,&sga,0);
    ret += sigaction(SIGQUIT,&sga,0);
    ret += sigaction(SIGTERM,&sga,0);
    ret += sigaction(SIGCHLD,&sga,0);
    ret += sigaction(SIGTSTP,&sga,0);
    ret += sigaction(SIGPOLL,&sga,0);
    ret += sigaction(SIGPROF,&sga,0);
    ret += sigaction(SIGURG,&sga,0);
    ret += sigaction(SIGVTALRM,&sga,0);
   
    if( ret < 0 )
      syslog(BBFTPD_ERR,"Sigaction: %s", strerror( errno ) );
    
    return( ret );
}


int bbftpd_setdaemonsignals() {
    struct sigaction sga;
    int ret;

    ret = bbftpd_resetallsignals( SIG_IGN );
    
    sga.sa_handler = bbftpd_sigusr1;
    sigemptyset(&(sga.sa_mask));
    sga.sa_flags = 0;

    ret += sigaction(SIGUSR1,&sga,0);
    
    if( ret < 0 )
      syslog(BBFTPD_ERR,"Sigaction: %s", strerror( errno ) );
    
    return( ret );
}


int bbftpd_setfathersignals() {
    struct sigaction sga;
    int ret;

    ret = bbftpd_resetallsignals( SIG_DFL );

    sga.sa_handler = bbftpd_sigtermfather;
    sigemptyset(&(sga.sa_mask));
    sga.sa_flags = 0;

    ret += sigaction(SIGTERM,&sga,0);
    ret += sigaction(SIGINT,&sga,0);

    sga.sa_handler = SIG_IGN;
    sigemptyset(&(sga.sa_mask));
    sga.sa_flags = 0;
    
    ret += sigaction(SIGPIPE,&sga,0);
    
    if( ret < 0 )
      syslog(BBFTPD_ERR,"Sigaction: %s", strerror( errno ) );

    return( ret );
}

int bbftpd_setsonsignals() {
  struct sigaction sga;
  int ret;

  ret = bbftpd_resetallsignals( SIG_IGN );	/* SIGPIPE too */

  sga.sa_handler = bbftpd_sigtermson;
  sigemptyset(&(sga.sa_mask));
  sga.sa_flags = 0;

  ret += sigaction(SIGTERM,&sga,0);
  ret += sigaction(SIGINT,&sga,0);

  sga.sa_handler= SIG_DFL;
  sigemptyset(&(sga.sa_mask));
  sga.sa_flags = 0;
  
  ret += sigaction(SIGCHLD,&sga,0);
  
  if( ret < 0 )
    syslog(BBFTPD_ERR,"Sigaction: %s", strerror( errno ) );

  return( ret );
}

int bbftpd_setchildsignals() {
  struct sigaction sga;
  int ret;
	      
  ret = bbftpd_resetallsignals( SIG_IGN );	/* SIGPIPE too */
	        
  sga.sa_handler = bbftpd_sigtermchild;
  sigemptyset(&(sga.sa_mask));
  sga.sa_flags = 0;
		      
  ret += sigaction(SIGTERM,&sga,0);
  ret += sigaction(SIGINT,&sga,0);
				  
  if( ret < 0 )
    syslog(BBFTPD_ERR,"Sigaction: %s", strerror( errno ) );
				    
  return( ret );
}
