/*
 * bbftpd/bbftpd_socket.c
 * Copyright (C) 1999, 2000, 2001, 2002 IN2P3, CNRS
 * bbftp@in2p3.fr
 * http://doc.in2p3.fr/bbftp
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */ 

/****************************************************************************

 
 bbftpd_socket.c    v 2.0.0 2000/12/19  - Routines creation
                    v 2.0.1 2001/04/23  - Correct indentation
                                        - Port to IRIX
               
*****************************************************************************/
#include "bbftpd.h"
#include <sys/types.h>

#include <errno.h>
#include <syslog.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <netinet/tcp.h>

#include <unistd.h>

#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>

#include "common.h"
#include "daemon.h"
#include "structures.h"
#include "daemon_proto.h"


extern unsigned short int     *myports ;
extern int	*mysockets ;
extern int	transferoption ;
extern int	pasvport_min ;
extern int	pasvport_max ;

extern int	be_daemon;
extern pid_t	daemonpid;

extern int	ctrl_port;

extern int	*myips;
extern char	*localips;

extern int  requestedstreamnumber;

extern int natip;


int get_ip( char *host, char **fhost ) {
  struct sockaddr_in sn;
  char *alist[1] = { NULL };
  struct in_addr defaddr;
  struct hostent *hp, def;

  hp = gethostbyname( host );
  if( NULL == hp ) {
    defaddr.s_addr = inet_addr( host );
    if( -1 == defaddr.s_addr )
      return( INADDR_ANY );     /* zero */

    def.h_name = host;
    def.h_addr_list = alist;
    def.h_addr = (char *)&defaddr;
    def.h_length = sizeof( struct in_addr );
    def.h_addrtype = AF_INET;
    def.h_aliases = 0;
    hp = &def;
  }
  if( hp->h_length > (int)sizeof( sn.sin_addr ) )
    hp->h_length = sizeof( sn.sin_addr );
  memcpy( &sn.sin_addr, hp->h_addr, hp->h_length );

  if( fhost )
    *fhost = strdup( hp->h_name );

  return( sn.sin_addr.s_addr );
}


int *nips = NULL;
int sip = 0;

static int nextip() {
  static int no = 0;
  static int crt = 0;
  static int any = INADDR_ANY;
  int nip;

  if( 0 == no ) {
    char *ip;

    ip = strtok( localips, ", " );
    while( ip ) {
      int *ptr;

      ptr = (int *)(realloc( nips, (no + 1) * sizeof( int ) ));
      if( NULL == ptr ) {
	syslog( BBFTPD_ERR, "Realloc: %s", strerror( errno ) );
	break;
      }
      nips = ptr;
      nips[no] = get_ip( ip, NULL );
      ++no;
      ip = strtok( NULL, ", " );
    }
    if( 0 == no ) {
      sip = 1;
      nips = &any;
      ++no;
    }
  }
  nip = nips[crt];
  crt = (crt + 1) % no;
  return( nip );
}


int bbftpd_makedatasock( int nbsock ) {
  int     on = 1 ;
#if defined(SUNOS) || defined(_HPUX_SOURCE) || defined(IRIX) || defined(OSF1)
  int     addrlen ;
#else
  size_t  addrlen ;
#endif		
  int     sock ;
  struct  sockaddr_in data_source ;
  int     i ;

  unsigned short int *sports;
 
  sports = (unsigned short int *)(malloc( nbsock * sizeof( unsigned short int ) )) ; /* List of active source ports */
  if( NULL == sports ) {
	  syslog( BBFTPD_ERR, "malloc : %s", strerror( errno ) );
	  return( -1 );
  }

  for( i = 0; i < nbsock; ++i ) {
    sock = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP ) ;
    if ( sock < 0 ) {
      syslog(BBFTPD_ERR, "Cannot create receive socket : %s",strerror(errno));
      goto bye001;
    }

    /* make active and eventually reading sockets nonblock !! */
    if ( -1 == fcntl( sock, F_SETFL, O_NONBLOCK ) ) {
      syslog(BBFTPD_ERR, "fcntl : %s", strerror(errno) );
      goto bye001;
    }
   
    if ( setsockopt(sock,SOL_SOCKET, SO_KEEPALIVE,(char *)&on,sizeof(on)) < 0 ) {
      syslog(BBFTPD_ERR,"Cannot set SO_KEEPALIVE on receive socket : %s",strerror(errno));
      goto bye001;
    }

    if ( setsockopt(sock,IPPROTO_TCP, TCP_NODELAY,(char *)&on,sizeof(on)) < 0 ) {
      syslog(BBFTPD_ERR,"Cannot set TCP_NODELAY on receive socket : %s",strerror(errno)) ;
      goto bye001;
    }

    data_source.sin_family = AF_INET;
    data_source.sin_addr.s_addr = nextip();
    data_source.sin_port = 0;
    
    if ( bind( sock, (struct sockaddr *) &data_source,sizeof(data_source)) < 0) {
      syslog(BBFTPD_ERR,"Cannot bind on receive socket port : %s",strerror(errno)) ;
      goto bye001;
    }
    addrlen = sizeof(data_source) ;
    if (getsockname(sock,(struct sockaddr *)&data_source, (socklen_t *)&addrlen) < 0) {
      syslog(BBFTPD_ERR, "Cannot getsockname on data socket : %s",strerror(errno)) ;
      goto bye001;
    }

    mysockets[i]  = sock;

    sports[i] = data_source.sin_port;
  }

  {
	  int retcode;

	  retcode = writepackage( MSG_POSITION, requestedstreamnumber * sizeof( unsigned short int ), sports, requestedstreamnumber * sizeof( unsigned short int ) );
	  if( retcode < 0 ) {
		  syslog( BBFTPD_ERR, "Error sending %s message", MESSAGE_NAMES[MSG_POSITION] );
		  goto bye001;
	  }

	  syslog( BBFTPD_DEBUG, "%s, sent!", MESSAGE_NAMES[MSG_POSITION] );

	  FREE( sports )
	  return( 0 );
  }

  {
    int j;

bye001:
    for( j = 0; j < i; ++j )
      close( mysockets[j] );
    if( sock >= 0 )
      close( sock );

    FREE( sports )
    return( -1 );
  }
}

static void killer1() {
  if( -1 == kill( daemonpid, SIGUSR1 ) )      /* Open doors request */
    syslog( BBFTPD_ERR, "Kill -SIGUSR1 %d : %s", daemonpid, strerror( errno ) );
}


/*******************************************************************************
** bbftpd_getdatasock :                                                       *
**                                                                             *
**      Routine to create all the needed sockets to send data                  *
**                                                                             *
**      INPUT variable :                                                       *
**           nbsock   :  number of sockets to create    NOT MODIFIED           *
**                                                                             *
**      RETURN:                                                                *
**          -1   Creation failed unrecoverable error                           *
**           0   OK                                                            *
**                                                                             *
*******************************************************************************/
int bbftpd_getdatasock(int nbsock) 
{
    int     i ;
    int     on = 1 ;
#if defined(SUNOS) || defined(_HPUX_SOURCE) || defined(IRIX) || defined(OSF1)
    int     addrlen ;
#else
    size_t  addrlen ;
#endif
    struct  linger li ;
    struct  sockaddr_in sck ;
    int     sock ;   
    /* Declaring the variables need to change the value of the TOS */
    int qbss_value = 0x20;

    int port = pasvport_min;


    sck.sin_family = AF_INET;
    
    for( i = 0; i < nbsock; ++i ) {

	sck.sin_addr.s_addr = nextip();

        sock = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP ) ;
        if ( sock < 0 ) {
            syslog(BBFTPD_ERR, "Cannot create data socket : %s",strerror(errno));
            break;
	}

        if ( setsockopt(sock,SOL_SOCKET, SO_REUSEADDR,(char *)&on,sizeof(on)) < 0 ) {
            syslog(BBFTPD_ERR, "Cannot set SO_REUSEADDR on data socket : %s",strerror(errno));
            break;
        }

        if ( setsockopt(sock,SOL_SOCKET, SO_KEEPALIVE,(char *)&on,sizeof(on)) < 0 ) {
            syslog(BBFTPD_ERR, "Cannot set SO_KEEPALIVE on data socket : %s",strerror(errno));
            break;
        }

        if ( setsockopt(sock,IPPROTO_TCP, TCP_NODELAY,(char *)&on,sizeof(on)) < 0 ) {
            syslog(BBFTPD_ERR, "Cannot set TCP_NODELAY on data socket : %s",strerror(errno));
            break;
        }

		if( transferoption & TROPT_QBSS ) {
			/* Setting the value for IP_TOS to be 0x20 */
			if ( setsockopt(sock,IPPROTO_IP, IP_TOS, (char *)&qbss_value, sizeof(qbss_value)) < 0 ) {
                syslog(BBFTPD_WARNING, "Cannot set IP_TOS on data socket : %s",strerror(errno));
			}
		}

        li.l_onoff = 1 ;
        li.l_linger = 1 ;
        if ( setsockopt(sock,SOL_SOCKET,SO_LINGER,(char *)&li,sizeof(li)) < 0 ) {
            syslog(BBFTPD_ERR, "Cannot set SO_LINGER on data socket : %s",strerror(errno));
            break;
        }

        if( pasvport_min != 0 ) {	/* Try to bind within a range */
 	  for( ; port <= pasvport_max; ++port ) {
     	    sck.sin_port = htons( port );
 	    if( bind( sock, (struct sockaddr *)&sck, sizeof(sck) ) >= 0 ) break;
 	  }
	  if( port > pasvport_max ) {
	    if( i > 0 ) {
	      int j;
	      int k;

	      khl:
	      /* round robin reuse of sockets, no more keyhole... */
              for( j = i, k = 0; j < nbsock; ++j ) {
		      myips[j] = myips[k];
                      myports[j] = myports[k];
                      mysockets[j] = mysockets[k];
                      ++k;
                      if( k == i ) k = 0;
	      }
              return( 0 );
            }
	    syslog(BBFTPD_ERR, "Cannot bind on data socket : %s",strerror(errno));
            break;
	  }
	  else ++port;
        } 
	else if( 0 == pasvport_max ) {	/* Free ports */
	  sck.sin_port = 0;
          if( bind( sock, (struct sockaddr *)&sck, sizeof(sck) ) < 0 ) {
            syslog(BBFTPD_ERR, "Cannot bind on data socket : %s",strerror(errno));
            break;
 	  }
	} 
	else { 				/* Use control port */
	  int grace;
	
	  sck.sin_port = ctrl_port;
	  if( 2 == be_daemon )
	    killer1();
	  for( grace = 0; ; ) {
	    if( bind( sock, (struct sockaddr *)&sck, sizeof(sck) ) < 0 ) {
	      if( 2 == be_daemon ) {
		if( grace < 6000 ) {	/* sixty seconds */
		  ++grace;
		  usleep( 10000 );	/* 0.01 seconds */
		  if( 0 == grace % 3 )	/* 0.30 seconds */
		    killer1();
		  continue;
		}
		syslog( BBFTPD_DEBUG, "Bind on control port, for data : %s", strerror( errno ) );
	      }
	      syslog(BBFTPD_ERR, "Cannot bind on data socket : %s",strerror(errno));
	      
	      goto err001;
	    }
	    break;
	  }
        }
	
        addrlen = sizeof(sck) ;
        if (getsockname(sock,(struct sockaddr *)&sck, (socklen_t *)&addrlen) < 0) {
            syslog(BBFTPD_ERR, "Cannot getsockname on data socket : %s",strerror(errno));
            break;
        }
	
        if( listen( sock, nbsock - i ) < 0 ) {
	    syslog(BBFTPD_ERR, "Cannot listen on data socket : %s",strerror(errno));
            break;
        }

	mysockets[i] = sock;
        myports[i] =  sck.sin_port ;
	myips[i] = natip ? natip : sck.sin_addr.s_addr;

	/*XXX*/syslog(BBFTPD_DEBUG,"Listen(%d) on %s:%d", i, inet_ntoa(sck.sin_addr), ntohs(sck.sin_port));

        if( 0 == pasvport_min && pasvport_max != 0 ) { ++i; goto khl; }
    }

    if( i < nbsock ) {
      int j;
      
err001:
      for( j = 0; j < i; ++j )
        close( mysockets[j] );
      if( sock >= 0 )
        close( sock );

      return( -1 );
    }
    
    return( 0 );
}
