/*
 * bbftpd/bbftpd_utils.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 
#include "bbftpd.h"

#include <string.h>

/* To make sure of optional arguments; Should go to Makefile */
#include "getopt.c"

#include <stdarg.h>
#include <sys/time.h>
#include <time.h>

#include "common.h"


extern	char *curfilename;
extern	char *realfilename;
extern	unsigned short int *myports;
extern	int *mysockets;

extern	int *myips;

extern	int *nips;
extern	int sip;


void free_n_var() {
  if( 0 == sip ) FREE( nips )
  FREE( myips )
  FREE( myports )
  FREE( mysockets )
}

void free_f_var() {
  free_n_var();
  FREE( realfilename )
  FREE( curfilename )
}

char *appendxb( xbuffer *xb, char *str, int len ) {
  if( len > xb->lsize ) {
    char *ptr;
    int ext;
    int rem;

    ext = len - xb->lsize;
    rem = ext % GROWTH;
    if( rem )
      ext += GROWTH - rem;

    ptr = (char *)(malloc( xb->tsize + ext ));
    if( NULL == ptr ) {
      FREE( xb->line )
      xb->tsize = 0;
      xb->lsize = 0;
      return( NULL );
    }

    if( xb->tsize - xb->lsize )
      memcpy( ptr, xb->line, xb->tsize - xb->lsize );
    xb->tsize += ext;
    xb->lsize += ext;
    FREE( xb->line )
    xb->line = ptr;
  }
  if( len > 0 ) {
    if( str )
      memcpy( xb->line + xb->tsize - xb->lsize, str, len );
    xb->lsize -= len;
  }
  
  return( xb->line );
}

void emptyxb( xbuffer *xb ) {
  xb->lsize = xb->tsize;
}

void shrinkxb( xbuffer *xb, int kept ) {
  if( kept <= xb->tsize )
    xb->lsize = xb->tsize - kept;
}

char *chainxb( xbuffer *xb, char *str, int len ) {
  if( xb->lsize < xb->tsize )
    xb->lsize++;
  return( appendxb( xb, str, len ) );
}

char *catxb( xbuffer *xb, char *strz ) {
  if( xb->lsize < xb->tsize )
    xb->lsize++;
  return( appendxb( xb, strz, strlen( strz ) + 1 ) );
}

void freexb( xbuffer *xb ) {
  FREE( xb->line )
  xb->tsize = 0;
  xb->lsize = 0;
}

int lenxb( xbuffer *xb ) {
  return( xb->tsize - xb->lsize );
}

char *ptrxb( xbuffer *xb ) {
  return( xb->line );
}

char *printxb( xbuffer *xb, char *format, ... ) {
  va_list args;
  char *proc;
  char *s;
  int d;
  static char dtoa[128];
 
  emptyxb( xb );
 
  va_start( args, format );
  do {
    int len;
    
    proc = strchr( format, '%' );
    len = (proc ? proc - format : strlen( format ));
    if( len && NULL == appendxb( xb, format, len ) )
      break;
    if( NULL == proc ) {
      appendxb( xb, "", 1 );
      break;
    }
    format = proc + 2;
    switch( *(proc + 1) ) {
      case 's':
	s = va_arg( args, char * );
	len = strlen( s );
	if( len && NULL == appendxb( xb, s, len ) )
	  break;
	continue;
      case 'd':
	d = va_arg( args, int );
	sprintf( dtoa, "%d", d );
	len = strlen( dtoa );
	if( len && NULL == appendxb( xb, dtoa, len ) )
          break;
        continue;
      default:
	break;
    }
    break;
  }while( 1 );
  va_end( args );

  s = ptrxb( xb );
  return( s ? s : "" );
}

my64_t stopwatch( const int stsp ) {
  static struct timespec tst;

  if( 0 == stsp ) {
    struct timespec tnd;
    register struct timeval tdf;

    (void) clock_gettime( CLOCK_MONOTONIC, &tnd );
    tdf.tv_sec = tnd.tv_sec - tst.tv_sec;
    tdf.tv_usec = (tnd.tv_nsec - tst.tv_nsec) / 1000;
    if( tdf.tv_usec < 0 ) tdf.tv_sec--, tdf.tv_usec += 1000000;
    return( tdf.tv_sec * 1000000 + tdf.tv_usec );
  }

  clock_gettime( CLOCK_MONOTONIC, &tst );
  return( 0 );
}
