/*
 * chdir.c
 * Copyright (C) Dan Schrager
 * Dan.Schrager@weizmann.ac.il
 */ 

#include "bbftp.h"
#include <sys/types.h>

#include <errno.h>
#include <stdio.h>
#include <netinet/in.h>

#include <unistd.h>
#include <string.h>

#include "client.h"
#include "client_proto.h"
#include "common.h"
#include "structures.h"


#ifdef	CLIENT
extern	int	timestamp;
extern	int	verbose;
extern	int	errcode;
#else
extern	xbuffer	logmessage;
#endif	/*CLIENT*/

xbuffer thecd = { NULL, 0, 0 };


#ifdef	CLIENT
int 
#else
void
#endif	/*CLIENT*/
bbftp_cd( char *dirpath ) {
  int err;
	
#ifdef	CLIENT
  if( verbose ) 
    printmessage( stderr, CASE_NORMAL, 0, timestamp, ">> COMMAND : lcd %s\n", dirpath );
#else
  syslog( BBFTPD_DEBUG, "Changing directory to %s", dirname );
#endif	/*CLIENT*/
  
  if( CHDIR( dirpath ) < 0 ) {
	
    err = ERRNO;

#ifdef	CLIENT
    printmessage( stderr, CASE_ERROR, 71, timestamp, "Error changing directory(%s): %s\n", dirpath, strerror( err ) );
    errcode = 71;
    return( ENOMEM == err || EBUSY == err ? BB_RET_ERROR :  BB_RET_FT_NR );
#else
    printxb( &logmessage, "Chdir(%s): %s ", dirname, strerror( err) );
    syslog( BBFTPD_ERR, ptrxb( &logmessage ) );
    reply( (ENOMEM == err || EBUSY == err ? MSG_BAD : MSG_BAD_NO_RETRY), ptrxb( &logmessage ) );
    return;
#endif	/*CLIENT*/
  } 
  
  emptyxb( &thecd ); 
  for(;;) {
    if( NULL == appendxb( &thecd, NULL, 128 ) ) {

#ifdef	CLIENT
      errcode = 1;
      printmessage( stderr, CASE_ERROR, 1, timestamp, "Realloc: %s\n", strerror( errno ) );
      return( BB_RET_ERROR );
#else
      printxb( &logmessage, "Realloc: %s", strerror( errno ) );
      mal:
      syslog( BBFTPD_ERR, ptrxb( &logmessage ) );
      reply( MSG_BAD, ptrxb( &logmessage ) );
      return;
#endif	/*CLIENT*/
    }
			          
    if( NULL == GETCWD( ptrxb( &thecd ), lenxb( &thecd ) ) ) {
          
      err = ERRNO;
      if( ERANGE == errno )
        continue;

#ifdef	CLIENT      
      printmessage( stderr, CASE_ERROR, 71, timestamp, "Getcwd : %s\n", strerror( err ) );
      errcode = 71;
#else
      printxb( &logmessage, "Getcwd : %s ", strerror( err ) );
#endif	/*CLIENT*/
	
      emptyxb( &thecd );
      
#ifdef	CLIENT
      return( BB_RET_FT_NR );
#else
      goto mal;
#endif	/*CLIENT*/
      
    }
    break;
  }
    
#ifdef	CLIENT
  if( verbose ) 
    printmessage( stderr, CASE_NORMAL, 0, timestamp, "<< OK : Current local directory is %s\n", ptrxb( &thecd ) );
  return( BB_RET_OK );
#else
  reply( MSG_OK, ptrxb( &thecd ) );
#endif	/*CLIENT*/
}
