/*
 * includes/client_proto.h
 * Copyright (C) 1999, 2000, 2001, 2002 IN2P3, CNRS
 * bbftp@in2p3.fr
 * http://doc.in2p3.fr/bbftp
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */ 

/****************************************************************************

  
  
 client_proto.h v 2.0.0 2001/03/21    - Header creation

*****************************************************************************/

#include <utime.h>
#include <stdio.h>

#ifdef WITH_SSL
#include <openssl/rsa.h>
#include <openssl/err.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <openssl/rand.h>
#include <openssl/md5.h>
#include "openssl.h"
#endif /* WITH_SSL */

/*
** Prototype for store routines
*/
int bbftp_storecheckfile( char *filename );
int bbftp_storechmod( char *filename, int mode );
int bbftp_storeclosecastfile( char *filename );
int bbftp_storecreatefile( char *filename );
int bbftp_storemkdir( char *dirname );
int bbftp_storerename( char *oldpath, char *newpath );
int bbftp_storemklnk( char *oldpath, char *newpath );
int bbftp_storeunlink( char *filename );

/*
** Prototype for retr routines
*/
int bbftp_retrlistdir( char *prefixdir, char *pattern, char **filelist, int *filelistlen );
int bbftp_retrcheckfile( char *filename );

int start_pumps();

my64_t filelength( char *filename );

/*
** Prototype for mains routines
*/
int bbftp_cd( char *dirpath );
int bbftp_lcd( char *dirpath );
int bbftp_list( char *prefixdir, char *pattern, char **filelist, int *filelistlen );
int bbftp_get( char *remotefilename );
int bbftp_mget( char *remotefile, char *localdir );
int bbftp_mkdir( char *dirpath );
int bbftp_mput( char *localfile, char *remotedir );
int bbftp_put( char *remotefilename );
int bbftp_setremotecos( int cos );
int bbftp_setremoteumask( int mask );
int bbftp_rget( char *line );
int bbftp_rput( char *line );
/*
** Prototype for utilities
*/
void bbftp_close_control()  ;
void bbftp_free_all_var()  ;
void bbftp_clean_child();
void printmessage(FILE *strm , int flag, int errcode, int tok, char *fmt, ...) ;
void Usage() ;
/*
** Prototype for connection routines
*/
int connectviassh() ;
int connectviapassword() ;
int todoafterconnection();
void reconnecttoserver() ;
/*
** Prototype for sockets routines
*/
int bbftp_connect( int sock, int port, int ip );

int getdatasock( int nbsock );

int readmessage( char *buffer, int msglen, int to );

int sendproto() ;

int writemessage( char *buffer, int msglen, int to );

unsigned char *readpackage( int *length, int *code, unsigned char *bucket );

int writepackage( int code, int msglen, ... );

int handlepackage( int handle );
int bbftp_setfathersignals();
int get_ip( char *host, char **fhost );
void interpreter( int handle, int rhandle, char *bbftpcmd );
void stir();
int bbftp_storeutime( char *filename, struct utimbuf *ftime );
int makedatasock( int nbsock/*, unsigned short int *localports*/ );
int readpipe( int p, char *buffer, int len );
int writepipe( int p, char *buffer, int len );
void bbftp_free_f_var();
int dochildren();
int sendattributes( char *remotefilename );
int startgame();
int bbftp_setchildsignals();
int bbftp_setsonsignals();
void bbftp_free_n_var();
void endgame();
int bbftp_delete( char *path );
int bbftp_mklnk( char *oldpath, char *newpath );
int bbftp_rmdir( char *path );

#ifdef WITH_SSL
RSA *key_demote(RSA *k);
RSA *key_load_private(const char *filename, const char *passphrase, char **commentp);
RSA *key_new();
u_char *respond_to_challenge(BIGNUM * challenge);
void recrypt();
char *readpassphrase(const char *prompt, char *buf, size_t bufsiz, int flags);
#endif /* WITH_SSL */
