/*
 * includes/structures.h
 * Copyright (C) 1999, 2000, 2001, 2002 IN2P3, CNRS
 * bbftp@in2p3.fr
 * http://doc.in2p3.fr/bbftp
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */ 

/****************************************************************************

  
  
 structures.h v 0.0.0  1999/11/24
              v 1.2.0  2000/11/24
              v 1.4.0  2000/03/22   - Add message for crypt mode for
                                      username and password
              v 1.6.0  2000/03/24   - New parameters for get implementation
              v 1.6.1  2000/03/30   - Portage to OSF1
              v 1.8.0  2000/04/14   - Introduce RSA Cryptage
              v 1.8.10 2000/08/10   - Portage to Linux
              v 1.9.0  2000/08/18   - Use configure to help portage
              v 2.0.0  2000/12/13   - Introduce V2 protocol
              v 2.0.2  2001/05/04   - Add padding on mess_store in order to avoid usage
                                      of malign-double on Linux
              v 2.1.0  2001/05/25   - Add private authentication
			  v 2.2.0  2001/10/03   - Add certificate authentication message
			  
*****************************************************************************/

#include <netinet/in.h>

/*
** BBFTP V1 protocol definitions
*/

#define MAXLEN  200
#define MAXLENFILE 500		/* do not touch */
#define MAXPORT 10
#define READBUFLEN 256000

/* Unified definitions */

extern const char * const MESSAGE_NAMES[];

#define	MESSAGE_LIST		\
	T(MSG_OK)		\
	T(MSG_LOG)		\
	T(MSG_STORE)		\
	T(MSG_BAD)		\
	T(MSG_ACK)		\
	T(MSG_STORE_C)		\
	T(MSG_CRYPT)		\
	T(MSG_RETR)		\
	T(MSG_RETR_C)		\
	T(MSG_RETR_OK)		\
	T(MSG_RETR_RFIO)	\
	T(MSG_RETR_RFIO_C)	\
	T(MSG_RETR_START)	\
	T(MSG_ABR)		\
	T(MSG_BAD_NO_RETRY)	\
	T(MSG_MKDIR)		\
	T(MSG_CHDIR)		\
	T(MSG_LIST)		\
	T(MSG_LIST_REPL)	\
	T(MSG_CREATE_ZERO)	\
	T(MSG_LOGGED_STDIN)	\
	T(MSG_PROT)		\
	T(MSG_STORE_V2)		\
	T(MSG_RETR_V2)		\
	T(MSG_ABORT)		\
	T(MSG_TRANS_OK_V2)	\
	T(MSG_TRANS_START_V2)	\
	T(MSG_INFO)		\
	T(MSG_FILENAME)		\
	T(MSG_CHUMASK)		\
	T(MSG_MKDIR_V2)		\
	T(MSG_CHDIR_V2)		\
	T(MSG_LIST_V2)		\
	T(MSG_LIST_REPL_V2)	\
	T(MSG_CHCOS)		\
	T(MSG_PRIV_LOG)		\
	T(MSG_PRIV_DATA)	\
	T(MSG_CLOSE_CONN)	\
	T(MSG_CERT_LOG)		\
	T(MSG_TRANS_START_V3)	\
	T(MSG_TRANS_OK_V3)	\
	T(MSG_PORTS)		\
	T(MSG_USER)		\
	T(MSG_PASS)		\
	T(MSG_NKEY)		\
	T(MSG_KEY)		\
	T(MSG_CHALLENGE)	\
	T(MSG_RESPONSE)		\
	T(MSG_HIDDEN)		\
	T(MSG_MKLNK)		\
	T(MSG_GOAHEAD)		\
	T(MSG_DELETE)		\
	T(MSG_RMDIR)		\
	T(MSG_POSITION)		\
	T(MSG_NONE)

/*
** Message code 
*/

#define	T(x)	x,
enum MESSAGE_TYPE {
	MESSAGE_LIST
};
#undef	T

#define	MSG_WARN	 MSG_LOGGED_STDIN	/* overlapped */

#define VALID_MESSAGE(x)	((x) >= 0 && (x) <= MSG_NONE)

/*
** End Message code ************************************************************
*/

/*
** Crypt type
*/

enum CRYPT_TYPE {
	NO_CRYPT,			/* No encoding */
	CRYPT_BR,			/* Simple "brouillage" */
	CRYPT_RSA_PKCS1_OAEP_PADDING	/* Cryptage RSA */
};

enum DATA_TYPE {
	DATA_NOCOMPRESS,
	DATA_COMPRESS
};

#define NBYTESRSA	128
/*
** End Crypt type **************************************************************
*/

/*
** Structures 
*/
/*
** message is only used as the minimum structure to get a message 
*/

struct message {
	int	code ;			/* Message code */
	int	msglen ;		/* Message length in bytes */
} ;

struct mess_rsa {
	unsigned char	cryptuser[NBYTESRSA] ;	
	unsigned char	cryptpass[NBYTESRSA] ;	
	int		numuser ;
	int		numpass ;
} ;

struct mess_store {
    char    filename[MAXLENFILE] ;		/* Filename to store or to get*/
    char    pad[4] ;          /*   Pad to be on a double boundary if MAXLENFILE = 500 */
    my64_t  filesize ;			/* Size of the file (0 for a get) */
    int     blocksize ;			/* Blocksize (not used)*/
    struct  sockaddr_in remote_host ;	/* Adresse of remote host */
    int     nbport ;			/* Number of ports */
    int     port[MAXPORT] ;			/* Ports Number */
} ;

struct mess_sec {
	int	crtype ;
	int pubkeylen ;
	int expolen ;
} ;

struct mess_retr_ok {
	my64_t	filesize ;
} ;

#define USERMESS struct mess_user
#define USERMESSLEN sizeof(USERMESS)

#define MINMESS  struct message
#define RSAMESS struct mess_rsa
#define STORMESS struct mess_store
#define COMPMESS struct mess_compress
#define CRYPTMESS struct mess_sec
#define RETROKMESS struct mess_retr_ok
#define MINMESSLEN sizeof(MINMESS)
#define RSAMESSLEN sizeof(RSAMESS)
#define STORMESSLEN sizeof(STORMESS)
#define COMPMESSLEN sizeof(COMPMESS)
#define CRYPTMESSLEN sizeof(CRYPTMESS)
#define RETROKMESSLEN sizeof(RETROKMESS)
#define MAXMESSLEN STORMESSLEN
/*
** End Structures **************************************************************
*/

/*
** BBFTP V2 protocol definition
*/

/*
** Crypt type are the same the version 1
*/
/*
** End Crypt type **************************************************************
*/

/*
** Structures 
*/
/*
* The following structures from version 1 are still used
*			message
*			mess_rsa
*			mess_compress
*			mess_sec
*/
struct mess_store_v2 {
	my64_t	filesize ;			/* file size in bytes */
	int		transferoption ;	/* transfer otions : see TROPT_XXX */
	int		filemode ;			/* file mode */
	char	lastaccess[8] ;		/* last access in hexa form */
	char	lastmodif[8] ;		/* last modif in hexa form */
	int		sendwinsize ;		/* Send window size in KB */
	int		recvwinsize ;		/* Receive window size in KB */
	int		buffersizeperstream ;	/* Buffer size per stream */
	int		nbstream ;			/* Number of stream */
} ;

struct mess_dir {
    int     transferoption ;
    char    dirname[1] ;
} ;

struct mess_private {
	unsigned char	cryptdata[NBYTESRSA] ;	
	int		lengthdata ;
} ;

	
#define STORMESS_V2 struct mess_store_v2
#define STORMESSLEN_V2 sizeof(STORMESS_V2)
#define PRIVRSAMESS struct mess_private
#define PRIVRSAMESSLEN sizeof(PRIVRSAMESS)

/*
** End Structures **************************************************************
*/
/*
** Define options bits
*/

enum TROPT_TYPE {
	TROPT_TMP =	1,	/* Always use temporary name */
	TROPT_ACC =	2,	/* Keep Access and modification time  */
	TROPT_MODE =	4,	/* Keep Mode */
	TROPT_DIR =	8,	/* Create all dir in path */
	TROPT_GZIP =	16,	/* Compress with GZIP algorythme */
	TROPT_RFIO =	32,	/* RFIO on the daemon side, OBSOLETE */
	TROPT_RFIO_O =	64,	/* RFIO on the client side, OBSOLETE */
	TROPT_QBSS =	128,	/* Mark traffic for QBSS */
	TROPT_HIDDEN =  256,	/* Data full encryption */
	TROPT_RECS =    512,	/* Recursively */
	TROPT_EXCL =	1024,	/* Exclude already transfered files */
	TROPT_UNIX =	2048,	/* On the fly tar XXX */
	TROPT_PIPE =	4096,	/* Pipe XXX */
	TROPT_MPATH =	8192	/* Multipath */
};

/*
** End Define options bits *****************************************************
*/
